@@ -17,11 +17,12 @@ Pluggable Back-end for Account Server
 """
 
 from uuid import uuid4
-import six.moves.cPickle as pickle
 
 import sqlite3
 
-from swift.common.utils import Timestamp
+import six
+
+from swift.common.utils import Timestamp, RESERVED_BYTE
 from swift.common.db import DatabaseBroker, utf8encode, zero_like
 
 DATADIR = 'accounts'
@@ -188,28 +189,13 @@ class AccountBroker(DatabaseBroker):
                 self._db_version = 1
         return self._db_version
 
-    def _delete_db(self, conn, timestamp, force=False):
-        """
-        Mark the DB as deleted.
-
-        :param conn: DB connection object
-        :param timestamp: timestamp to mark as deleted
-        """
-        conn.execute("""
-            UPDATE account_stat
-            SET delete_timestamp = ?,
-                status = 'DELETED',
-                status_changed_at = ?
-            WHERE delete_timestamp < ? """, (timestamp, timestamp, timestamp))
-
     def _commit_puts_load(self, item_list, entry):
         """See :func:`swift.common.db.DatabaseBroker._commit_puts_load`"""
-        loaded = pickle.loads(entry.decode('base64'))
         # check to see if the update includes policy_index or not
         (name, put_timestamp, delete_timestamp, object_count, bytes_used,
-         deleted) = loaded[:6]
-        if len(loaded) > 6:
-            storage_policy_index = loaded[6]
+         deleted) = entry[:6]
+        if len(entry) > 6:
+            storage_policy_index = entry[6]
         else:
             # legacy support during upgrade until first non legacy storage
             # policy is defined
@@ -246,7 +232,7 @@ class AccountBroker(DatabaseBroker):
         """
         Create a container with the given attributes.
 
-        :param name: name of the container to create
+        :param name: name of the container to create (a native string)
         :param put_timestamp: put_timestamp of the container to create
         :param delete_timestamp: delete_timestamp of the container to create
         :param object_count: number of objects in the container
@@ -369,7 +355,7 @@ class AccountBroker(DatabaseBroker):
             ''').fetchone())
 
     def list_containers_iter(self, limit, marker, end_marker, prefix,
-                             delimiter, reverse=False):
+                             delimiter, reverse=False, allow_reserved=False):
         """
         Get a list of containers sorted by name starting at marker onward, up
         to limit entries. Entries will begin with the prefix and will not have
@@ -381,13 +367,15 @@ class AccountBroker(DatabaseBroker):
         :param prefix: prefix query
         :param delimiter: delimiter for query
         :param reverse: reverse the result order.
+        :param allow_reserved: exclude names with reserved-byte by default
 
         :returns: list of tuples of (name, object_count, bytes_used,
                   put_timestamp, 0)
         """
         delim_force_gte = False
-        (marker, end_marker, prefix, delimiter) = utf8encode(
-            marker, end_marker, prefix, delimiter)
+        if six.PY2:
+            (marker, end_marker, prefix, delimiter) = utf8encode(
+                marker, end_marker, prefix, delimiter)
         if reverse:
             # Reverse the markers if we are reversing the listing.
             marker, end_marker = end_marker, marker
@@ -417,12 +405,15 @@ class AccountBroker(DatabaseBroker):
                     query_args.append(marker)
                     # Always set back to False
                     delim_force_gte = False
-                elif marker and marker >= prefix:
+                elif marker and (not prefix or marker >= prefix):
                     query += ' name > ? AND'
                     query_args.append(marker)
                 elif prefix:
                     query += ' name >= ? AND'
                     query_args.append(prefix)
+                if not allow_reserved:
+                    query += ' name >= ? AND'
+                    query_args.append(chr(ord(RESERVED_BYTE) + 1))
                 if self.get_db_version(conn) < 1:
                     query += ' +deleted = 0'
                 else:
@@ -454,14 +445,18 @@ class AccountBroker(DatabaseBroker):
                         curs.close()
                         return results
                     end = name.find(delimiter, len(prefix))
-                    if end > 0:
+                    if end >= 0:
                         if reverse:
-                            end_marker = name[:end + 1]
+                            end_marker = name[:end + len(delimiter)]
                         else:
-                            marker = name[:end] + chr(ord(delimiter) + 1)
+                            marker = ''.join([
+                                name[:end],
+                                delimiter[:-1],
+                                chr(ord(delimiter[-1:]) + 1),
+                            ])
                             # we want result to be inclusive of delim+1
                             delim_force_gte = True
-                        dir_name = name[:end + 1]
+                        dir_name = name[:end + len(delimiter)]
                         if dir_name != orig_marker:
                             results.append([dir_name, 0, 0, '0', 1])
                         curs.close()
