@@ -16,7 +16,6 @@
 import os
 import random
 import socket
-from swift import gettext_ as _
 from logging import DEBUG
 from math import sqrt
 from time import time
@@ -32,6 +31,7 @@ from swift.common.constraints import check_drive
 from swift.common.direct_client import direct_delete_container, \
     direct_delete_object, direct_get_container
 from swift.common.exceptions import ClientException
+from swift.common.request_helpers import USE_REPLICATION_NETWORK_HEADER
 from swift.common.ring import Ring
 from swift.common.ring.utils import is_local_device
 from swift.common.utils import get_logger, whataremyips, config_true_value, \
@@ -142,10 +142,10 @@ class AccountReaper(Daemon):
                     continue
                 self.reap_device(device)
         except (Exception, Timeout):
-            self.logger.exception(_("Exception in top-level account reaper "
-                                    "loop"))
+            self.logger.exception("Exception in top-level account reaper "
+                                  "loop")
         elapsed = time() - begin
-        self.logger.info(_('Devices pass completed: %.02fs'), elapsed)
+        self.logger.info('Devices pass completed: %.02fs', elapsed)
 
     def reap_device(self, device):
         """
@@ -194,7 +194,8 @@ class AccountReaper(Daemon):
                         elif fname.endswith('.db'):
                             self.start_time = time()
                             broker = \
-                                AccountBroker(os.path.join(hsh_path, fname))
+                                AccountBroker(os.path.join(hsh_path, fname),
+                                              logger=self.logger)
                             if broker.is_status_deleted() and \
                                     not broker.empty():
                                 self.reap_account(
@@ -255,22 +256,22 @@ class AccountReaper(Daemon):
                 self.delay_reaping:
             return False
         account = info['account']
-        self.logger.info(_('Beginning pass on account %s'), account)
+        self.logger.info('Beginning pass on account %s', account)
         self.reset_stats()
         container_limit = 1000
         if container_shard is not None:
             container_limit *= len(nodes)
         try:
-            marker = ''
-            while True:
-                containers = \
-                    list(broker.list_containers_iter(container_limit, marker,
-                                                     None, None, None))
-                if not containers:
-                    break
+            containers = list(broker.list_containers_iter(
+                container_limit, '', None, None, None, allow_reserved=True))
+            while containers:
                 try:
                     for (container, _junk, _junk, _junk, _junk) in containers:
-                        this_shard = int(md5(container).hexdigest(), 16) % \
+                        if six.PY3:
+                            container_ = container.encode('utf-8')
+                        else:
+                            container_ = container
+                        this_shard = int(md5(container_).hexdigest(), 16) % \
                             len(nodes)
                         if container_shard not in (this_shard, None):
                             continue
@@ -280,43 +281,45 @@ class AccountReaper(Daemon):
                     self.container_pool.waitall()
                 except (Exception, Timeout):
                     self.logger.exception(
-                        _('Exception with containers for account %s'), account)
-                marker = containers[-1][0]
-                if marker == '':
-                    break
-            log = 'Completed pass on account %s' % account
+                        'Exception with containers for account %s', account)
+                containers = list(broker.list_containers_iter(
+                    container_limit, containers[-1][0], None, None, None,
+                    allow_reserved=True))
+            log_buf = ['Completed pass on account %s' % account]
         except (Exception, Timeout):
-            self.logger.exception(
-                _('Exception with account %s'), account)
-            log = _('Incomplete pass on account %s') % account
+            self.logger.exception('Exception with account %s', account)
+            log_buf = ['Incomplete pass on account %s' % account]
         if self.stats_containers_deleted:
-            log += _(', %s containers deleted') % self.stats_containers_deleted
+            log_buf.append(', %s containers deleted' %
+                           self.stats_containers_deleted)
         if self.stats_objects_deleted:
-            log += _(', %s objects deleted') % self.stats_objects_deleted
+            log_buf.append(', %s objects deleted' % self.stats_objects_deleted)
         if self.stats_containers_remaining:
-            log += _(', %s containers remaining') % \
-                self.stats_containers_remaining
+            log_buf.append(', %s containers remaining' %
+                           self.stats_containers_remaining)
         if self.stats_objects_remaining:
-            log += _(', %s objects remaining') % self.stats_objects_remaining
+            log_buf.append(', %s objects remaining' %
+                           self.stats_objects_remaining)
         if self.stats_containers_possibly_remaining:
-            log += _(', %s containers possibly remaining') % \
-                self.stats_containers_possibly_remaining
+            log_buf.append(', %s containers possibly remaining' %
+                           self.stats_containers_possibly_remaining)
         if self.stats_objects_possibly_remaining:
-            log += _(', %s objects possibly remaining') % \
-                self.stats_objects_possibly_remaining
+            log_buf.append(', %s objects possibly remaining' %
+                           self.stats_objects_possibly_remaining)
         if self.stats_return_codes:
-            log += _(', return codes: ')
+            log_buf.append(', return codes: ')
             for code in sorted(self.stats_return_codes):
-                log += '%s %sxxs, ' % (self.stats_return_codes[code], code)
-            log = log[:-2]
-        log += _(', elapsed: %.02fs') % (time() - begin)
-        self.logger.info(log)
+                log_buf.append('%s %sxxs, ' % (self.stats_return_codes[code],
+                                               code))
+            log_buf[-1] = log_buf[-1][:-2]
+        log_buf.append(', elapsed: %.02fs' % (time() - begin))
+        self.logger.info(''.join(log_buf))
         self.logger.timing_since('timing', self.start_time)
         delete_timestamp = Timestamp(info['delete_timestamp'])
         if self.stats_containers_remaining and \
            begin - float(delete_timestamp) >= self.reap_not_done_after:
             self.logger.warning(
-                _('Account %(account)s has not been reaped since %(time)s') %
+                'Account %(account)s has not been reaped since %(time)s' %
                 {'account': account, 'time': delete_timestamp.isoformat})
         return True
 
@@ -368,21 +371,22 @@ class AccountReaper(Daemon):
                     node, part, account, container,
                     marker=marker,
                     conn_timeout=self.conn_timeout,
-                    response_timeout=self.node_timeout)
+                    response_timeout=self.node_timeout,
+                    headers={USE_REPLICATION_NETWORK_HEADER: 'true'})
                 self.stats_return_codes[2] = \
                     self.stats_return_codes.get(2, 0) + 1
                 self.logger.increment('return_codes.2')
             except ClientException as err:
                 if self.logger.getEffectiveLevel() <= DEBUG:
                     self.logger.exception(
-                        _('Exception with %(ip)s:%(port)s/%(device)s'), node)
+                        'Exception with %(ip)s:%(port)s/%(device)s', node)
                 self.stats_return_codes[err.http_status // 100] = \
                     self.stats_return_codes.get(err.http_status // 100, 0) + 1
                 self.logger.increment(
                     'return_codes.%d' % (err.http_status // 100,))
-            except (Timeout, socket.error) as err:
+            except (Timeout, socket.error):
                 self.logger.error(
-                    _('Timeout Exception with %(ip)s:%(port)s/%(device)s'),
+                    'Timeout Exception with %(ip)s:%(port)s/%(device)s',
                     node)
             if not objects:
                 break
@@ -393,20 +397,15 @@ class AccountReaper(Daemon):
                     self.logger.error('ERROR: invalid storage policy index: %r'
                                       % policy_index)
                 for obj in objects:
-                    if isinstance(obj['name'], six.text_type):
-                        obj['name'] = obj['name'].encode('utf8')
                     pool.spawn(self.reap_object, account, container, part,
                                nodes, obj['name'], policy_index)
                 pool.waitall()
             except (Exception, Timeout):
-                self.logger.exception(_('Exception with objects for container '
-                                        '%(container)s for account %(account)s'
-                                        ),
+                self.logger.exception('Exception with objects for container '
+                                      '%(container)s for account %(account)s',
                                       {'container': container,
                                        'account': account})
             marker = objects[-1]['name']
-            if marker == '':
-                break
         successes = 0
         failures = 0
         timestamp = Timestamp.now()
@@ -421,7 +420,8 @@ class AccountReaper(Daemon):
                              'X-Account-Partition': str(account_partition),
                              'X-Account-Device': anode['device'],
                              'X-Account-Override-Deleted': 'yes',
-                             'X-Timestamp': timestamp.internal})
+                             'X-Timestamp': timestamp.internal,
+                             USE_REPLICATION_NETWORK_HEADER: 'true'})
                 successes += 1
                 self.stats_return_codes[2] = \
                     self.stats_return_codes.get(2, 0) + 1
@@ -429,16 +429,16 @@ class AccountReaper(Daemon):
             except ClientException as err:
                 if self.logger.getEffectiveLevel() <= DEBUG:
                     self.logger.exception(
-                        _('Exception with %(ip)s:%(port)s/%(device)s'), node)
+                        'Exception with %(ip)s:%(port)s/%(device)s', node)
                 failures += 1
                 self.logger.increment('containers_failures')
                 self.stats_return_codes[err.http_status // 100] = \
                     self.stats_return_codes.get(err.http_status // 100, 0) + 1
                 self.logger.increment(
                     'return_codes.%d' % (err.http_status // 100,))
-            except (Timeout, socket.error) as err:
+            except (Timeout, socket.error):
                 self.logger.error(
-                    _('Timeout Exception with %(ip)s:%(port)s/%(device)s'),
+                    'Timeout Exception with %(ip)s:%(port)s/%(device)s',
                     node)
                 failures += 1
                 self.logger.increment('containers_failures')
@@ -497,7 +497,8 @@ class AccountReaper(Daemon):
                              'X-Container-Partition': str(container_partition),
                              'X-Container-Device': cnode['device'],
                              'X-Backend-Storage-Policy-Index': policy_index,
-                             'X-Timestamp': timestamp.internal})
+                             'X-Timestamp': timestamp.internal,
+                             USE_REPLICATION_NETWORK_HEADER: 'true'})
                 successes += 1
                 self.stats_return_codes[2] = \
                     self.stats_return_codes.get(2, 0) + 1
@@ -505,18 +506,18 @@ class AccountReaper(Daemon):
             except ClientException as err:
                 if self.logger.getEffectiveLevel() <= DEBUG:
                     self.logger.exception(
-                        _('Exception with %(ip)s:%(port)s/%(device)s'), node)
+                        'Exception with %(ip)s:%(port)s/%(device)s', node)
                 failures += 1
                 self.logger.increment('objects_failures')
                 self.stats_return_codes[err.http_status // 100] = \
                     self.stats_return_codes.get(err.http_status // 100, 0) + 1
                 self.logger.increment(
                     'return_codes.%d' % (err.http_status // 100,))
-            except (Timeout, socket.error) as err:
+            except (Timeout, socket.error):
                 failures += 1
                 self.logger.increment('objects_failures')
                 self.logger.error(
-                    _('Timeout Exception with %(ip)s:%(port)s/%(device)s'),
+                    'Timeout Exception with %(ip)s:%(port)s/%(device)s',
                     node)
             if successes > failures:
                 self.stats_objects_deleted += 1
