@@ -146,7 +146,7 @@ be performed in isolation or, as shown below, using a single command.
 All three steps may be performed with one sub-command::
 
     $ swift-manage-shard-ranges <path_to_db> find_and_replace 500000 --enable \
-    > --force
+--force
     Loaded db broker for AUTH_test/c1.
     No shard ranges found to delete.
     Injected 7 shard ranges.
@@ -158,6 +158,7 @@ All three steps may be performed with one sub-command::
 from __future__ import print_function
 import argparse
 import json
+import os.path
 import sys
 import time
 
@@ -171,7 +172,7 @@ from swift.container.sharder import make_shard_ranges, sharding_enabled, \
 
 def _load_and_validate_shard_data(args):
     try:
-        with open(args.input, 'rb') as fd:
+        with open(args.input, 'r') as fd:
             try:
                 data = json.load(fd)
                 if not isinstance(data, list):
@@ -329,7 +330,7 @@ def delete_shard_ranges(broker, args):
     return 0
 
 
-def _replace_shard_ranges(broker, args, shard_data, timeout=None):
+def _replace_shard_ranges(broker, args, shard_data, timeout=0):
     own_shard_range = _check_own_shard_range(broker, args)
     shard_ranges = make_shard_ranges(
         broker, shard_data, args.shards_account_prefix)
@@ -342,7 +343,9 @@ def _replace_shard_ranges(broker, args, shard_data, timeout=None):
 
     # Crank up the timeout in an effort to *make sure* this succeeds
     with broker.updated_timeout(max(timeout, args.replace_timeout)):
-        delete_shard_ranges(broker, args)
+        delete_status = delete_shard_ranges(broker, args)
+        if delete_status != 0:
+            return delete_status
         broker.merge_shard_ranges(shard_ranges)
 
     print('Injected %d shard ranges.' % len(shard_ranges))
@@ -435,10 +438,10 @@ def _add_enable_args(parser):
 def _make_parser():
     parser = argparse.ArgumentParser(description='Manage shard ranges')
     parser.add_argument('container_db')
-    parser.add_argument('--verbose', '-v', action='count',
+    parser.add_argument('--verbose', '-v', action='count', default=0,
                         help='Increase output verbosity')
     subparsers = parser.add_subparsers(
-        help='Sub-command help', title='Sub-commands')
+        dest='subcommand', help='Sub-command help', title='Sub-commands')
 
     # find
     find_parser = subparsers.add_parser(
@@ -503,10 +506,24 @@ def _make_parser():
 def main(args=None):
     parser = _make_parser()
     args = parser.parse_args(args)
+    if not args.subcommand:
+        # On py2, subparsers are required; on py3 they are not; see
+        # https://bugs.python.org/issue9253. py37 added a `required` kwarg
+        # to let you control it, but prior to that, there was no choice in
+        # the matter. So, check whether the destination was set and bomb
+        # out if not.
+        parser.print_help()
+        print('\nA sub-command is required.')
+        return 1
     logger = get_logger({}, name='ContainerBroker', log_to_console=True)
-    broker = ContainerBroker(args.container_db, logger=logger,
-                             skip_commits=True)
-    broker.get_info()
+    broker = ContainerBroker(os.path.realpath(args.container_db),
+                             logger=logger, skip_commits=True)
+    try:
+        broker.get_info()
+    except Exception as exc:
+        print('Error opening container DB %s: %s' % (args.container_db, exc),
+              file=sys.stderr)
+        return 2
     print('Loaded db broker for %s.' % broker.path, file=sys.stderr)
     return args.func(broker, args)
 
