@@ -169,8 +169,7 @@ class SwiftRecon(object):
             ret_dict = {'low': min(cstats), 'high': max(cstats),
                         'total': sum(cstats), 'reported': len(cstats),
                         'number_none': len(stats) - len(cstats), 'name': name}
-            ret_dict['average'] = \
-                ret_dict['total'] / float(len(cstats))
+            ret_dict['average'] = ret_dict['total'] / float(len(cstats))
             ret_dict['perc_none'] = \
                 ret_dict['number_none'] * 100.0 / len(stats)
         else:
@@ -222,7 +221,7 @@ class SwiftRecon(object):
         Compare ring md5sum's with those on remote host
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         :param swift_dir: The local directory with the ring files.
         """
         matches = 0
@@ -276,7 +275,7 @@ class SwiftRecon(object):
         Compare swift.conf md5sum with that on remote hosts
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         :param printfn: function to print text; defaults to print()
         """
         matches = 0
@@ -308,7 +307,7 @@ class SwiftRecon(object):
         Obtain and print async pending statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         scan = {}
         recon = Scout("async", self.verbose, self.suppress_errors,
@@ -330,7 +329,7 @@ class SwiftRecon(object):
         Obtain and print drive audit error statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)]
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)]
         """
         scan = {}
         recon = Scout("driveaudit", self.verbose, self.suppress_errors,
@@ -352,7 +351,7 @@ class SwiftRecon(object):
         Check for and print unmounted drives
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         unmounted = {}
         errors = {}
@@ -385,7 +384,7 @@ class SwiftRecon(object):
         Check for server types on the ring
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         errors = {}
         recon = Scout("server_type_check", self.verbose, self.suppress_errors,
@@ -408,7 +407,7 @@ class SwiftRecon(object):
         Obtain and print expirer statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         stats = {'object_expiration_pass': [], 'expired_last_pass': []}
         recon = Scout("expirer/%s" % self.server_type, self.verbose,
@@ -437,7 +436,7 @@ class SwiftRecon(object):
         Obtain and print replication statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         stats = {'replication_time': [], 'failure': [], 'success': [],
                  'attempted': []}
@@ -460,6 +459,8 @@ class SwiftRecon(object):
                         stats[stat_key].append(repl_stats.get(stat_key))
                 last = response.get('replication_last',
                                     response.get('object_replication_last', 0))
+                if last is None:
+                    continue
                 if last < least_recent_time:
                     least_recent_time = last
                     least_recent_url = url
@@ -503,7 +504,7 @@ class SwiftRecon(object):
         Obtain and print updater statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         stats = []
         recon = Scout("updater/%s" % self.server_type, self.verbose,
@@ -530,7 +531,7 @@ class SwiftRecon(object):
         Obtain and print obj auditor statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         scan = {}
         adone = '%s_auditor_pass_completed' % self.server_type
@@ -602,7 +603,7 @@ class SwiftRecon(object):
         Obtain and print obj auditor statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         all_scan = {}
         zbf_scan = {}
@@ -678,7 +679,7 @@ class SwiftRecon(object):
         Obtain and print load average statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         load1 = {}
         load5 = {}
@@ -707,7 +708,7 @@ class SwiftRecon(object):
         Obtain and print quarantine statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         objq = {}
         conq = {}
@@ -741,7 +742,7 @@ class SwiftRecon(object):
         Obtain and print /proc/net/sockstat statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         inuse4 = {}
         mem = {}
@@ -775,7 +776,7 @@ class SwiftRecon(object):
         Obtain and print disk usage statistics
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         stats = {}
         highs = []
@@ -874,7 +875,7 @@ class SwiftRecon(object):
         Check a time synchronization of hosts with current time
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         :param jitter: Maximal allowed time jitter
         """
 
@@ -913,7 +914,7 @@ class SwiftRecon(object):
         Check OS Swift version of hosts. Inform if differs.
 
         :param hosts: set of hosts to check. in the format of:
-            set([('127.0.0.1', 6020), ('127.0.0.2', 6030)])
+            set([('127.0.0.1', 6220), ('127.0.0.2', 6230)])
         """
         versions = set()
         errors = 0
@@ -986,7 +987,8 @@ class SwiftRecon(object):
                         help="Print verbose info")
         args.add_option('--suppress', action="store_true",
                         help="Suppress most connection related errors")
-        args.add_option('--async', '-a', action="store_true",
+        args.add_option('--async', '-a',
+                        action="store_true", dest="async_check",
                         help="Get async stats")
         args.add_option('--replication', '-r', action="store_true",
                         help="Get replication stats")
@@ -1104,7 +1106,7 @@ class SwiftRecon(object):
                 self.time_check(hosts, options.jitter)
                 self.version_check(hosts)
             else:
-                if options.async:
+                if options.async_check:
                     if self.server_type == 'object':
                         self.async_check(hosts)
                     else:
