@@ -13,6 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import json
 import os
 import random
 import math
@@ -29,12 +30,11 @@ from eventlet.green import subprocess
 
 import swift.common.db
 from swift.common.constraints import check_drive
-from swift.common.direct_client import quote
 from swift.common.utils import get_logger, whataremyips, storage_directory, \
     renamer, mkdirs, lock_parent_directory, config_true_value, \
     unlink_older_than, dump_recon_cache, rsync_module_interpolation, \
-    json, parse_override_options, round_robin_iter, Everything, get_db_files, \
-    parse_db_filename
+    parse_override_options, round_robin_iter, Everything, get_db_files, \
+    parse_db_filename, quote, RateLimitedIterator
 from swift.common import ring
 from swift.common.ring.utils import is_local_device
 from swift.common.http import HTTP_NOT_FOUND, HTTP_INSUFFICIENT_STORAGE, \
@@ -205,6 +205,8 @@ class Replicator(Daemon):
                                 ' to use option %(type)s-replicator/'
                                 'interval.'
                                 % {'type': self.server_type})
+        self.databases_per_second = float(
+            conf.get('databases_per_second', 50))
         self.node_timeout = float(conf.get('node_timeout', 10))
         self.conn_timeout = float(conf.get('conn_timeout', 0.5))
         self.rsync_compress = config_true_value(
@@ -215,6 +217,8 @@ class Replicator(Daemon):
         self.reclaim_age = float(conf.get('reclaim_age', 86400 * 7))
         swift.common.db.DB_PREALLOCATION = \
             config_true_value(conf.get('db_preallocation', 'f'))
+        swift.common.db.QUERY_LOGGING = \
+            config_true_value(conf.get('db_query_logging', 'f'))
         self._zero_stats()
         self.recon_cache_path = conf.get('recon_cache_path',
                                          '/var/cache/swift')
@@ -424,7 +428,7 @@ class Replicator(Daemon):
         Make an http_connection using ReplConnection
 
         :param node: node dictionary from the ring
-        :param partition: partition partition to send in the url
+        :param partition: partition to send in the url
         :param db_file: DB file
 
         :returns: ReplConnection object
@@ -561,7 +565,7 @@ class Replicator(Daemon):
 
         :param partition: partition to be replicated to
         :param object_file: DB file name to be replicated
-        :param node_id: node id of the node to be replicated to
+        :param node_id: node id of the node to be replicated from
         :returns: a tuple (success, responses). ``success`` is a boolean that
             is True if the method completed successfully, False otherwise.
             ``responses`` is a list of booleans each of which indicates the
@@ -577,7 +581,8 @@ class Replicator(Daemon):
         shouldbehere = True
         responses = []
         try:
-            broker = self.brokerclass(object_file, pending_timeout=30)
+            broker = self.brokerclass(object_file, pending_timeout=30,
+                                      logger=self.logger)
             broker.reclaim(now - self.reclaim_age,
                            now - (self.reclaim_age * 2))
             info = broker.get_replication_info()
@@ -734,6 +739,11 @@ class Replicator(Daemon):
     def report_up_to_date(self, full_info):
         return True
 
+    def roundrobin_datadirs(self, dirs):
+        return RateLimitedIterator(
+            roundrobin_datadirs(dirs),
+            elements_per_second=self.databases_per_second)
+
     def run_once(self, *args, **kwargs):
         """Run a replication pass once."""
         override_options = parse_override_options(once=True, **kwargs)
@@ -790,7 +800,7 @@ class Replicator(Daemon):
                               "file, not replicating",
                               ", ".join(ips), self.port)
         self.logger.info(_('Beginning replication run'))
-        for part, object_file, node_id in roundrobin_datadirs(dirs):
+        for part, object_file, node_id in self.roundrobin_datadirs(dirs):
             self.cpool.spawn_n(
                 self._replicate_object, part, object_file, node_id)
         self.cpool.waitall()
@@ -853,7 +863,8 @@ class ReplicatorRpc(object):
             mkdirs(os.path.join(self.root, drive, 'tmp'))
             if not self._db_file_exists(db_file):
                 return HTTPNotFound()
-            return getattr(self, op)(self.broker_class(db_file), args)
+            return getattr(self, op)(
+                self.broker_class(db_file, logger=self.logger), args)
 
     @contextmanager
     def debug_timing(self, name):
@@ -963,7 +974,7 @@ class ReplicatorRpc(object):
             return HTTPNotFound()
         if not os.path.exists(old_filename):
             return HTTPNotFound()
-        broker = self.broker_class(old_filename)
+        broker = self.broker_class(old_filename, logger=self.logger)
         broker.newid(args[0])
         renamer(old_filename, db_file)
         return HTTPNoContent()
@@ -980,8 +991,8 @@ class ReplicatorRpc(object):
         tmp_filename = os.path.join(self.root, drive, 'tmp', args[0])
         if self._abort_rsync_then_merge(db_file, tmp_filename):
             return HTTPNotFound()
-        new_broker = self.broker_class(tmp_filename)
-        existing_broker = self.broker_class(db_file)
+        new_broker = self.broker_class(tmp_filename, logger=self.logger)
+        existing_broker = self.broker_class(db_file, logger=self.logger)
         db_file = existing_broker.db_file
         point = -1
         objects = existing_broker.get_items_since(point, 1000)
