@@ -25,10 +25,12 @@ import zlib
 from time import gmtime, strftime, time
 from zlib import compressobj
 
+from swift.common.constraints import AUTO_CREATE_ACCOUNT_PREFIX
 from swift.common.exceptions import ClientException
 from swift.common.http import (HTTP_NOT_FOUND, HTTP_MULTIPLE_CHOICES,
-                               is_server_error)
-from swift.common.swob import Request
+                               is_client_error, is_server_error)
+from swift.common.request_helpers import USE_REPLICATION_NETWORK_HEADER
+from swift.common.swob import Request, bytes_to_wsgi
 from swift.common.utils import quote, closing_if_possible
 from swift.common.wsgi import loadapp, pipeline_property
 
@@ -95,7 +97,7 @@ class CompressingFileReader(object):
         """
 
         if self.done:
-            return ''
+            return b''
         x = self._f.read(*a, **kw)
         if x:
             self.crc32 = zlib.crc32(x, self.crc32) & 0xffffffff
@@ -112,19 +114,21 @@ class CompressingFileReader(object):
             self.done = True
         if self.first:
             self.first = False
-            header = '\037\213\010\000\000\000\000\000\002\377'
+            header = b'\037\213\010\000\000\000\000\000\002\377'
             compressed = header + compressed
         return compressed
 
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         chunk = self.read(self.chunk_size)
         if chunk:
             return chunk
         raise StopIteration
 
+    next = __next__
+
     def seek(self, offset, whence=0):
         if not (offset == 0 and whence == 0):
             raise NotImplementedError('Seek implemented on offset 0 only')
@@ -144,17 +148,20 @@ class InternalClient(object):
     """
 
     def __init__(self, conf_path, user_agent, request_tries,
-                 allow_modify_pipeline=False):
+                 allow_modify_pipeline=False, use_replication_network=False):
+        if request_tries < 1:
+            raise ValueError('request_tries must be positive')
         self.app = loadapp(conf_path,
                            allow_modify_pipeline=allow_modify_pipeline)
         self.user_agent = user_agent
         self.request_tries = request_tries
+        self.use_replication_network = use_replication_network
 
     get_object_ring = pipeline_property('get_object_ring')
     container_ring = pipeline_property('container_ring')
     account_ring = pipeline_property('account_ring')
     auto_create_account_prefix = pipeline_property(
-        'auto_create_account_prefix', default='.')
+        'auto_create_account_prefix', default=AUTO_CREATE_ACCOUNT_PREFIX)
 
     def make_request(
             self, method, path, headers, acceptable_statuses, body_file=None,
@@ -180,6 +187,10 @@ class InternalClient(object):
 
         headers = dict(headers)
         headers['user-agent'] = self.user_agent
+        headers.setdefault('x-backend-allow-reserved-names', 'true')
+        if self.use_replication_network:
+            headers.setdefault(USE_REPLICATION_NETWORK_HEADER, 'true')
+
         for attempt in range(self.request_tries):
             resp = exc_type = exc_value = exc_traceback = None
             req = Request.blank(
@@ -225,7 +236,7 @@ class InternalClient(object):
 
     def _get_metadata(
             self, path, metadata_prefix='', acceptable_statuses=(2,),
-            headers=None):
+            headers=None, params=None):
         """
         Gets metadata by doing a HEAD on a path and using the metadata_prefix
         to get values from the headers returned.
@@ -248,7 +259,8 @@ class InternalClient(object):
         """
 
         headers = headers or {}
-        resp = self.make_request('HEAD', path, headers, acceptable_statuses)
+        resp = self.make_request('HEAD', path, headers, acceptable_statuses,
+                                 params=params)
         metadata_prefix = metadata_prefix.lower()
         metadata = {}
         for k, v in resp.headers.items():
@@ -276,15 +288,23 @@ class InternalClient(object):
         :raises Exception: Exception is raised when code fails in an
                            unexpected way.
         """
+        if not isinstance(marker, bytes):
+            marker = marker.encode('utf8')
+        if not isinstance(end_marker, bytes):
+            end_marker = end_marker.encode('utf8')
+        if not isinstance(prefix, bytes):
+            prefix = prefix.encode('utf8')
 
         while True:
             resp = self.make_request(
                 'GET', '%s?format=json&marker=%s&end_marker=%s&prefix=%s' %
-                (path, quote(marker), quote(end_marker), quote(prefix)),
+                (path, bytes_to_wsgi(quote(marker)),
+                 bytes_to_wsgi(quote(end_marker)),
+                 bytes_to_wsgi(quote(prefix))),
                 {}, acceptable_statuses)
             if not resp.status_int == 200:
                 if resp.status_int >= HTTP_MULTIPLE_CHOICES:
-                    ''.join(resp.app_iter)
+                    b''.join(resp.app_iter)
                 break
             data = json.loads(resp.body)
             if not data:
@@ -371,6 +391,34 @@ class InternalClient(object):
         return self._iter_items(path, marker, end_marker, prefix,
                                 acceptable_statuses)
 
+    def create_account(self, account):
+        """
+        Creates an account.
+
+        :param account: Account to create.
+        :raises UnexpectedResponse: Exception raised when requests fail
+                                    to get a response with an acceptable status
+        :raises Exception: Exception is raised when code fails in an
+                           unexpected way.
+        """
+        path = self.make_path(account)
+        self.make_request('PUT', path, {}, (201, 202))
+
+    def delete_account(self, account, acceptable_statuses=(2, HTTP_NOT_FOUND)):
+        """
+        Deletes an account.
+
+        :param account: Account to delete.
+        :param acceptable_statuses: List of status for valid responses,
+                                    defaults to (2, HTTP_NOT_FOUND).
+        :raises UnexpectedResponse: Exception raised when requests fail
+                                    to get a response with an acceptable status
+        :raises Exception: Exception is raised when code fails in an
+                           unexpected way.
+        """
+        path = self.make_path(account)
+        self.make_request('DELETE', path, {}, acceptable_statuses)
+
     def get_account_info(
             self, account, acceptable_statuses=(2, HTTP_NOT_FOUND)):
         """
@@ -394,7 +442,8 @@ class InternalClient(object):
                 int(resp.headers.get('x-account-object-count', 0)))
 
     def get_account_metadata(
-            self, account, metadata_prefix='', acceptable_statuses=(2,)):
+            self, account, metadata_prefix='', acceptable_statuses=(2,),
+            params=None):
         """Gets account metadata.
 
         :param account: Account on which to get the metadata.
@@ -413,7 +462,8 @@ class InternalClient(object):
         """
 
         path = self.make_path(account)
-        return self._get_metadata(path, metadata_prefix, acceptable_statuses)
+        return self._get_metadata(path, metadata_prefix, acceptable_statuses,
+                                  headers=None, params=params)
 
     def set_account_metadata(
             self, account, metadata, metadata_prefix='',
@@ -484,7 +534,8 @@ class InternalClient(object):
         self.make_request('PUT', path, headers, acceptable_statuses)
 
     def delete_container(
-            self, account, container, acceptable_statuses=(2, HTTP_NOT_FOUND)):
+            self, account, container, headers=None,
+            acceptable_statuses=(2, HTTP_NOT_FOUND)):
         """
         Deletes a container.
 
@@ -499,12 +550,13 @@ class InternalClient(object):
                            unexpected way.
         """
 
+        headers = headers or {}
         path = self.make_path(account, container)
-        self.make_request('DELETE', path, {}, acceptable_statuses)
+        self.make_request('DELETE', path, headers, acceptable_statuses)
 
     def get_container_metadata(
             self, account, container, metadata_prefix='',
-            acceptable_statuses=(2,)):
+            acceptable_statuses=(2,), params=None):
         """Gets container metadata.
 
         :param account: The container's account.
@@ -524,7 +576,8 @@ class InternalClient(object):
         """
 
         path = self.make_path(account, container)
-        return self._get_metadata(path, metadata_prefix, acceptable_statuses)
+        return self._get_metadata(path, metadata_prefix, acceptable_statuses,
+                                  params=params)
 
     def iter_objects(
             self, account, container, marker='', end_marker='', prefix='',
@@ -602,11 +655,18 @@ class InternalClient(object):
         """
 
         path = self.make_path(account, container, obj)
-        self.make_request('DELETE', path, (headers or {}), acceptable_statuses)
+        resp = self.make_request('DELETE', path, (headers or {}),
+                                 acceptable_statuses)
+
+        # Drain the response body to prevent unexpected disconnect
+        # in proxy-server
+        with closing_if_possible(resp.app_iter):
+            for iter_body in resp.app_iter:
+                pass
 
     def get_object_metadata(
             self, account, container, obj, metadata_prefix='',
-            acceptable_statuses=(2,), headers=None):
+            acceptable_statuses=(2,), headers=None, params=None):
         """Gets object metadata.
 
         :param account: The object's account.
@@ -629,9 +689,9 @@ class InternalClient(object):
 
         path = self.make_path(account, container, obj)
         return self._get_metadata(path, metadata_prefix, acceptable_statuses,
-                                  headers=headers)
+                                  headers=headers, params=params)
 
-    def get_object(self, account, container, obj, headers,
+    def get_object(self, account, container, obj, headers=None,
                    acceptable_statuses=(2,), params=None):
         """
         Gets an object.
@@ -685,7 +745,7 @@ class InternalClient(object):
         if not resp.status_int // 100 == 2:
             return
 
-        last_part = ''
+        last_part = b''
         compressed = obj.endswith('.gz')
         # magic in the following zlib.decompressobj argument is courtesy of
         # Python decompressing gzip chunk-by-chunk
@@ -694,7 +754,7 @@ class InternalClient(object):
         for chunk in resp.app_iter:
             if compressed:
                 chunk = d.decompress(chunk)
-            parts = chunk.split('\n')
+            parts = chunk.split(b'\n')
             if len(parts) == 1:
                 last_part = last_part + parts[0]
             else:
@@ -733,13 +793,17 @@ class InternalClient(object):
             path, metadata, metadata_prefix, acceptable_statuses)
 
     def upload_object(
-            self, fobj, account, container, obj, headers=None):
+            self, fobj, account, container, obj, headers=None,
+            acceptable_statuses=(2,), params=None):
         """
         :param fobj: File object to read object's content from.
         :param account: The object's account.
         :param container: The object's container.
         :param obj: The object.
         :param headers: Headers to send with request, defaults to empty dict.
+        :param acceptable_statuses: List of acceptable statuses for request.
+        :param params: A dict of params to be set in request query string,
+                       defaults to None.
 
         :raises UnexpectedResponse: Exception raised when requests fail
                                     to get a response with an acceptable status
@@ -751,7 +815,8 @@ class InternalClient(object):
         if 'Content-Length' not in headers:
             headers['Transfer-Encoding'] = 'chunked'
         path = self.make_path(account, container, obj)
-        self.make_request('PUT', path, headers, (2,), fobj)
+        self.make_request('PUT', path, headers, acceptable_statuses, fobj,
+                          params=params)
 
 
 def get_auth(url, user, key, auth_version='1.0', **kwargs):
@@ -870,14 +935,16 @@ class SimpleClient(object):
             self.attempts += 1
             try:
                 return self.base_request(method, **kwargs)
-            except (socket.error, httplib.HTTPException, urllib2.URLError) \
-                    as err:
+            except urllib2.HTTPError as err:
+                if is_client_error(err.getcode() or 500):
+                    raise ClientException('Client error',
+                                          http_status=err.getcode())
+                elif self.attempts > retries:
+                    raise ClientException('Raise too many retries',
+                                          http_status=err.getcode())
+            except (socket.error, httplib.HTTPException, urllib2.URLError):
                 if self.attempts > retries:
-                    if isinstance(err, urllib2.HTTPError):
-                        raise ClientException('Raise too many retries',
-                                              http_status=err.getcode())
-                    else:
-                        raise
+                    raise
             sleep(backoff)
             backoff = min(backoff * 2, self.max_backoff)
 
