@@ -98,6 +98,10 @@ The bulk middleware  will handle xattrs stored by both GNU and BSD tar (2).
 Only xattrs ``user.mime_type`` and ``user.meta.*`` are processed. Other
 attributes are ignored.
 
+In addition to the extended attributes, the object metadata and the
+x-delete-at/x-delete-after headers set in the request are also assigned to the
+extracted objects.
+
 Notes:
 
 (1) The POSIX 1003.1-2001 (pax) format. The default format on GNU tar
@@ -191,7 +195,7 @@ payload sent to the proxy (the list of objects/containers to be deleted).
 """
 
 import json
-from six.moves.urllib.parse import quote, unquote
+import six
 import tarfile
 from xml.sax import saxutils
 from time import time
@@ -200,11 +204,14 @@ import zlib
 from swift.common.swob import Request, HTTPBadGateway, \
     HTTPCreated, HTTPBadRequest, HTTPNotFound, HTTPUnauthorized, HTTPOk, \
     HTTPPreconditionFailed, HTTPRequestEntityTooLarge, HTTPNotAcceptable, \
-    HTTPLengthRequired, HTTPException, HTTPServerError, wsgify
+    HTTPLengthRequired, HTTPException, HTTPServerError, wsgify, \
+    bytes_to_wsgi, str_to_wsgi, wsgi_unquote, wsgi_quote, wsgi_to_str
 from swift.common.utils import get_logger, register_swift_info, \
     StreamingPile
 from swift.common import constraints
 from swift.common.http import HTTP_UNAUTHORIZED, HTTP_NOT_FOUND, HTTP_CONFLICT
+from swift.common.request_helpers import is_user_meta
+from swift.common.wsgi import make_subrequest
 
 
 class CreateContainerError(Exception):
@@ -233,7 +240,7 @@ def get_response_body(data_format, data_dict, error_list, root_tag):
     """
     if data_format == 'application/json':
         data_dict['Errors'] = error_list
-        return json.dumps(data_dict)
+        return json.dumps(data_dict).encode('ascii')
     if data_format and data_format.endswith('/xml'):
         output = ['<', root_tag, '>\n']
         for key in sorted(data_dict):
@@ -250,7 +257,9 @@ def get_response_body(data_format, data_dict, error_list, root_tag):
                 saxutils.escape(status), '</status></object>\n',
             ])
         output.extend(['</errors>\n</', root_tag, '>\n'])
-        return ''.join(output)
+        if six.PY2:
+            return ''.join(output)
+        return ''.join(output).encode('utf-8')
 
     output = []
     for key in sorted(data_dict):
@@ -259,7 +268,9 @@ def get_response_body(data_format, data_dict, error_list, root_tag):
     output.extend(
         '%s, %s\n' % (name, status)
         for name, status in error_list)
-    return ''.join(output)
+    if six.PY2:
+        return ''.join(output)
+    return ''.join(output).encode('utf-8')
 
 
 def pax_key_to_swift_header(pax_key):
@@ -268,10 +279,14 @@ def pax_key_to_swift_header(pax_key):
         return "Content-Type"
     elif pax_key.startswith(u"SCHILY.xattr.user.meta."):
         useful_part = pax_key[len(u"SCHILY.xattr.user.meta."):]
-        return "X-Object-Meta-" + useful_part.encode("utf-8")
+        if six.PY2:
+            return "X-Object-Meta-" + useful_part.encode("utf-8")
+        return str_to_wsgi("X-Object-Meta-" + useful_part)
     elif pax_key.startswith(u"LIBARCHIVE.xattr.user.meta."):
         useful_part = pax_key[len(u"LIBARCHIVE.xattr.user.meta."):]
-        return "X-Object-Meta-" + useful_part.encode("utf-8")
+        if six.PY2:
+            return "X-Object-Meta-" + useful_part.encode("utf-8")
+        return str_to_wsgi("X-Object-Meta-" + useful_part)
     else:
         # You can get things like atime/mtime/ctime or filesystem ACLs in
         # pax headers; those aren't really user metadata. The same goes for
@@ -306,20 +321,18 @@ class Bulk(object):
         :returns: True if created container, False if container exists
         :raises CreateContainerError: when unable to create container
         """
-        new_env = req.environ.copy()
-        new_env['PATH_INFO'] = container_path
-        new_env['swift.source'] = 'EA'
-        new_env['REQUEST_METHOD'] = 'HEAD'
-        head_cont_req = Request.blank(container_path, environ=new_env)
+        head_cont_req = make_subrequest(
+            req.environ, method='HEAD', path=wsgi_quote(container_path),
+            headers={'X-Auth-Token': req.headers.get('X-Auth-Token')},
+            swift_source='EA')
         resp = head_cont_req.get_response(self.app)
         if resp.is_success:
             return False
         if resp.status_int == HTTP_NOT_FOUND:
-            new_env = req.environ.copy()
-            new_env['PATH_INFO'] = container_path
-            new_env['swift.source'] = 'EA'
-            new_env['REQUEST_METHOD'] = 'PUT'
-            create_cont_req = Request.blank(container_path, environ=new_env)
+            create_cont_req = make_subrequest(
+                req.environ, method='PUT', path=wsgi_quote(container_path),
+                headers={'X-Auth-Token': req.headers.get('X-Auth-Token')},
+                swift_source='EA')
             resp = create_cont_req.get_response(self.app)
             if resp.is_success:
                 return True
@@ -334,7 +347,7 @@ class Bulk(object):
         :returns: a list of the contents of req.body when separated by newline.
         :raises HTTPException: on failures
         """
-        line = ''
+        line = b''
         data_remaining = True
         objs_to_delete = []
         if req.content_length is None and \
@@ -342,21 +355,31 @@ class Bulk(object):
             raise HTTPLengthRequired(request=req)
 
         while data_remaining:
-            if '\n' in line:
-                obj_to_delete, line = line.split('\n', 1)
-                obj_to_delete = obj_to_delete.strip()
-                objs_to_delete.append(
-                    {'name': unquote(obj_to_delete)})
+            if b'\n' in line:
+                obj_to_delete, line = line.split(b'\n', 1)
+                if six.PY2:
+                    obj_to_delete = wsgi_unquote(obj_to_delete.strip())
+                else:
+                    # yeah, all this chaining is pretty terrible...
+                    # but it gets even worse trying to use UTF-8 and
+                    # errors='surrogateescape' when dealing with terrible
+                    # input like b'\xe2%98\x83'
+                    obj_to_delete = wsgi_to_str(wsgi_unquote(
+                        bytes_to_wsgi(obj_to_delete.strip())))
+                objs_to_delete.append({'name': obj_to_delete})
             else:
                 data = req.body_file.read(self.max_path_length)
                 if data:
                     line += data
                 else:
                     data_remaining = False
-                    obj_to_delete = line.strip()
+                    if six.PY2:
+                        obj_to_delete = wsgi_unquote(line.strip())
+                    else:
+                        obj_to_delete = wsgi_to_str(wsgi_unquote(
+                            bytes_to_wsgi(line.strip())))
                     if obj_to_delete:
-                        objs_to_delete.append(
-                            {'name': unquote(obj_to_delete)})
+                        objs_to_delete.append({'name': obj_to_delete})
             if len(objs_to_delete) > self.max_deletes_per_request:
                 raise HTTPRequestEntityTooLarge(
                     'Maximum Bulk Deletes: %d per request' %
@@ -377,11 +400,15 @@ class Bulk(object):
 
         :params req: a swob Request
         :params objs_to_delete: a list of dictionaries that specifies the
-            objects to be deleted. If None, uses self.get_objs_to_delete to
-            query request.
+            (native string) objects to be deleted. If None, uses
+            self.get_objs_to_delete to query request.
         """
         last_yield = time()
-        separator = ''
+        if out_content_type and out_content_type.endswith('/xml'):
+            to_yield = b'<?xml version="1.0" encoding="UTF-8"?>\n'
+        else:
+            to_yield = b' '
+        separator = b''
         failed_files = []
         resp_dict = {'Response Status': HTTPOk().status,
                      'Response Body': '',
@@ -391,13 +418,13 @@ class Bulk(object):
         try:
             if not out_content_type:
                 raise HTTPNotAcceptable(request=req)
-            if out_content_type.endswith('/xml'):
-                yield '<?xml version="1.0" encoding="UTF-8"?>\n'
 
             try:
                 vrs, account, _junk = req.split_path(2, 3, True)
             except ValueError:
                 raise HTTPNotFound(request=req)
+            vrs = wsgi_to_str(vrs)
+            account = wsgi_to_str(account)
 
             incoming_format = req.headers.get('Content-Type')
             if incoming_format and \
@@ -421,16 +448,17 @@ class Bulk(object):
                             resp_dict['Number Not Found'] += 1
                         else:
                             failed_files.append([
-                                quote(obj_name),
+                                wsgi_quote(str_to_wsgi(obj_name)),
                                 obj_to_delete['error']['message']])
                         continue
                     delete_path = '/'.join(['', vrs, account,
                                             obj_name.lstrip('/')])
                     if not constraints.check_utf8(delete_path):
-                        failed_files.append([quote(obj_name),
+                        failed_files.append([wsgi_quote(str_to_wsgi(obj_name)),
                                              HTTPPreconditionFailed().status])
                         continue
-                    yield (obj_name, delete_path)
+                    yield (obj_name, delete_path,
+                           obj_to_delete.get('version_id'))
 
             def objs_then_containers(objs_to_delete):
                 # process all objects first
@@ -440,16 +468,17 @@ class Bulk(object):
                 yield delete_filter(lambda name: '/' not in name.strip('/'),
                                     objs_to_delete)
 
-            def do_delete(obj_name, delete_path):
-                new_env = req.environ.copy()
-                new_env['PATH_INFO'] = delete_path
-                del(new_env['wsgi.input'])
-                new_env['CONTENT_LENGTH'] = 0
-                new_env['REQUEST_METHOD'] = 'DELETE'
-                new_env['HTTP_USER_AGENT'] = '%s %s' % (
-                    req.environ.get('HTTP_USER_AGENT'), user_agent)
-                new_env['swift.source'] = swift_source
-                delete_obj_req = Request.blank(delete_path, new_env)
+            def do_delete(obj_name, delete_path, version_id):
+                delete_obj_req = make_subrequest(
+                    req.environ, method='DELETE',
+                    path=wsgi_quote(str_to_wsgi(delete_path)),
+                    headers={'X-Auth-Token': req.headers.get('X-Auth-Token')},
+                    body='', agent='%(orig)s ' + user_agent,
+                    swift_source=swift_source)
+                if version_id is None:
+                    delete_obj_req.params = {}
+                else:
+                    delete_obj_req.params = {'version-id': version_id}
                 return (delete_obj_req.get_response(self.app), obj_name, 0)
 
             with StreamingPile(self.delete_concurrency) as pile:
@@ -457,9 +486,9 @@ class Bulk(object):
                     for resp, obj_name, retry in pile.asyncstarmap(
                             do_delete, names_to_delete):
                         if last_yield + self.yield_frequency < time():
-                            separator = '\r\n\r\n'
                             last_yield = time()
-                            yield ' '
+                            yield to_yield
+                            to_yield, separator = b' ', b'\r\n\r\n'
                         self._process_delete(resp, pile, obj_name,
                                              resp_dict, failed_files,
                                              failed_file_response, retry)
@@ -467,9 +496,9 @@ class Bulk(object):
                             # Abort, but drain off the in-progress deletes
                             for resp, obj_name, retry in pile:
                                 if last_yield + self.yield_frequency < time():
-                                    separator = '\r\n\r\n'
                                     last_yield = time()
-                                    yield ' '
+                                    yield to_yield
+                                    to_yield, separator = b' ', b'\r\n\r\n'
                                 # Don't pass in the pile, as we shouldn't retry
                                 self._process_delete(
                                     resp, None, obj_name, resp_dict,
@@ -487,7 +516,7 @@ class Bulk(object):
 
         except HTTPException as err:
             resp_dict['Response Status'] = err.status
-            resp_dict['Response Body'] = err.body
+            resp_dict['Response Body'] = err.body.decode('utf-8')
         except Exception:
             self.logger.exception('Error in bulk delete.')
             resp_dict['Response Status'] = HTTPServerError().status
@@ -513,14 +542,16 @@ class Bulk(object):
                      'Response Body': '', 'Number Files Created': 0}
         failed_files = []
         last_yield = time()
-        separator = ''
+        if out_content_type and out_content_type.endswith('/xml'):
+            to_yield = b'<?xml version="1.0" encoding="UTF-8"?>\n'
+        else:
+            to_yield = b' '
+        separator = b''
         containers_accessed = set()
         req.environ['eventlet.minimum_write_chunk_size'] = 0
         try:
             if not out_content_type:
                 raise HTTPNotAcceptable(request=req)
-            if out_content_type.endswith('/xml'):
-                yield '<?xml version="1.0" encoding="UTF-8"?>\n'
 
             if req.content_length is None and \
                     req.headers.get('transfer-encoding',
@@ -538,15 +569,18 @@ class Bulk(object):
             containers_created = 0
             while True:
                 if last_yield + self.yield_frequency < time():
-                    separator = '\r\n\r\n'
                     last_yield = time()
-                    yield ' '
-                tar_info = next(tar)
+                    yield to_yield
+                    to_yield, separator = b' ', b'\r\n\r\n'
+                tar_info = tar.next()
                 if tar_info is None or \
                         len(failed_files) >= self.max_failed_extractions:
                     break
                 if tar_info.isfile():
                     obj_path = tar_info.name
+                    if not six.PY2:
+                        obj_path = obj_path.encode('utf-8', 'surrogateescape')
+                    obj_path = bytes_to_wsgi(obj_path)
                     if obj_path.startswith('./'):
                         obj_path = obj_path[2:]
                     obj_path = obj_path.lstrip('/')
@@ -558,14 +592,14 @@ class Bulk(object):
                     destination = '/'.join(
                         ['', vrs, account, obj_path])
                     container = obj_path.split('/', 1)[0]
-                    if not constraints.check_utf8(destination):
+                    if not constraints.check_utf8(wsgi_to_str(destination)):
                         failed_files.append(
-                            [quote(obj_path[:self.max_path_length]),
+                            [wsgi_quote(obj_path[:self.max_path_length]),
                              HTTPPreconditionFailed().status])
                         continue
                     if tar_info.size > constraints.MAX_FILE_SIZE:
                         failed_files.append([
-                            quote(obj_path[:self.max_path_length]),
+                            wsgi_quote(obj_path[:self.max_path_length]),
                             HTTPRequestEntityTooLarge().status])
                         continue
                     container_failure = None
@@ -582,26 +616,34 @@ class Bulk(object):
                             # the object PUT to this container still may
                             # succeed if acls are set
                             container_failure = [
-                                quote(cont_path[:self.max_path_length]),
+                                wsgi_quote(cont_path[:self.max_path_length]),
                                 err.status]
                             if err.status_int == HTTP_UNAUTHORIZED:
                                 raise HTTPUnauthorized(request=req)
                         except ValueError:
                             failed_files.append([
-                                quote(obj_path[:self.max_path_length]),
+                                wsgi_quote(obj_path[:self.max_path_length]),
                                 HTTPBadRequest().status])
                             continue
 
                     tar_file = tar.extractfile(tar_info)
-                    new_env = req.environ.copy()
-                    new_env['REQUEST_METHOD'] = 'PUT'
-                    new_env['wsgi.input'] = tar_file
-                    new_env['PATH_INFO'] = destination
-                    new_env['CONTENT_LENGTH'] = tar_info.size
-                    new_env['swift.source'] = 'EA'
-                    new_env['HTTP_USER_AGENT'] = \
-                        '%s BulkExpand' % req.environ.get('HTTP_USER_AGENT')
-                    create_obj_req = Request.blank(destination, new_env)
+                    create_headers = {
+                        'Content-Length': tar_info.size,
+                        'X-Auth-Token': req.headers.get('X-Auth-Token'),
+                    }
+
+                    # Copy some whitelisted headers to the subrequest
+                    for k, v in req.headers.items():
+                        if ((k.lower() in ('x-delete-at', 'x-delete-after'))
+                                or is_user_meta('object', k)):
+                            create_headers[k] = v
+
+                    create_obj_req = make_subrequest(
+                        req.environ, method='PUT',
+                        path=wsgi_quote(destination),
+                        headers=create_headers,
+                        agent='%(orig)s BulkExpand', swift_source='EA')
+                    create_obj_req.environ['wsgi.input'] = tar_file
 
                     for pax_key, pax_value in tar_info.pax_headers.items():
                         header_name = pax_key_to_swift_header(pax_key)
@@ -621,13 +663,13 @@ class Bulk(object):
                             failed_files.append(container_failure)
                         if resp.status_int == HTTP_UNAUTHORIZED:
                             failed_files.append([
-                                quote(obj_path[:self.max_path_length]),
+                                wsgi_quote(obj_path[:self.max_path_length]),
                                 HTTPUnauthorized().status])
                             raise HTTPUnauthorized(request=req)
                         if resp.status_int // 100 == 5:
                             failed_response_type = HTTPBadGateway
                         failed_files.append([
-                            quote(obj_path[:self.max_path_length]),
+                            wsgi_quote(obj_path[:self.max_path_length]),
                             resp.status])
 
             if failed_files:
@@ -638,7 +680,7 @@ class Bulk(object):
 
         except HTTPException as err:
             resp_dict['Response Status'] = err.status
-            resp_dict['Response Body'] = err.body
+            resp_dict['Response Body'] = err.body.decode('utf-8')
         except (tarfile.TarError, zlib.error) as tar_error:
             resp_dict['Response Status'] = HTTPBadRequest().status
             resp_dict['Response Body'] = 'Invalid Tar File: %s' % tar_error
@@ -656,7 +698,7 @@ class Bulk(object):
         elif resp.status_int == HTTP_NOT_FOUND:
             resp_dict['Number Not Found'] += 1
         elif resp.status_int == HTTP_UNAUTHORIZED:
-            failed_files.append([quote(obj_name),
+            failed_files.append([wsgi_quote(str_to_wsgi(obj_name)),
                                  HTTPUnauthorized().status])
         elif resp.status_int == HTTP_CONFLICT and pile and \
                 self.retry_count > 0 and self.retry_count > retry:
@@ -671,7 +713,8 @@ class Bulk(object):
         else:
             if resp.status_int // 100 == 5:
                 failed_file_response['type'] = HTTPBadGateway
-            failed_files.append([quote(obj_name), resp.status])
+            failed_files.append([wsgi_quote(str_to_wsgi(obj_name)),
+                                 resp.status])
 
     @wsgify
     def __call__(self, req):
