@@ -40,7 +40,7 @@ class Crypto(object):
     cipher = 'AES_CTR_256'
     # AES will accept several key sizes - we are using 256 bits i.e. 32 bytes
     key_length = 32
-    iv_length = algorithms.AES.block_size / 8
+    iv_length = algorithms.AES.block_size // 8
 
     def __init__(self, conf=None):
         self.logger = get_logger(conf, log_route="crypto")
@@ -80,7 +80,7 @@ class Crypto(object):
             offset_blocks, offset_in_block = divmod(offset, self.iv_length)
             ivl = int(binascii.hexlify(iv), 16) + offset_blocks
             ivl %= 1 << algorithms.AES.block_size
-            iv = str(bytearray.fromhex(format(
+            iv = bytes(bytearray.fromhex(format(
                 ivl, '0%dx' % (2 * self.iv_length))))
         else:
             offset_in_block = 0
@@ -89,7 +89,7 @@ class Crypto(object):
                         backend=self.backend)
         dec = engine.decryptor()
         # Adjust decryption boundary within current AES block
-        dec.update('*' * offset_in_block)
+        dec.update(b'*' * offset_in_block)
         return dec
 
     def create_iv(self):
@@ -231,7 +231,7 @@ def dump_crypto_meta(crypto_meta):
         json.dumps(b64_encode_meta(crypto_meta), sort_keys=True))
 
 
-def load_crypto_meta(value):
+def load_crypto_meta(value, b64decode=True):
     """
     Build the crypto_meta from the json object.
 
@@ -243,15 +243,17 @@ def load_crypto_meta(value):
           as native unicode strings on py3).
 
     :param value: a string serialization of a crypto meta dict
+    :param b64decode: decode the 'key' and 'iv' values to bytes, default True
     :returns: a dict containing crypto meta items
     :raises EncryptionException: if an error occurs while parsing the
                                  crypto meta
     """
     def b64_decode_meta(crypto_meta):
         return {
-            str(name): (base64.b64decode(val) if name in ('iv', 'key')
-                        else b64_decode_meta(val) if isinstance(val, dict)
-                        else val.encode('utf8') if six.PY2 else val)
+            str(name): (
+                base64.b64decode(val) if name in ('iv', 'key') and b64decode
+                else b64_decode_meta(val) if isinstance(val, dict)
+                else val.encode('utf8') if six.PY2 else val)
             for name, val in crypto_meta.items()}
 
     try:
@@ -274,6 +276,8 @@ def append_crypto_meta(value, crypto_meta):
     :param crypto_meta: a dict of crypto meta
     :return: a string of the form <value>; swift_meta=<serialized crypto meta>
     """
+    if not isinstance(value, str):
+        raise ValueError
     return '%s; swift_meta=%s' % (value, dump_crypto_meta(crypto_meta))
 
 
