@@ -22,9 +22,10 @@ from swift.common.http import is_success
 from swift.common.middleware.crypto.crypto_utils import CryptoWSGIContext, \
     dump_crypto_meta, append_crypto_meta, Crypto
 from swift.common.request_helpers import get_object_transient_sysmeta, \
-    strip_user_meta_prefix, is_user_meta, update_etag_is_at_header
+    strip_user_meta_prefix, is_user_meta, update_etag_is_at_header, \
+    get_container_update_override_key
 from swift.common.swob import Request, Match, HTTPException, \
-    HTTPUnprocessableEntity
+    HTTPUnprocessableEntity, wsgi_to_bytes, bytes_to_wsgi, normalize_etag
 from swift.common.utils import get_logger, config_true_value, \
     MD5_OF_EMPTY_STRING
 
@@ -46,7 +47,8 @@ def encrypt_header_val(crypto, value, key):
 
     crypto_meta = crypto.create_crypto_meta()
     crypto_ctxt = crypto.create_encryption_ctxt(key, crypto_meta['iv'])
-    enc_val = base64.b64encode(crypto_ctxt.update(value))
+    enc_val = bytes_to_wsgi(base64.b64encode(
+        crypto_ctxt.update(wsgi_to_bytes(value))))
     return enc_val, crypto_meta
 
 
@@ -58,6 +60,8 @@ def _hmac_etag(key, etag):
     :param etag: The etag to hash.
     :returns: a Base64-encoded representation of the HMAC
     """
+    if not isinstance(etag, bytes):
+        etag = wsgi_to_bytes(etag)
     result = hmac.new(key, etag, digestmod=hashlib.sha256).digest()
     return base64.b64encode(result).decode()
 
@@ -97,8 +101,8 @@ class EncInputWrapper(object):
         # remove any Etag from headers, it won't be valid for ciphertext and
         # we'll send the ciphertext Etag later in footer metadata
         client_etag = req.headers.pop('etag', None)
-        container_listing_etag_header = req.headers.get(
-            'X-Object-Sysmeta-Container-Update-Override-Etag')
+        override_header = get_container_update_override_key('etag')
+        container_listing_etag_header = req.headers.get(override_header)
 
         def footers_callback(footers):
             if inner_callback:
@@ -146,12 +150,15 @@ class EncInputWrapper(object):
             #   * override in the footer, otherwise
             #   * override in the header, and finally
             #   * MD5 of the plaintext received
-            # This may be None if no override was set and no data was read
+            # This may be None if no override was set and no data was read. An
+            # override value of '' will be passed on.
             container_listing_etag = footers.get(
-                'X-Object-Sysmeta-Container-Update-Override-Etag',
-                container_listing_etag_header) or plaintext_etag
+                override_header, container_listing_etag_header)
 
-            if (container_listing_etag is not None and
+            if container_listing_etag is None:
+                container_listing_etag = plaintext_etag
+
+            if (container_listing_etag and
                     (container_listing_etag != MD5_OF_EMPTY_STRING or
                      plaintext_etag)):
                 # Encrypt the container-listing etag using the container key
@@ -167,7 +174,7 @@ class EncInputWrapper(object):
                     self.crypto, container_listing_etag,
                     self.keys['container'])
                 crypto_meta['key_id'] = self.keys['id']
-                footers['X-Object-Sysmeta-Container-Update-Override-Etag'] = \
+                footers[override_header] = \
                     append_crypto_meta(val, crypto_meta)
             # else: no override was set and no data was read
 
@@ -256,7 +263,7 @@ class EncrypterObjContext(CryptoWSGIContext):
             ciphertext_etag = enc_input_proxy.ciphertext_md5.hexdigest()
             mod_resp_headers = [
                 (h, v if (h.lower() != 'etag' or
-                          v.strip('"') != ciphertext_etag)
+                          normalize_etag(v) != ciphertext_etag)
                     else plaintext_etag)
                 for h, v in mod_resp_headers]
 
@@ -362,7 +369,10 @@ class Encrypter(object):
             return self.app(env, start_response)
         try:
             req.split_path(4, 4, True)
+            is_object_request = True
         except ValueError:
+            is_object_request = False
+        if not is_object_request:
             return self.app(env, start_response)
 
         if req.method in ('GET', 'HEAD'):
