@@ -59,27 +59,35 @@ Static Large Object when the multipart upload is completed.
 
 """
 
+import binascii
+import copy
+from hashlib import md5
 import os
 import re
+import time
 
-from swift.common.swob import Range
-from swift.common.utils import json, public
+import six
+
+from swift.common.swob import Range, bytes_to_wsgi, normalize_etag
+from swift.common.utils import json, public, reiterate
 from swift.common.db import utf8encode
+from swift.common.request_helpers import get_container_update_override_key
 
-from six.moves.urllib.parse import urlparse  # pylint: disable=F0401
+from six.moves.urllib.parse import quote, urlparse
 
 from swift.common.middleware.s3api.controllers.base import Controller, \
     bucket_operation, object_operation, check_container_existence
 from swift.common.middleware.s3api.s3response import InvalidArgument, \
-    ErrorResponse, MalformedXML, \
+    ErrorResponse, MalformedXML, BadDigest, \
     InvalidPart, BucketAlreadyExists, EntityTooSmall, InvalidPartOrder, \
     InvalidRequest, HTTPOk, HTTPNoContent, NoSuchKey, NoSuchUpload, \
-    NoSuchBucket
+    NoSuchBucket, BucketAlreadyOwnedByYou
 from swift.common.middleware.s3api.exception import BadSwiftRequest
 from swift.common.middleware.s3api.utils import unique_id, \
     MULTIUPLOAD_SUFFIX, S3Timestamp, sysmeta_header
 from swift.common.middleware.s3api.etree import Element, SubElement, \
     fromstring, tostring, XMLSyntaxError, DocumentInvalid
+from swift.common.storage_policy import POLICIES
 
 DEFAULT_MAX_PARTS_LISTING = 1000
 DEFAULT_MAX_UPLOADS = 1000
@@ -92,15 +100,56 @@ def _get_upload_info(req, app, upload_id):
     container = req.container_name + MULTIUPLOAD_SUFFIX
     obj = '%s/%s' % (req.object_name, upload_id)
 
+    # XXX: if we leave the copy-source header, somewhere later we might
+    # drop in a ?version-id=... query string that's utterly inappropriate
+    # for the upload marker. Until we get around to fixing that, just pop
+    # it off for now...
+    copy_source = req.headers.pop('X-Amz-Copy-Source', None)
     try:
         return req.get_response(app, 'HEAD', container=container, obj=obj)
     except NoSuchKey:
+        try:
+            resp = req.get_response(app, 'HEAD')
+            if resp.sysmeta_headers.get(sysmeta_header(
+                    'object', 'upload-id')) == upload_id:
+                return resp
+        except NoSuchKey:
+            pass
         raise NoSuchUpload(upload_id=upload_id)
+    finally:
+        # ...making sure to restore any copy-source before returning
+        if copy_source is not None:
+            req.headers['X-Amz-Copy-Source'] = copy_source
 
 
-def _check_upload_info(req, app, upload_id):
+def _make_complete_body(req, s3_etag, yielded_anything):
+    result_elem = Element('CompleteMultipartUploadResult')
 
-    _get_upload_info(req, app, upload_id)
+    # NOTE: boto with sig v4 appends port to HTTP_HOST value at
+    # the request header when the port is non default value and it
+    # makes req.host_url like as http://localhost:8080:8080/path
+    # that obviously invalid. Probably it should be resolved at
+    # swift.common.swob though, tentatively we are parsing and
+    # reconstructing the correct host_url info here.
+    # in detail, https://github.com/boto/boto/pull/3513
+    parsed_url = urlparse(req.host_url)
+    host_url = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)
+    # Why are we doing our own port parsing? Because py3 decided
+    # to start raising ValueErrors on access after parsing such
+    # an invalid port
+    netloc = parsed_url.netloc.split('@')[-1].split(']')[-1]
+    if ':' in netloc:
+        port = netloc.split(':', 2)[1]
+        host_url += ':%s' % port
+
+    SubElement(result_elem, 'Location').text = host_url + req.path
+    SubElement(result_elem, 'Bucket').text = req.container_name
+    SubElement(result_elem, 'Key').text = req.object_name
+    SubElement(result_elem, 'ETag').text = '"%s"' % s3_etag
+    body = tostring(result_elem, xml_declaration=not yielded_anything)
+    if yielded_anything:
+        return b'\n' + body
+    return body
 
 
 class PartController(Controller):
@@ -135,7 +184,7 @@ class PartController(Controller):
                                   err_msg)
 
         upload_id = req.params['uploadId']
-        _check_upload_info(req, self.app, upload_id)
+        _get_upload_info(req, self.app, upload_id)
 
         req.container_name += MULTIUPLOAD_SUFFIX
         req.object_name = '%s/%s/%d' % (req.object_name, upload_id,
@@ -170,6 +219,15 @@ class PartController(Controller):
 
             req.headers['Range'] = rng
             del req.headers['X-Amz-Copy-Source-Range']
+        if 'X-Amz-Copy-Source' in req.headers:
+            # Clear some problematic headers that might be on the source
+            req.headers.update({
+                sysmeta_header('object', 'etag'): '',
+                'X-Object-Sysmeta-Swift3-Etag': '',  # for legacy data
+                'X-Object-Sysmeta-Slo-Etag': '',
+                'X-Object-Sysmeta-Slo-Size': '',
+                get_container_update_override_key('etag'): '',
+            })
         resp = req.get_response(self.app)
 
         if 'X-Amz-Copy-Source' in req.headers:
@@ -220,8 +278,8 @@ class UploadsController(Controller):
 
             :return (non_delimited_uploads, common_prefixes)
             """
-            (prefix, delimiter) = \
-                utf8encode(prefix, delimiter)
+            if six.PY2:
+                (prefix, delimiter) = utf8encode(prefix, delimiter)
             non_delimited_uploads = []
             common_prefixes = set()
             for upload in uploads:
@@ -282,8 +340,7 @@ class UploadsController(Controller):
         if 'delimiter' in req.params:
             prefix = req.params.get('prefix', '')
             delimiter = req.params['delimiter']
-            uploads, prefixes = \
-                separate_uploads(uploads, prefix, delimiter)
+            uploads, prefixes = separate_uploads(uploads, prefix, delimiter)
 
         if len(uploads) > maxuploads:
             uploads = uploads[:maxuploads]
@@ -304,8 +361,7 @@ class UploadsController(Controller):
         SubElement(result_elem, 'NextKeyMarker').text = nextkeymarker
         SubElement(result_elem, 'NextUploadIdMarker').text = nextuploadmarker
         if 'delimiter' in req.params:
-            SubElement(result_elem, 'Delimiter').text = \
-                req.params['delimiter']
+            SubElement(result_elem, 'Delimiter').text = req.params['delimiter']
         if 'prefix' in req.params:
             SubElement(result_elem, 'Prefix').text = req.params['prefix']
         SubElement(result_elem, 'MaxUploads').text = str(maxuploads)
@@ -318,7 +374,10 @@ class UploadsController(Controller):
         # created.
         for u in uploads:
             upload_elem = SubElement(result_elem, 'Upload')
-            SubElement(upload_elem, 'Key').text = u['key']
+            name = u['key']
+            if encoding_type == 'url':
+                name = quote(name)
+            SubElement(upload_elem, 'Key').text = name
             SubElement(upload_elem, 'UploadId').text = u['upload_id']
             initiator_elem = SubElement(upload_elem, 'Initiator')
             SubElement(initiator_elem, 'ID').text = req.user_id
@@ -334,7 +393,7 @@ class UploadsController(Controller):
             elem = SubElement(result_elem, 'CommonPrefixes')
             SubElement(elem, 'Prefix').text = p
 
-        body = tostring(result_elem, encoding_type=encoding_type)
+        body = tostring(result_elem)
 
         return HTTPOk(body=body, content_type='application/xml')
 
@@ -345,11 +404,10 @@ class UploadsController(Controller):
         """
         Handles Initiate Multipart Upload.
         """
-
         # Create a unique S3 upload id from UUID to avoid duplicates.
         upload_id = unique_id()
 
-        container = req.container_name + MULTIUPLOAD_SUFFIX
+        seg_container = req.container_name + MULTIUPLOAD_SUFFIX
         content_type = req.headers.get('Content-Type')
         if content_type:
             req.headers[sysmeta_header('object', 'has-content-type')] = 'yes'
@@ -360,16 +418,28 @@ class UploadsController(Controller):
         req.headers['Content-Type'] = 'application/directory'
 
         try:
-            req.get_response(self.app, 'PUT', container, '')
-        except BucketAlreadyExists:
-            pass
+            seg_req = copy.copy(req)
+            seg_req.environ = copy.copy(req.environ)
+            seg_req.container_name = seg_container
+            seg_req.get_container_info(self.app)
+        except NoSuchBucket:
+            try:
+                # multi-upload bucket doesn't exist, create one with
+                # same storage policy as the primary bucket
+                info = req.get_container_info(self.app)
+                policy_name = POLICIES[info['storage_policy']].name
+                hdrs = {'X-Storage-Policy': policy_name}
+                seg_req.get_response(self.app, 'PUT', seg_container, '',
+                                     headers=hdrs)
+            except (BucketAlreadyExists, BucketAlreadyOwnedByYou):
+                pass
 
         obj = '%s/%s' % (req.object_name, upload_id)
 
         req.headers.pop('Etag', None)
         req.headers.pop('Content-Md5', None)
 
-        req.get_response(self.app, 'PUT', container, obj, body='')
+        req.get_response(self.app, 'PUT', seg_container, obj, body='')
 
         result_elem = Element('InitiateMultipartUploadResult')
         SubElement(result_elem, 'Bucket').text = req.container_name
@@ -411,7 +481,7 @@ class UploadController(Controller):
             raise InvalidArgument('encoding-type', encoding_type, err_msg)
 
         upload_id = req.params['uploadId']
-        _check_upload_info(req, self.app, upload_id)
+        _get_upload_info(req, self.app, upload_id)
 
         maxparts = req.get_validated_param(
             'max-parts', DEFAULT_MAX_PARTS_LISTING,
@@ -435,7 +505,7 @@ class UploadController(Controller):
 
         # If the caller requested a list starting at a specific part number,
         # construct a sub-set of the object list.
-        objList = filter(filter_part_num_marker, objects)
+        objList = [obj for obj in objects if filter_part_num_marker(obj)]
 
         # pylint: disable-msg=E1103
         objList.sort(key=lambda o: int(o['name'].split('/')[-1]))
@@ -455,7 +525,10 @@ class UploadController(Controller):
 
         result_elem = Element('ListPartsResult')
         SubElement(result_elem, 'Bucket').text = req.container_name
-        SubElement(result_elem, 'Key').text = req.object_name
+        name = req.object_name
+        if encoding_type == 'url':
+            name = quote(name)
+        SubElement(result_elem, 'Key').text = name
         SubElement(result_elem, 'UploadId').text = upload_id
 
         initiator_elem = SubElement(result_elem, 'Initiator')
@@ -483,7 +556,7 @@ class UploadController(Controller):
             SubElement(part_elem, 'ETag').text = '"%s"' % i['hash']
             SubElement(part_elem, 'Size').text = str(i['bytes'])
 
-        body = tostring(result_elem, encoding_type=encoding_type)
+        body = tostring(result_elem)
 
         return HTTPOk(body=body, content_type='application/xml')
 
@@ -495,7 +568,7 @@ class UploadController(Controller):
         Handles Abort Multipart Upload.
         """
         upload_id = req.params['uploadId']
-        _check_upload_info(req, self.app, upload_id)
+        _get_upload_info(req, self.app, upload_id)
 
         # First check to see if this multi-part upload was already
         # completed.  Look in the primary container, if the object exists,
@@ -518,9 +591,15 @@ class UploadController(Controller):
 
         #  Iterate over the segment objects and delete them individually
         objects = json.loads(resp.body)
-        for o in objects:
-            container = req.container_name + MULTIUPLOAD_SUFFIX
-            req.get_response(self.app, container=container, obj=o['name'])
+        while objects:
+            for o in objects:
+                container = req.container_name + MULTIUPLOAD_SUFFIX
+                obj = bytes_to_wsgi(o['name'].encode('utf-8'))
+                req.get_response(self.app, container=container, obj=obj)
+            query['marker'] = objects[-1]['name']
+            resp = req.get_response(self.app, 'GET', container, '',
+                                    query=query)
+            objects = json.loads(resp.body)
 
         return HTTPNoContent()
 
@@ -533,8 +612,9 @@ class UploadController(Controller):
         """
         upload_id = req.params['uploadId']
         resp = _get_upload_info(req, self.app, upload_id)
-        headers = {}
-        for key, val in resp.headers.iteritems():
+        headers = {'Accept': 'application/json',
+                   sysmeta_header('object', 'upload-id'): upload_id}
+        for key, val in resp.headers.items():
             _key = key.lower()
             if _key.startswith('x-amz-meta-'):
                 headers['x-object-meta-' + _key[11:]] = val
@@ -555,27 +635,23 @@ class UploadController(Controller):
         if content_type:
             headers['Content-Type'] = content_type
 
-        # Query for the objects in the segments area to make sure it completed
-        query = {
-            'format': 'json',
-            'prefix': '%s/%s/' % (req.object_name, upload_id),
-            'delimiter': '/'
-        }
-
         container = req.container_name + MULTIUPLOAD_SUFFIX
-        resp = req.get_response(self.app, 'GET', container, '', query=query)
-        objinfo = json.loads(resp.body)
-        objtable = dict((o['name'],
-                         {'path': '/'.join(['', container, o['name']]),
-                          'etag': o['hash'],
-                          'size_bytes': o['bytes']}) for o in objinfo)
-
+        s3_etag_hasher = md5()
         manifest = []
         previous_number = 0
         try:
             xml = req.xml(MAX_COMPLETE_UPLOAD_BODY_SIZE)
             if not xml:
                 raise InvalidRequest(msg='You must specify at least one part')
+            if 'content-md5' in req.headers:
+                # If an MD5 was provided, we need to verify it.
+                # Note that S3Request already took care of translating to ETag
+                if req.headers['etag'] != md5(xml).hexdigest():
+                    raise BadDigest(content_md5=req.headers['content-md5'])
+                # We're only interested in the body here, in the
+                # multipart-upload controller -- *don't* let it get
+                # plumbed down to the object-server
+                del req.headers['etag']
 
             complete_elem = fromstring(
                 xml, 'CompleteMultipartUpload', self.logger)
@@ -586,20 +662,19 @@ class UploadController(Controller):
                     raise InvalidPartOrder(upload_id=upload_id)
                 previous_number = part_number
 
-                etag = part_elem.find('./ETag').text
-                if len(etag) >= 2 and etag[0] == '"' and etag[-1] == '"':
-                    # strip double quotes
-                    etag = etag[1:-1]
-
-                info = objtable.get("%s/%s/%s" % (req.object_name, upload_id,
-                                                  part_number))
-                if info is None or info['etag'] != etag:
+                etag = normalize_etag(part_elem.find('./ETag').text)
+                if len(etag) != 32 or any(c not in '0123456789abcdef'
+                                          for c in etag):
                     raise InvalidPart(upload_id=upload_id,
                                       part_number=part_number)
 
-                info['size_bytes'] = int(info['size_bytes'])
-                manifest.append(info)
+                manifest.append({
+                    'path': '/%s/%s/%s/%d' % (
+                        container, req.object_name, upload_id, part_number),
+                    'etag': etag})
+                s3_etag_hasher.update(binascii.a2b_hex(etag))
         except (XMLSyntaxError, DocumentInvalid):
+            # NB: our schema definitions catch uploads with no parts here
             raise MalformedXML()
         except ErrorResponse:
             raise
@@ -607,63 +682,114 @@ class UploadController(Controller):
             self.logger.error(e)
             raise
 
-        # Check the size of each segment except the last and make sure they are
-        # all more than the minimum upload chunk size
-        for info in manifest[:-1]:
-            if info['size_bytes'] < self.conf.min_segment_size:
-                raise EntityTooSmall()
+        s3_etag = '%s-%d' % (s3_etag_hasher.hexdigest(), len(manifest))
+        s3_etag_header = sysmeta_header('object', 'etag')
+        if resp.sysmeta_headers.get(s3_etag_header) == s3_etag:
+            # This header should only already be present if the upload marker
+            # has been cleaned up and the current target uses the same
+            # upload-id; assuming the segments to use haven't changed, the work
+            # is already done
+            return HTTPOk(body=_make_complete_body(req, s3_etag, False),
+                          content_type='application/xml')
+        headers[s3_etag_header] = s3_etag
+        # Leave base header value blank; SLO will populate
+        c_etag = '; s3_etag=%s' % s3_etag
+        headers[get_container_update_override_key('etag')] = c_etag
 
-        try:
-            # TODO: add support for versioning
-            if manifest:
-                resp = req.get_response(self.app, 'PUT',
-                                        body=json.dumps(manifest),
-                                        query={'multipart-manifest': 'put'},
-                                        headers=headers)
-            else:
-                # the upload must have consisted of a single zero-length part
-                # just write it directly
-                resp = req.get_response(self.app, 'PUT', body='',
-                                        headers=headers)
-        except BadSwiftRequest as e:
-            msg = str(e)
-            expected_msg = 'too small; each segment must be at least 1 byte'
-            if expected_msg in msg:
-                # FIXME: AWS S3 allows a smaller object than 5 MB if there is
-                # only one part.  Use a COPY request to copy the part object
-                # from the segments container instead.
-                raise EntityTooSmall(msg)
-            else:
-                raise
+        too_small_message = ('s3api requires that each segment be at least '
+                             '%d bytes' % self.conf.min_segment_size)
 
-        # clean up the multipart-upload record
-        obj = '%s/%s' % (req.object_name, upload_id)
-        try:
-            req.get_response(self.app, 'DELETE', container, obj)
-        except NoSuchKey:
-            pass  # We know that this existed long enough for us to HEAD
+        def size_checker(manifest):
+            # Check the size of each segment except the last and make sure
+            # they are all more than the minimum upload chunk size.
+            # Note that we need to use the *internal* keys, since we're
+            # looking at the manifest that's about to be written.
+            return [
+                (item['name'], too_small_message)
+                for item in manifest[:-1]
+                if item and item['bytes'] < self.conf.min_segment_size]
 
-        result_elem = Element('CompleteMultipartUploadResult')
+        req.environ['swift.callback.slo_manifest_hook'] = size_checker
+        start_time = time.time()
 
-        # NOTE: boto with sig v4 appends port to HTTP_HOST value at the
-        # request header when the port is non default value and it makes
-        # req.host_url like as http://localhost:8080:8080/path
-        # that obviously invalid. Probably it should be resolved at
-        # swift.common.swob though, tentatively we are parsing and
-        # reconstructing the correct host_url info here.
-        # in detail, https://github.com/boto/boto/pull/3513
-        parsed_url = urlparse(req.host_url)
-        host_url = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)
-        if parsed_url.port:
-            host_url += ':%s' % parsed_url.port
+        def response_iter():
+            # NB: XML requires that the XML declaration, if present, be at the
+            # very start of the document. Clients *will* call us out on not
+            # being valid XML if we pass through whitespace before it.
+            # Track whether we've sent anything yet so we can yield out that
+            # declaration *first*
+            yielded_anything = False
 
-        SubElement(result_elem, 'Location').text = host_url + req.path
-        SubElement(result_elem, 'Bucket').text = req.container_name
-        SubElement(result_elem, 'Key').text = req.object_name
-        SubElement(result_elem, 'ETag').text = resp.etag
+            try:
+                try:
+                    # TODO: add support for versioning
+                    put_resp = req.get_response(
+                        self.app, 'PUT', body=json.dumps(manifest),
+                        query={'multipart-manifest': 'put',
+                               'heartbeat': 'on'},
+                        headers=headers)
+                    if put_resp.status_int == 202:
+                        body = []
+                        put_resp.fix_conditional_response()
+                        for chunk in put_resp.response_iter:
+                            if not chunk.strip():
+                                if time.time() - start_time < 10:
+                                    # Include some grace period to keep
+                                    # ceph-s3tests happy
+                                    continue
+                                if not yielded_anything:
+                                    yield (b'<?xml version="1.0" '
+                                           b'encoding="UTF-8"?>\n')
+                                yielded_anything = True
+                                yield chunk
+                                continue
+                            body.append(chunk)
+                        body = json.loads(b''.join(body))
+                        if body['Response Status'] != '201 Created':
+                            for seg, err in body['Errors']:
+                                if err == too_small_message:
+                                    raise EntityTooSmall()
+                                elif err in ('Etag Mismatch', '404 Not Found'):
+                                    raise InvalidPart(upload_id=upload_id)
+                            raise InvalidRequest(
+                                status=body['Response Status'],
+                                msg='\n'.join(': '.join(err)
+                                              for err in body['Errors']))
+                except BadSwiftRequest as e:
+                    msg = str(e)
+                    if too_small_message in msg:
+                        raise EntityTooSmall(msg)
+                    elif ', Etag Mismatch' in msg:
+                        raise InvalidPart(upload_id=upload_id)
+                    elif ', 404 Not Found' in msg:
+                        raise InvalidPart(upload_id=upload_id)
+                    else:
+                        raise
 
-        resp.body = tostring(result_elem)
-        resp.status = 200
+                # clean up the multipart-upload record
+                obj = '%s/%s' % (req.object_name, upload_id)
+                try:
+                    req.get_response(self.app, 'DELETE', container, obj)
+                except NoSuchKey:
+                    # The important thing is that we wrote out a tombstone to
+                    # make sure the marker got cleaned up. If it's already
+                    # gone (e.g., because of concurrent completes or a retried
+                    # complete), so much the better.
+                    pass
+
+                yield _make_complete_body(req, s3_etag, yielded_anything)
+            except ErrorResponse as err_resp:
+                if yielded_anything:
+                    err_resp.xml_declaration = False
+                    yield b'\n'
+                else:
+                    # Oh good, we can still change HTTP status code, too!
+                    resp.status = err_resp.status
+                for chunk in err_resp({}, lambda *a: None):
+                    yield chunk
+
+        resp = HTTPOk()  # assume we're good for now... but see above!
+        resp.app_iter = reiterate(response_iter())
         resp.content_type = "application/xml"
 
         return resp
