@@ -82,16 +82,16 @@ Multiple Reseller Prefix Items
 
 The reseller prefix specifies which parts of the account namespace this
 middleware is responsible for managing authentication and authorization.
-By default, the prefix is 'AUTH' so accounts and tokens are prefixed
-by 'AUTH\_'. When a request's token and/or path start with 'AUTH\_', this
+By default, the prefix is ``AUTH`` so accounts and tokens are prefixed
+by ``AUTH_``. When a request's token and/or path start with ``AUTH_``, this
 middleware knows it is responsible.
 
 We allow the reseller prefix to be a list. In tempauth, the first item
 in the list is used as the prefix for tokens and user groups. The
 other prefixes provide alternate accounts that user's can access. For
-example if the reseller prefix list is 'AUTH, OTHER', a user with
-admin access to 'AUTH_account' also has admin access to
-'OTHER_account'.
+example if the reseller prefix list is ``AUTH, OTHER``, a user with
+admin access to ``AUTH_account`` also has admin access to
+``OTHER_account``.
 
 Required Group
 ^^^^^^^^^^^^^^
@@ -112,7 +112,7 @@ derived from the token are appended to the roles derived from
 
 The ``X-Service-Token`` is useful when combined with multiple reseller
 prefix items. In the following configuration, accounts prefixed
-``SERVICE\_`` are only accessible if ``X-Auth-Token`` is from the end-user
+``SERVICE_`` are only accessible if ``X-Auth-Token`` is from the end-user
 and ``X-Service-Token`` is from the ``glance`` user::
 
    [filter:tempauth]
@@ -182,9 +182,9 @@ import base64
 
 from eventlet import Timeout
 import six
-from swift.common.swob import Response, Request
+from swift.common.swob import Response, Request, wsgi_to_str
 from swift.common.swob import HTTPBadRequest, HTTPForbidden, HTTPNotFound, \
-    HTTPUnauthorized
+    HTTPUnauthorized, HTTPMethodNotAllowed
 
 from swift.common.request_helpers import get_sys_meta_prefix
 from swift.common.middleware.acl import (
@@ -239,6 +239,9 @@ class TempAuth(object):
                     account = base64.b64decode(account)
                     username += '=' * (len(username) % 4)
                     username = base64.b64decode(username)
+                    if not six.PY2:
+                        account = account.decode('utf8')
+                        username = username.decode('utf8')
                 values = conf[conf_key].split()
                 if not values:
                     raise ValueError('%s has no key set' % conf_key)
@@ -275,7 +278,7 @@ class TempAuth(object):
             return self.app(env, start_response)
         if env.get('PATH_INFO', '').startswith(self.auth_prefix):
             return self.handle(env, start_response)
-        s3 = env.get('s3api.auth_details')
+        s3 = env.get('s3api.auth_details') or env.get('swift3.auth_details')
         token = env.get('HTTP_X_AUTH_TOKEN', env.get('HTTP_X_STORAGE_TOKEN'))
         service_token = env.get('HTTP_X_SERVICE_TOKEN')
         if s3 or (token and token.startswith(self.reseller_prefix)):
@@ -438,10 +441,11 @@ class TempAuth(object):
             expires, groups = cached_auth_data
             if expires < time():
                 groups = None
-            else:
+            elif six.PY2:
                 groups = groups.encode('utf8')
 
-        s3_auth_details = env.get('s3api.auth_details')
+        s3_auth_details = env.get('s3api.auth_details') or\
+            env.get('swift3.auth_details')
         if s3_auth_details:
             if 'check_signature' not in s3_auth_details:
                 self.logger.warning(
@@ -497,7 +501,7 @@ class TempAuth(object):
         or None if there are no errors.
         """
         acl_header = 'x-account-access-control'
-        acl_data = req.headers.get(acl_header)
+        acl_data = wsgi_to_str(req.headers.get(acl_header))
         result = parse_acl(version=2, data=acl_data)
         if result is None:
             return 'Syntax error in input (%r)' % acl_data
@@ -508,19 +512,17 @@ class TempAuth(object):
             # on ACLs, TempAuth is not such an auth system.  At this point,
             # it thinks it is authoritative.
             if key not in tempauth_acl_keys:
-                return "Key %s not recognized" % json.dumps(
-                    key).encode('ascii')
+                return "Key %s not recognized" % json.dumps(key)
 
         for key in tempauth_acl_keys:
             if key not in result:
                 continue
             if not isinstance(result[key], list):
-                return "Value for key %s must be a list" % json.dumps(
-                    key).encode('ascii')
+                return "Value for key %s must be a list" % json.dumps(key)
             for grantee in result[key]:
                 if not isinstance(grantee, six.string_types):
                     return "Elements of %s list must be strings" % json.dumps(
-                        key).encode('ascii')
+                        key)
 
         # Everything looks fine, no errors found
         internal_hdr = get_sys_meta_prefix('account') + 'core-access-control'
@@ -567,7 +569,7 @@ class TempAuth(object):
                               % account_user)
             return None
 
-        if account in user_groups and \
+        if wsgi_to_str(account) in user_groups and \
                 (req.method not in ('DELETE', 'PUT') or container):
             # The user is admin for the account and is not trying to do an
             # account DELETE or PUT
@@ -664,8 +666,6 @@ class TempAuth(object):
             req = Request(env)
             if self.auth_prefix:
                 req.path_info_pop()
-            req.bytes_transferred = '-'
-            req.client_disconnect = False
             if 'x-storage-token' in req.headers and \
                     'x-auth-token' not in req.headers:
                 req.headers['x-auth-token'] = req.headers['x-storage-token']
@@ -686,6 +686,9 @@ class TempAuth(object):
         """
         req.start_time = time()
         handler = None
+        if req.method != 'GET':
+            req.response = HTTPMethodNotAllowed(request=req)
+            return req.response
         try:
             version, account, user, _junk = split_path(req.path_info,
                                                        1, 4, True)
@@ -743,7 +746,7 @@ class TempAuth(object):
                     return HTTPUnauthorized(request=req,
                                             headers={'Www-Authenticate': auth})
                 account2, user = user.split(':', 1)
-                if account != account2:
+                if wsgi_to_str(account) != account2:
                     self.logger.increment('token_denied')
                     auth = 'Swift realm="%s"' % account
                     return HTTPUnauthorized(request=req,
@@ -799,7 +802,7 @@ class TempAuth(object):
             cached_auth_data = memcache_client.get(memcache_token_key)
             if cached_auth_data:
                 expires, old_groups = cached_auth_data
-                old_groups = [group.encode('utf8')
+                old_groups = [group.encode('utf8') if six.PY2 else group
                               for group in old_groups.split(',')]
                 new_groups = self._get_user_groups(account, account_user,
                                                    account_id)
