@@ -23,6 +23,7 @@ from random import choice, random
 from struct import unpack_from
 
 from eventlet import sleep, Timeout
+from six.moves.urllib.parse import urlparse
 
 import swift.common.db
 from swift.common.db import DatabaseConnectionError
@@ -35,13 +36,16 @@ from swift.common.internal_client import (
 from swift.common.exceptions import ClientException
 from swift.common.ring import Ring
 from swift.common.ring.utils import is_local_device
+from swift.common.swob import normalize_etag
 from swift.common.utils import (
     clean_content_type, config_true_value,
-    FileLikeIter, get_logger, hash_path, quote, urlparse, validate_sync_to,
+    FileLikeIter, get_logger, hash_path, quote, validate_sync_to,
     whataremyips, Timestamp, decode_timestamps)
 from swift.common.daemon import Daemon
-from swift.common.http import HTTP_UNAUTHORIZED, HTTP_NOT_FOUND
+from swift.common.http import HTTP_UNAUTHORIZED, HTTP_NOT_FOUND, HTTP_CONFLICT
 from swift.common.wsgi import ConfigString
+from swift.common.middleware.versioned_writes.object_versioning import (
+    SYSMETA_VERSIONS_CONT, SYSMETA_VERSIONS_SYMLINK)
 
 
 # The default internal client config body is to support upgrades without
@@ -73,10 +77,15 @@ ic_conf_body = """
 # log_statsd_metric_prefix =
 
 [pipeline:main]
-pipeline = catch_errors proxy-logging cache proxy-server
+pipeline = catch_errors proxy-logging cache symlink proxy-server
 
 [app:proxy-server]
 use = egg:swift#proxy
+account_autocreate = true
+# See proxy-server.conf-sample for options
+
+[filter:symlink]
+use = egg:swift#symlink
 # See proxy-server.conf-sample for options
 
 [filter:cache]
@@ -232,9 +241,11 @@ class ContainerSync(Daemon):
             internal_client_conf = internal_client_conf_path
         try:
             self.swift = InternalClient(
-                internal_client_conf, 'Swift Container Sync', request_tries)
-        except IOError as err:
-            if err.errno != errno.ENOENT:
+                internal_client_conf, 'Swift Container Sync', request_tries,
+                use_replication_network=True)
+        except (OSError, IOError) as err:
+            if err.errno != errno.ENOENT and \
+                    not str(err).endswith(' not found'):
                 raise
             raise SystemExit(
                 _('Unable to load internal client from config: '
@@ -327,7 +338,7 @@ class ContainerSync(Daemon):
         """
         broker = None
         try:
-            broker = ContainerBroker(path)
+            broker = ContainerBroker(path, logger=self.logger)
             # The path we pass to the ContainerBroker is a real path of
             # a container DB. If we get here, however, it means that this
             # path is linked from the sync_containers dir. In rare cases
@@ -350,6 +361,13 @@ class ContainerSync(Daemon):
                     break
             else:
                 return
+            if broker.metadata.get(SYSMETA_VERSIONS_CONT):
+                self.container_skips += 1
+                self.logger.increment('skips')
+                self.logger.warning('Skipping container %s/%s with '
+                                    'object versioning configured' % (
+                                        info['account'], info['container']))
+                return
             if not broker.is_deleted():
                 sync_to = None
                 user_key = None
@@ -429,8 +447,6 @@ class ContainerSync(Daemon):
                         sync_stage_time = time()
                     self.container_syncs += 1
                     self.logger.increment('syncs')
-                except Exception as ex:
-                    raise ex
                 finally:
                     self.container_report(start_at, sync_stage_time,
                                           sync_point1,
@@ -555,7 +571,8 @@ class ContainerSync(Daemon):
                                   logger=self.logger,
                                   timeout=self.conn_timeout)
                 except ClientException as err:
-                    if err.http_status != HTTP_NOT_FOUND:
+                    if err.http_status not in (
+                            HTTP_NOT_FOUND, HTTP_CONFLICT):
                         raise
                 self.container_deletes += 1
                 self.container_stats['deletes'] += 1
@@ -569,7 +586,10 @@ class ContainerSync(Daemon):
                                                     realm_key, ts_meta):
                     return True
                 exc = None
-                # look up for the newest one
+                # look up for the newest one; the symlink=get query-string has
+                # no effect unless symlinks are enabled in the internal client
+                # in which case it ensures that symlink objects retain their
+                # symlink property when sync'd.
                 headers_out = {'X-Newest': True,
                                'X-Backend-Storage-Policy-Index':
                                str(info['storage_policy_index'])}
@@ -578,19 +598,30 @@ class ContainerSync(Daemon):
                         self.swift.get_object(info['account'],
                                               info['container'], row['name'],
                                               headers=headers_out,
-                                              acceptable_statuses=(2, 4))
+                                              acceptable_statuses=(2, 4),
+                                              params={'symlink': 'get'})
 
                 except (Exception, UnexpectedResponse, Timeout) as err:
                     headers = {}
                     body = None
                     exc = err
+
+                # skip object_versioning links; this is in case the container
+                # metadata is out of date
+                if headers.get(SYSMETA_VERSIONS_SYMLINK):
+                    self.logger.info(
+                        'Skipping versioning symlink %s/%s/%s ' % (
+                            info['account'], info['container'],
+                            row['name']))
+                    return True
+
                 timestamp = Timestamp(headers.get('x-timestamp', 0))
                 if timestamp < ts_meta:
                     if exc:
                         raise exc
                     raise Exception(
-                        _('Unknown exception trying to GET: '
-                          '%(account)r %(container)r %(object)r'),
+                        'Unknown exception trying to GET: '
+                        '%(account)r %(container)r %(object)r' %
                         {'account': info['account'],
                          'container': info['container'],
                          'object': row['name']})
@@ -598,7 +629,7 @@ class ContainerSync(Daemon):
                     if key in headers:
                         del headers[key]
                 if 'etag' in headers:
-                    headers['etag'] = headers['etag'].strip('"')
+                    headers['etag'] = normalize_etag(headers['etag'])
                 if 'content-type' in headers:
                     headers['content-type'] = clean_content_type(
                         headers['content-type'])
@@ -634,7 +665,7 @@ class ContainerSync(Daemon):
             self.container_failures += 1
             self.logger.increment('failures')
             return False
-        except (Exception, Timeout) as err:
+        except (Exception, Timeout):
             self.logger.exception(
                 _('ERROR Syncing %(db_file)s %(row)s'),
                 {'db_file': str(broker), 'row': row})
