@@ -18,7 +18,6 @@ import os
 import sys
 import time
 import signal
-import re
 from os.path import basename, dirname, join
 from random import shuffle
 from swift import gettext_ as _
@@ -28,15 +27,12 @@ from eventlet import Timeout
 from swift.obj import diskfile, replicator
 from swift.common.utils import (
     get_logger, ratelimit_sleep, dump_recon_cache, list_from_csv, listdir,
-    unlink_paths_older_than, readconf, config_auto_int_value)
+    unlink_paths_older_than, readconf, config_auto_int_value, round_robin_iter)
 from swift.common.exceptions import DiskFileQuarantined, DiskFileNotExist,\
     DiskFileDeleted, DiskFileExpired
 from swift.common.daemon import Daemon
 from swift.common.storage_policy import POLICIES
 
-# This matches rsync tempfiles, like ".<timestamp>.data.Xy095a"
-RE_RSYNC_TEMPFILE = re.compile(r'^\..*\.([a-zA-Z0-9_]){6}$')
-
 
 class AuditorWorker(object):
     """Walk through file system to audit objects"""
@@ -54,9 +50,10 @@ class AuditorWorker(object):
             # configure for their replicator
             replicator_config = readconf(self.conf['__file__'],
                                          'object-replicator')
-        except (KeyError, SystemExit):
+        except (KeyError, ValueError, IOError):
             # if we can't parse the real config (generally a KeyError on
-            # __file__, or SystemExit on no object-replicator section) we use
+            # __file__, or ValueError on no object-replicator section, or
+            # IOError if reading the file failed) we use
             # a very conservative default for rsync_timeout
             default_rsync_timeout = 86400
         else:
@@ -123,18 +120,17 @@ class AuditorWorker(object):
         total_quarantines = 0
         total_errors = 0
         time_auditing = 0
-        # TODO: we should move audit-location generation to the storage policy,
-        # as we may (conceivably) have a different filesystem layout for each.
-        # We'd still need to generate the policies to audit from the actual
-        # directories found on-disk, and have appropriate error reporting if we
-        # find a directory that doesn't correspond to any known policy. This
-        # will require a sizable refactor, but currently all diskfile managers
-        # can find all diskfile locations regardless of policy -- so for now
-        # just use Policy-0's manager.
-        all_locs = (self.diskfile_router[POLICIES[0]]
+
+        # get AuditLocations for each policy
+        loc_generators = []
+        for policy in POLICIES:
+            loc_generators.append(
+                self.diskfile_router[policy]
                     .object_audit_location_generator(
-                        device_dirs=device_dirs,
+                        policy, device_dirs=device_dirs,
                         auditor_type=self.auditor_type))
+
+        all_locs = round_robin_iter(loc_generators)
         for location in all_locs:
             loop_time = time.time()
             self.failsafe_object_audit(location)
@@ -195,8 +191,11 @@ class AuditorWorker(object):
             self.logger.info(
                 _('Object audit stats: %s') % json.dumps(self.stats_buckets))
 
-        # Unset remaining partitions to not skip them in the next run
-        diskfile.clear_auditor_status(self.devices, self.auditor_type)
+        for policy in POLICIES:
+            # Unset remaining partitions to not skip them in the next run
+            self.diskfile_router[policy].clear_auditor_status(
+                policy,
+                self.auditor_type)
 
     def record_stats(self, obj_size):
         """
@@ -243,7 +242,7 @@ class AuditorWorker(object):
         df = diskfile_mgr.get_diskfile_from_audit_location(location)
         reader = None
         try:
-            with df.open():
+            with df.open(modernize=True):
                 metadata = df.get_metadata()
                 obj_size = int(metadata['Content-Length'])
                 if self.stats_sizes:
@@ -281,8 +280,8 @@ class AuditorWorker(object):
         # _ondisk_info attr is initialized to None and filled in by open
         ondisk_info_dict = df._ondisk_info or {}
         if 'unexpected' in ondisk_info_dict:
-            is_rsync_tempfile = lambda fpath: RE_RSYNC_TEMPFILE.match(
-                basename(fpath))
+            is_rsync_tempfile = lambda fpath: (
+                diskfile.RE_RSYNC_TEMPFILE.match(basename(fpath)))
             rsync_tempfile_paths = filter(is_rsync_tempfile,
                                           ondisk_info_dict['unexpected'])
             mtime = time.time() - self.rsync_tempfile_timeout
@@ -292,9 +291,9 @@ class AuditorWorker(object):
 class ObjectAuditor(Daemon):
     """Audit objects."""
 
-    def __init__(self, conf, **options):
+    def __init__(self, conf, logger=None, **options):
         self.conf = conf
-        self.logger = get_logger(conf, log_route='object-auditor')
+        self.logger = logger or get_logger(conf, log_route='object-auditor')
         self.devices = conf.get('devices', '/srv/node')
         self.concurrency = int(conf.get('concurrency', 1))
         self.conf_zero_byte_fps = int(
@@ -322,7 +321,8 @@ class ObjectAuditor(Daemon):
                                zero_byte_only_at_fps=zero_byte_only_at_fps)
         worker.audit_all_objects(mode=mode, device_dirs=device_dirs)
 
-    def fork_child(self, zero_byte_fps=False, **kwargs):
+    def fork_child(self, zero_byte_fps=False, sleep_between_zbf_scanner=False,
+                   **kwargs):
         """Child execution"""
         pid = os.fork()
         if pid:
@@ -331,6 +331,8 @@ class ObjectAuditor(Daemon):
             signal.signal(signal.SIGTERM, signal.SIG_DFL)
             if zero_byte_fps:
                 kwargs['zero_byte_fps'] = self.conf_zero_byte_fps
+                if sleep_between_zbf_scanner:
+                    self._sleep()
             try:
                 self.run_audit(**kwargs)
             except Exception as e:
@@ -394,8 +396,9 @@ class ObjectAuditor(Daemon):
                    len(pids) > 1 and not once:
                     kwargs['device_dirs'] = override_devices
                     # sleep between ZBF scanner forks
-                    self._sleep()
-                    zbf_pid = self.fork_child(zero_byte_fps=True, **kwargs)
+                    zbf_pid = self.fork_child(zero_byte_fps=True,
+                                              sleep_between_zbf_scanner=True,
+                                              **kwargs)
                     pids.add(zbf_pid)
                 pids.discard(pid)
 
