@@ -13,7 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from six.moves import urllib
+import six
 
 from random import random
 from time import time
@@ -25,16 +25,46 @@ import hashlib
 from eventlet import sleep, Timeout
 from eventlet.greenpool import GreenPool
 
+from swift.common.constraints import AUTO_CREATE_ACCOUNT_PREFIX
 from swift.common.daemon import Daemon
 from swift.common.internal_client import InternalClient, UnexpectedResponse
 from swift.common.utils import get_logger, dump_recon_cache, split_path, \
-    Timestamp
+    Timestamp, config_true_value, normalize_delete_at_timestamp
 from swift.common.http import HTTP_NOT_FOUND, HTTP_CONFLICT, \
     HTTP_PRECONDITION_FAILED
+from swift.common.swob import wsgi_quote, str_to_wsgi
 
 from swift.container.reconciler import direct_delete_container_entry
 
 MAX_OBJECTS_TO_CACHE = 100000
+ASYNC_DELETE_TYPE = 'application/async-deleted'
+
+
+def build_task_obj(timestamp, target_account, target_container,
+                   target_obj):
+    """
+    :return: a task object name in format of
+             "<timestamp>-<target_account>/<target_container>/<target_obj>"
+    """
+    timestamp = Timestamp(timestamp)
+    return '%s-%s/%s/%s' % (
+        normalize_delete_at_timestamp(timestamp),
+        target_account, target_container, target_obj)
+
+
+def parse_task_obj(task_obj):
+    """
+    :param task_obj: a task object name in format of
+                     "<timestamp>-<target_account>/<target_container>" +
+                     "/<target_obj>"
+    :return: 4-tuples of (delete_at_time, target_account, target_container,
+             target_obj)
+    """
+    timestamp, target_path = task_obj.split('-', 1)
+    timestamp = Timestamp(timestamp)
+    target_account, target_container, target_obj = \
+        split_path('/' + target_path, 3, 3, True)
+    return timestamp, target_account, target_container, target_obj
 
 
 class ObjectExpirer(Daemon):
@@ -50,6 +80,22 @@ class ObjectExpirer(Daemon):
         self.logger = logger or get_logger(conf, log_route='object-expirer')
         self.interval = int(conf.get('interval') or 300)
 
+        self.conf_path = \
+            self.conf.get('__file__') or '/etc/swift/object-expirer.conf'
+        # True, if the conf file is 'object-expirer.conf'.
+        is_legacy_conf = 'expirer' in self.conf_path
+        # object-expirer.conf supports only legacy queue
+        self.dequeue_from_legacy = \
+            True if is_legacy_conf else \
+            config_true_value(conf.get('dequeue_from_legacy', 'false'))
+
+        if is_legacy_conf:
+            self.ic_conf_path = self.conf_path
+        else:
+            self.ic_conf_path = \
+                self.conf.get('internal_client_conf_path') or \
+                '/etc/swift/internal-client.conf'
+
         self.read_conf_for_queue_access(swift)
 
         self.report_interval = int(conf.get('report_interval') or 300)
@@ -67,19 +113,29 @@ class ObjectExpirer(Daemon):
         self.reclaim_age = int(conf.get('reclaim_age', 604800))
 
     def read_conf_for_queue_access(self, swift):
-        self.expiring_objects_account = \
-            (self.conf.get('auto_create_account_prefix') or '.') + \
+        if self.conf.get('auto_create_account_prefix'):
+            self.logger.warning('Option auto_create_account_prefix is '
+                                'deprecated. Configure '
+                                'auto_create_account_prefix under the '
+                                'swift-constraints section of '
+                                'swift.conf. This option will '
+                                'be ignored in a future release.')
+            auto_create_account_prefix = \
+                self.conf['auto_create_account_prefix']
+        else:
+            auto_create_account_prefix = AUTO_CREATE_ACCOUNT_PREFIX
+
+        self.expiring_objects_account = auto_create_account_prefix + \
             (self.conf.get('expiring_objects_account_name') or
              'expiring_objects')
 
         # This is for common parameter with general task queue in future
         self.task_container_prefix = ''
 
-        self.ic_conf_path = \
-            self.conf.get('__file__') or '/etc/swift/object-expirer.conf'
         request_tries = int(self.conf.get('request_tries') or 3)
         self.swift = swift or InternalClient(
-            self.ic_conf_path, 'Swift Object Expirer', request_tries)
+            self.ic_conf_path, 'Swift Object Expirer', request_tries,
+            use_replication_network=True)
 
         self.processes = int(self.conf.get('processes', 0))
         self.process = int(self.conf.get('process', 0))
@@ -108,18 +164,7 @@ class ObjectExpirer(Daemon):
             self.report_last_time = time()
 
     def parse_task_obj(self, task_obj):
-        """
-        :param task_obj: a task object name in format of
-                         "<timestamp>-<target_account>/<target_container>" +
-                         "/<target_obj>"
-        :return: 4-tuples of (delete_at_time, target_account, target_container,
-                 target_obj)
-        """
-        timestamp, target_path = task_obj.split('-', 1)
-        timestamp = Timestamp(timestamp)
-        target_account, target_container, target_obj = \
-            split_path('/' + target_path, 3, 3, True)
-        return timestamp, target_account, target_container, target_obj
+        return parse_task_obj(task_obj)
 
     def round_robin_order(self, task_iter):
         """
@@ -168,6 +213,8 @@ class ObjectExpirer(Daemon):
         :param divisor: a divisor number
         :return: an integer to decide which expirer is assigned to the task
         """
+        if not isinstance(name, bytes):
+            name = name.encode('utf8')
         # md5 is only used for shuffling mod
         return int(hashlib.md5(name).hexdigest(), 16) % divisor
 
@@ -215,10 +262,13 @@ class ObjectExpirer(Daemon):
         """
         for task_account, task_container in task_account_container_list:
             for o in self.swift.iter_objects(task_account, task_container):
-                task_object = o['name'].encode('utf8')
+                if six.PY2:
+                    task_object = o['name'].encode('utf8')
+                else:
+                    task_object = o['name']
                 try:
                     delete_timestamp, target_account, target_container, \
-                        target_object = self.parse_task_obj(task_object)
+                        target_object = parse_task_obj(task_object)
                 except ValueError:
                     self.logger.exception('Unexcepted error handling task %r' %
                                           task_object)
@@ -233,12 +283,14 @@ class ObjectExpirer(Daemon):
                                  divisor) != my_index:
                     continue
 
+                is_async = o.get('content_type') == ASYNC_DELETE_TYPE
                 yield {'task_account': task_account,
                        'task_container': task_container,
                        'task_object': task_object,
                        'target_path': '/'.join([
                            target_account, target_container, target_object]),
-                       'delete_timestamp': delete_timestamp}
+                       'delete_timestamp': delete_timestamp,
+                       'is_async_delete': is_async}
 
     def run_once(self, *args, **kwargs):
         """
@@ -251,6 +303,16 @@ class ObjectExpirer(Daemon):
                        These will override the values from the config file if
                        provided.
         """
+        # This if-clause will be removed when general task queue feature is
+        # implemented.
+        if not self.dequeue_from_legacy:
+            self.logger.info('This node is not configured to dequeue tasks '
+                             'from the legacy queue.  This node will '
+                             'not process any expiration tasks.  At least '
+                             'one node in your cluster must be configured '
+                             'with dequeue_from_legacy == true.')
+            return
+
         self.get_process_values(kwargs)
         pool = GreenPool(self.concurrency)
         self.report_first_time = self.report_last_time = time()
@@ -360,11 +422,13 @@ class ObjectExpirer(Daemon):
                 'process must be less than processes')
 
     def delete_object(self, target_path, delete_timestamp,
-                      task_account, task_container, task_object):
+                      task_account, task_container, task_object,
+                      is_async_delete):
         start_time = time()
         try:
             try:
-                self.delete_actual_object(target_path, delete_timestamp)
+                self.delete_actual_object(target_path, delete_timestamp,
+                                          is_async_delete)
             except UnexpectedResponse as err:
                 if err.resp.status_int not in {HTTP_NOT_FOUND,
                                                HTTP_PRECONDITION_FAILED}:
@@ -401,7 +465,7 @@ class ObjectExpirer(Daemon):
         direct_delete_container_entry(self.swift.container_ring, task_account,
                                       task_container, task_object)
 
-    def delete_actual_object(self, actual_obj, timestamp):
+    def delete_actual_object(self, actual_obj, timestamp, is_async_delete):
         """
         Deletes the end-user object indicated by the actual object name given
         '<account>/<container>/<object>' if and only if the X-Delete-At value
@@ -412,13 +476,19 @@ class ObjectExpirer(Daemon):
         :param timestamp: The swift.common.utils.Timestamp instance the
                           X-Delete-At value must match to perform the actual
                           delete.
+        :param is_async_delete: False if the object should be deleted because
+                                of "normal" expiration, or True if it should
+                                be async-deleted.
         :raises UnexpectedResponse: if the delete was unsuccessful and
                                     should be retried later
         """
-        path = '/v1/' + urllib.parse.quote(actual_obj.lstrip('/'))
-        self.swift.make_request(
-            'DELETE', path,
-            {'X-If-Delete-At': timestamp.normal,
-             'X-Timestamp': timestamp.normal,
-             'X-Backend-Clean-Expiring-Object-Queue': 'no'},
-            (2, HTTP_CONFLICT))
+        path = '/v1/' + wsgi_quote(str_to_wsgi(actual_obj.lstrip('/')))
+        if is_async_delete:
+            headers = {'X-Timestamp': timestamp.normal}
+            acceptable_statuses = (2, HTTP_CONFLICT, HTTP_NOT_FOUND)
+        else:
+            headers = {'X-Timestamp': timestamp.normal,
+                       'X-If-Delete-At': timestamp.normal,
+                       'X-Backend-Clean-Expiring-Object-Queue': 'no'}
+            acceptable_statuses = (2, HTTP_CONFLICT)
+        self.swift.make_request('DELETE', path, headers, acceptable_statuses)
