@@ -90,13 +90,13 @@ class Stats(object):
         total.suffix_hash = self.suffix_hash + other.suffix_hash
         total.suffix_sync = self.suffix_sync + other.suffix_sync
 
-        all_failed_ips = (set(self.failure_nodes.keys() +
-                              other.failure_nodes.keys()))
+        all_failed_ips = (set(list(self.failure_nodes.keys()) +
+                              list(other.failure_nodes.keys())))
         for ip in all_failed_ips:
             self_devs = self.failure_nodes.get(ip, {})
             other_devs = other.failure_nodes.get(ip, {})
             this_ip_failures = {}
-            for dev in set(self_devs.keys() + other_devs.keys()):
+            for dev in set(list(self_devs.keys()) + list(other_devs.keys())):
                 this_ip_failures[dev] = (
                     self_devs.get(dev, 0) + other_devs.get(dev, 0))
             total.failure_nodes[ip] = this_ip_failures
@@ -139,6 +139,8 @@ class ObjectReplicator(Daemon):
             int(conf.get('bind_port', 6200))
         self.concurrency = int(conf.get('concurrency', 1))
         self.replicator_workers = int(conf.get('replicator_workers', 0))
+        self.policies = [policy for policy in POLICIES
+                         if policy.policy_type == REPL_POLICY]
         self.stats_interval = int(conf.get('stats_interval', '300'))
         self.ring_check_interval = int(conf.get('ring_check_interval', 15))
         self.next_check = time.time() + self.ring_check_interval
@@ -187,6 +189,7 @@ class ObjectReplicator(Daemon):
         self.is_multiprocess_worker = None
         self._df_router = DiskFileRouter(conf, self.logger)
         self._child_process_reaper_queue = queue.LightQueue()
+        self.rings_mtime = None
 
     def _zero_stats(self):
         self.stats_for_dev = defaultdict(Stats)
@@ -204,7 +207,7 @@ class ObjectReplicator(Daemon):
     def _get_my_replication_ips(self):
         my_replication_ips = set()
         ips = whataremyips()
-        for policy in POLICIES:
+        for policy in self.policies:
             self.load_object_ring(policy)
             for local_dev in [dev for dev in policy.object_ring.devs
                               if dev and dev['replication_ip'] in ips and
@@ -291,6 +294,11 @@ class ObjectReplicator(Daemon):
         if time.time() >= self._next_rcache_update:
             update = self.aggregate_recon_update()
             dump_recon_cache(update, self.rcache, self.logger)
+        rings_mtime = [os.path.getmtime(self.load_object_ring(
+                       policy).serialized_path) for policy in self.policies]
+        if self.rings_mtime == rings_mtime:
+            return True
+        self.rings_mtime = rings_mtime
         return self.get_local_devices() == self.all_local_devices
 
     def get_local_devices(self):
@@ -303,9 +311,7 @@ class ObjectReplicator(Daemon):
         """
         ips = whataremyips(self.bind_ip)
         local_devices = set()
-        for policy in POLICIES:
-            if policy.policy_type != REPL_POLICY:
-                continue
+        for policy in self.policies:
             self.load_object_ring(policy)
             for device in policy.object_ring.devs:
                 if device and is_local_device(
@@ -394,7 +400,7 @@ class ObjectReplicator(Daemon):
             return 1  # failure response code
 
         total_time = time.time() - start_time
-        for result in results.split('\n'):
+        for result in results.decode('utf8').split('\n'):
             if result == '':
                 continue
             if result.startswith('cd+'):
@@ -576,7 +582,12 @@ class ObjectReplicator(Daemon):
 
     def delete_partition(self, path):
         self.logger.info(_("Removing partition: %s"), path)
-        tpool.execute(shutil.rmtree, path)
+        try:
+            tpool.execute(shutil.rmtree, path)
+        except OSError as e:
+            if e.errno not in (errno.ENOENT, errno.ENOTEMPTY):
+                # If there was a race to create or delete, don't worry
+                raise
 
     def delete_handoff_objs(self, job, delete_objs):
         success_paths = []
@@ -676,6 +687,9 @@ class ObjectReplicator(Daemon):
                     suffixes = [suffix for suffix in local_hash if
                                 local_hash[suffix] !=
                                 remote_hash.get(suffix, -1)]
+                    if not suffixes:
+                        stats.hashmatch += 1
+                        continue
                     stats.rsync += 1
                     success, _junk = self.sync(node, job, suffixes)
                     with Timeout(self.http_timeout):
@@ -812,8 +826,8 @@ class ObjectReplicator(Daemon):
                     self.logger.exception('ERROR creating %s' % obj_path)
                 continue
             for partition in os.listdir(obj_path):
-                if (override_partitions is not None
-                        and partition not in override_partitions):
+                if (override_partitions is not None and partition.isdigit()
+                        and int(partition) not in override_partitions):
                     continue
 
                 if (partition.startswith('auditor_status_') and
@@ -872,7 +886,7 @@ class ObjectReplicator(Daemon):
         """
         jobs = []
         ips = whataremyips(self.bind_ip)
-        for policy in POLICIES:
+        for policy in self.policies:
             # Skip replication if next_part_power is set. In this case
             # every object is hard-linked twice, but the replicator can't
             # detect them and would create a second copy of the file if not
@@ -886,15 +900,14 @@ class ObjectReplicator(Daemon):
                     policy.name)
                 continue
 
-            if policy.policy_type == REPL_POLICY:
-                if (override_policies is not None and
-                        policy.idx not in override_policies):
-                    continue
-                # ensure rings are loaded for policy
-                self.load_object_ring(policy)
-                jobs += self.build_replication_jobs(
-                    policy, ips, override_devices=override_devices,
-                    override_partitions=override_partitions)
+            if (override_policies is not None and
+                    policy.idx not in override_policies):
+                continue
+            # ensure rings are loaded for policy
+            self.load_object_ring(policy)
+            jobs += self.build_replication_jobs(
+                policy, ips, override_devices=override_devices,
+                override_partitions=override_partitions)
         random.shuffle(jobs)
         if self.handoffs_first:
             # Move the handoff parts to the front of the list
