@@ -16,8 +16,10 @@
 from swift import gettext_ as _
 import json
 
+import six
 from six.moves.urllib.parse import unquote
-from swift.common.utils import public, csv_append, Timestamp, \
+
+from swift.common.utils import public, private, csv_append, Timestamp, \
     config_true_value, ShardRange
 from swift.common.constraints import check_metadata, CONTAINER_LISTING_LIMIT
 from swift.common.http import HTTP_ACCEPTED, is_success
@@ -26,7 +28,8 @@ from swift.proxy.controllers.base import Controller, delay_denial, \
     cors_validation, set_info_cache, clear_info_cache
 from swift.common.storage_policy import POLICIES
 from swift.common.swob import HTTPBadRequest, HTTPForbidden, \
-    HTTPNotFound, HTTPServerError
+    HTTPNotFound, HTTPServiceUnavailable, str_to_wsgi, wsgi_to_str, \
+    bytes_to_wsgi
 
 
 class ContainerController(Controller):
@@ -102,7 +105,7 @@ class ContainerController(Controller):
         part = self.app.container_ring.get_part(
             self.account_name, self.container_name)
         concurrency = self.app.container_ring.replica_count \
-            if self.app.concurrent_gets else 1
+            if self.app.get_policy_options(None).concurrent_gets else 1
         node_iter = self.app.iter_nodes(self.app.container_ring, part)
         params = req.params
         params['format'] = 'json'
@@ -156,13 +159,14 @@ class ContainerController(Controller):
             return resp
 
         objects = []
-        req_limit = int(req.params.get('limit', CONTAINER_LISTING_LIMIT))
+        req_limit = int(req.params.get('limit') or CONTAINER_LISTING_LIMIT)
         params = req.params.copy()
         params.pop('states', None)
         req.headers.pop('X-Backend-Record-Type', None)
         reverse = config_true_value(params.get('reverse'))
-        marker = params.get('marker')
-        end_marker = params.get('end_marker')
+        marker = wsgi_to_str(params.get('marker'))
+        end_marker = wsgi_to_str(params.get('end_marker'))
+        prefix = wsgi_to_str(params.get('prefix'))
 
         limit = req_limit
         for shard_range in shard_ranges:
@@ -175,9 +179,9 @@ class ContainerController(Controller):
             if objects:
                 last_name = objects[-1].get('name',
                                             objects[-1].get('subdir', u''))
-                params['marker'] = last_name.encode('utf-8')
+                params['marker'] = bytes_to_wsgi(last_name.encode('utf-8'))
             elif marker:
-                params['marker'] = marker
+                params['marker'] = str_to_wsgi(marker)
             else:
                 params['marker'] = ''
             # Always set end_marker to ensure that misplaced objects beyond the
@@ -185,11 +189,11 @@ class ContainerController(Controller):
             # object obscuring correctly placed objects in the next shard
             # range.
             if end_marker and end_marker in shard_range:
-                params['end_marker'] = end_marker
+                params['end_marker'] = str_to_wsgi(end_marker)
             elif reverse:
-                params['end_marker'] = shard_range.lower_str
+                params['end_marker'] = str_to_wsgi(shard_range.lower_str)
             else:
-                params['end_marker'] = shard_range.end_marker
+                params['end_marker'] = str_to_wsgi(shard_range.end_marker)
 
             if (shard_range.account == self.account_name and
                     shard_range.container == self.container_name):
@@ -197,6 +201,18 @@ class ContainerController(Controller):
                 headers = {'X-Backend-Record-Type': 'object'}
             else:
                 headers = None
+
+            if prefix:
+                if prefix > shard_range:
+                    continue
+                try:
+                    just_past = prefix[:-1] + chr(ord(prefix[-1]) + 1)
+                except ValueError:
+                    pass
+                else:
+                    if just_past < shard_range:
+                        continue
+
             self.app.logger.debug('Getting from %s %s with %s',
                                   shard_range, shard_range.name, headers)
             objs, shard_resp = self._get_container_listing(
@@ -212,14 +228,16 @@ class ContainerController(Controller):
 
             if limit <= 0:
                 break
-            elif (end_marker and reverse and
-                  end_marker >= objects[-1]['name'].encode('utf-8')):
+            last_name = objects[-1].get('name',
+                                        objects[-1].get('subdir', u''))
+            if six.PY2:
+                last_name = last_name.encode('utf8')
+            if end_marker and reverse and end_marker >= last_name:
                 break
-            elif (end_marker and not reverse and
-                  end_marker <= objects[-1]['name'].encode('utf-8')):
+            if end_marker and not reverse and end_marker <= last_name:
                 break
 
-        resp.body = json.dumps(objects)
+        resp.body = json.dumps(objects).encode('ascii')
         constrained = any(req.params.get(constraint) for constraint in (
             'marker', 'end_marker', 'path', 'prefix', 'delimiter'))
         if not constrained and len(objects) < req_limit:
@@ -267,15 +285,15 @@ class ContainerController(Controller):
                 str(config_true_value(req.headers['X-Container-Sharding']))
         length_limit = self.get_name_length_limit()
         if len(self.container_name) > length_limit:
-            resp = HTTPBadRequest(request=req)
-            resp.body = 'Container name length of %d longer than %d' % \
-                        (len(self.container_name), length_limit)
+            body = 'Container name length of %d longer than %d' % (
+                len(self.container_name), length_limit)
+            resp = HTTPBadRequest(request=req, body=body)
             return resp
         account_partition, accounts, container_count = \
             self.account_info(self.account_name, req)
         if not accounts and self.app.account_autocreate:
             if not self.autocreate_account(req, self.account_name):
-                return HTTPServerError(request=req)
+                return HTTPServiceUnavailable(request=req)
             account_partition, accounts, container_count = \
                 self.account_info(self.account_name, req)
         if not accounts:
@@ -286,9 +304,9 @@ class ContainerController(Controller):
                 self.container_info(self.account_name, self.container_name,
                                     req)
             if not is_success(container_info.get('status')):
-                resp = HTTPForbidden(request=req)
-                resp.body = 'Reached container limit of %s' % \
-                    self.app.max_containers_per_account
+                body = 'Reached container limit of %s' % (
+                    self.app.max_containers_per_account, )
+                resp = HTTPForbidden(request=req, body=body)
                 return resp
         container_partition, containers = self.app.container_ring.get_nodes(
             self.account_name, self.container_name)
@@ -353,6 +371,27 @@ class ContainerController(Controller):
             return HTTPNotFound(request=req)
         return resp
 
+    @private
+    def UPDATE(self, req):
+        """HTTP UPDATE request handler.
+
+        Method to perform bulk operations on container DBs,
+        similar to a merge_items REPLICATE request.
+
+        Not client facing; internal clients or middlewares must include
+        ``X-Backend-Allow-Method: UPDATE`` header to access.
+        """
+        container_partition, containers = self.app.container_ring.get_nodes(
+            self.account_name, self.container_name)
+        # Since this isn't client facing, expect callers to supply an index
+        policy_index = req.headers['X-Backend-Storage-Policy-Index']
+        headers = self._backend_requests(
+            req, len(containers), account_partition=None, accounts=[],
+            policy_index=policy_index)
+        return self.make_requests(
+            req, self.app.container_ring, container_partition, 'UPDATE',
+            req.swift_entity_path, headers, body=req.body)
+
     def _backend_requests(self, req, n_outgoing, account_partition, accounts,
                           policy_index=None):
         additional = {'X-Timestamp': Timestamp.now().internal}
