@@ -29,9 +29,10 @@ from eventlet import Timeout
 
 from swift import __canonical_version__ as swift_version
 from swift.common import constraints
+from swift.common.http import is_server_error
 from swift.common.storage_policy import POLICIES
 from swift.common.ring import Ring
-from swift.common.utils import cache_from_env, get_logger, \
+from swift.common.utils import Watchdog, get_logger, \
     get_remote_client, split_path, config_true_value, generate_trans_id, \
     affinity_key_function, affinity_locality_predicate, list_from_csv, \
     register_swift_info, readconf, config_auto_int_value
@@ -39,10 +40,12 @@ from swift.common.constraints import check_utf8, valid_api_version
 from swift.proxy.controllers import AccountController, ContainerController, \
     ObjectControllerRouter, InfoController
 from swift.proxy.controllers.base import get_container_info, NodeIter, \
-    DEFAULT_RECHECK_CONTAINER_EXISTENCE, DEFAULT_RECHECK_ACCOUNT_EXISTENCE
+    DEFAULT_RECHECK_CONTAINER_EXISTENCE, DEFAULT_RECHECK_ACCOUNT_EXISTENCE, \
+    DEFAULT_RECHECK_UPDATING_SHARD_RANGES
 from swift.common.swob import HTTPBadRequest, HTTPForbidden, \
     HTTPMethodNotAllowed, HTTPNotFound, HTTPPreconditionFailed, \
-    HTTPServerError, HTTPException, Request, HTTPServiceUnavailable
+    HTTPServerError, HTTPException, Request, HTTPServiceUnavailable, \
+    wsgi_to_str
 from swift.common.exceptions import APIVersionError
 
 
@@ -87,6 +90,9 @@ def _label_for_policy(policy):
     return '(default)'
 
 
+VALID_SORTING_METHODS = ('shuffle', 'timing', 'affinity')
+
+
 class ProxyOverrideOptions(object):
     """
     Encapsulates proxy server options that may be overridden e.g. for
@@ -95,11 +101,17 @@ class ProxyOverrideOptions(object):
     :param conf: the proxy-server config dict.
     :param override_conf: a dict of overriding configuration options.
     """
-    def __init__(self, base_conf, override_conf):
+    def __init__(self, base_conf, override_conf, app):
+
         def get(key, default):
             return override_conf.get(key, base_conf.get(key, default))
 
         self.sorting_method = get('sorting_method', 'shuffle').lower()
+        if self.sorting_method not in VALID_SORTING_METHODS:
+            raise ValueError(
+                'Invalid sorting_method value; must be one of %s, not %r' % (
+                    ', '.join(VALID_SORTING_METHODS), self.sorting_method))
+
         self.read_affinity = get('read_affinity', '')
         try:
             self.read_affinity_sort_key = affinity_key_function(
@@ -107,7 +119,7 @@ class ProxyOverrideOptions(object):
         except ValueError as err:
             # make the message a little more useful
             raise ValueError("Invalid read_affinity value: %r (%s)" %
-                             (self.read_affinity, err.message))
+                             (self.read_affinity, err.args[0]))
 
         self.write_affinity = get('write_affinity', '')
         try:
@@ -116,7 +128,7 @@ class ProxyOverrideOptions(object):
         except ValueError as err:
             # make the message a little more useful
             raise ValueError("Invalid write_affinity value: %r (%s)" %
-                             (self.write_affinity, err.message))
+                             (self.write_affinity, err.args[0]))
         self.write_affinity_node_count = get(
             'write_affinity_node_count', '2 * replicas').lower()
         value = self.write_affinity_node_count.split()
@@ -136,14 +148,28 @@ class ProxyOverrideOptions(object):
             get('write_affinity_handoff_delete_count', 'auto'), None
         )
 
+        self.rebalance_missing_suppression_count = int(get(
+            'rebalance_missing_suppression_count', 1))
+        self.concurrent_gets = config_true_value(get('concurrent_gets', False))
+        self.concurrency_timeout = float(get(
+            'concurrency_timeout', app.conn_timeout))
+        self.concurrent_ec_extra_requests = int(get(
+            'concurrent_ec_extra_requests', 0))
+
     def __repr__(self):
-        return '%s({}, {%s})' % (self.__class__.__name__, ', '.join(
-            '%r: %r' % (k, getattr(self, k)) for k in (
-                'sorting_method',
-                'read_affinity',
-                'write_affinity',
-                'write_affinity_node_count',
-                'write_affinity_handoff_delete_count')))
+        return '%s({}, {%s}, app)' % (
+            self.__class__.__name__, ', '.join(
+                '%r: %r' % (k, getattr(self, k)) for k in (
+                    'sorting_method',
+                    'read_affinity',
+                    'write_affinity',
+                    'write_affinity_node_count',
+                    'write_affinity_handoff_delete_count',
+                    'rebalance_missing_suppression_count',
+                    'concurrent_gets',
+                    'concurrency_timeout',
+                    'concurrent_ec_extra_requests',
+                )))
 
     def __eq__(self, other):
         if not isinstance(other, ProxyOverrideOptions):
@@ -153,13 +179,18 @@ class ProxyOverrideOptions(object):
             'read_affinity',
             'write_affinity',
             'write_affinity_node_count',
-            'write_affinity_handoff_delete_count'))
+            'write_affinity_handoff_delete_count',
+            'rebalance_missing_suppression_count',
+            'concurrent_gets',
+            'concurrency_timeout',
+            'concurrent_ec_extra_requests',
+        ))
 
 
 class Application(object):
     """WSGI application for the proxy server."""
 
-    def __init__(self, conf, memcache=None, logger=None, account_ring=None,
+    def __init__(self, conf, logger=None, account_ring=None,
                  container_ring=None):
         if conf is None:
             conf = {}
@@ -167,10 +198,6 @@ class Application(object):
             self.logger = get_logger(conf, log_route='proxy-server')
         else:
             self.logger = logger
-        self._override_options = self._load_per_policy_config(conf)
-        self.sorts_by_timing = any(pc.sorting_method == 'timing'
-                                   for pc in self._override_options.values())
-
         self._error_limiting = {}
 
         swift_dir = conf.get('swift_dir', '/etc/swift')
@@ -179,8 +206,7 @@ class Application(object):
         self.recoverable_node_timeout = float(
             conf.get('recoverable_node_timeout', self.node_timeout))
         self.conn_timeout = float(conf.get('conn_timeout', 0.5))
-        self.client_timeout = int(conf.get('client_timeout', 60))
-        self.put_queue_depth = int(conf.get('put_queue_depth', 10))
+        self.client_timeout = float(conf.get('client_timeout', 60))
         self.object_chunk_size = int(conf.get('object_chunk_size', 65536))
         self.client_chunk_size = int(conf.get('client_chunk_size', 65536))
         self.trans_id_suffix = conf.get('trans_id_suffix', '')
@@ -192,6 +218,9 @@ class Application(object):
         self.recheck_container_existence = \
             int(conf.get('recheck_container_existence',
                          DEFAULT_RECHECK_CONTAINER_EXISTENCE))
+        self.recheck_updating_shard_ranges = \
+            int(conf.get('recheck_updating_shard_ranges',
+                         DEFAULT_RECHECK_UPDATING_SHARD_RANGES))
         self.recheck_account_existence = \
             int(conf.get('recheck_account_existence',
                          DEFAULT_RECHECK_ACCOUNT_EXISTENCE))
@@ -205,13 +234,22 @@ class Application(object):
         for policy in POLICIES:
             policy.load_ring(swift_dir)
         self.obj_controller_router = ObjectControllerRouter()
-        self.memcache = memcache
         mimetypes.init(mimetypes.knownfiles +
                        [os.path.join(swift_dir, 'mime.types')])
         self.account_autocreate = \
             config_true_value(conf.get('account_autocreate', 'no'))
-        self.auto_create_account_prefix = (
-            conf.get('auto_create_account_prefix') or '.')
+        if conf.get('auto_create_account_prefix'):
+            self.logger.warning('Option auto_create_account_prefix is '
+                                'deprecated. Configure '
+                                'auto_create_account_prefix under the '
+                                'swift-constraints section of '
+                                'swift.conf. This option will '
+                                'be ignored in a future release.')
+            self.auto_create_account_prefix = \
+                conf['auto_create_account_prefix']
+        else:
+            self.auto_create_account_prefix = \
+                constraints.AUTO_CREATE_ACCOUNT_PREFIX
         self.expiring_objects_account = self.auto_create_account_prefix + \
             (conf.get('expiring_objects_account_name') or 'expiring_objects')
         self.expiring_objects_container_divisor = \
@@ -238,10 +276,6 @@ class Application(object):
             conf.get('strict_cors_mode', 't'))
         self.node_timings = {}
         self.timing_expiry = int(conf.get('timing_expiry', 300))
-        self.concurrent_gets = \
-            config_true_value(conf.get('concurrent_gets'))
-        self.concurrency_timeout = float(conf.get('concurrency_timeout',
-                                                  self.conn_timeout))
         value = conf.get('request_node_count', '2 * replicas').lower().split()
         if len(value) == 1:
             rnc_value = int(value[0])
@@ -283,8 +317,15 @@ class Application(object):
         self.expose_info = config_true_value(
             conf.get('expose_info', 'yes'))
         self.disallowed_sections = list_from_csv(
-            conf.get('disallowed_sections', 'swift.valid_api_versions'))
+            conf.get('disallowed_sections', ', '.join([
+                'swift.auto_create_account_prefix',
+                'swift.valid_api_versions',
+            ])))
         self.admin_key = conf.get('admin_key', None)
+        self._override_options = self._load_per_policy_config(conf)
+        self.sorts_by_timing = any(pc.sorting_method == 'timing'
+                                   for pc in self._override_options.values())
+
         register_swift_info(
             version=swift_version,
             strict_cors_mode=self.strict_cors_mode,
@@ -292,16 +333,18 @@ class Application(object):
             allow_account_management=self.allow_account_management,
             account_autocreate=self.account_autocreate,
             **constraints.EFFECTIVE_CONSTRAINTS)
+        self.watchdog = Watchdog()
+        self.watchdog.spawn()
 
     def _make_policy_override(self, policy, conf, override_conf):
         label_for_policy = _label_for_policy(policy)
         try:
-            override = ProxyOverrideOptions(conf, override_conf)
+            override = ProxyOverrideOptions(conf, override_conf, self)
             self.logger.debug("Loaded override config for %s: %r" %
                               (label_for_policy, override))
             return override
         except ValueError as err:
-            raise ValueError(err.message + ' for %s' % label_for_policy)
+            raise ValueError('%s for %s' % (err, label_for_policy))
 
     def _load_per_policy_config(self, conf):
         """
@@ -333,23 +376,25 @@ class Application(object):
                 raise ValueError(
                     "No policy found for override config, index: %s" % index)
             override = self._make_policy_override(policy, conf, override_conf)
-            overrides[policy] = override
+            overrides[index] = override
         return overrides
 
     def get_policy_options(self, policy):
         """
         Return policy specific options.
 
-        :param policy: an instance of :class:`BaseStoragePolicy`
+        :param policy: an instance of :class:`BaseStoragePolicy` or ``None``
         :return: an instance of :class:`ProxyOverrideOptions`
         """
-        return self._override_options[policy]
+        return self._override_options[policy and policy.idx]
 
     def check_config(self):
         """
         Check the configuration for possible errors
         """
-        for policy, options in self._override_options.items():
+        for policy_idx, options in self._override_options.items():
+            policy = (None if policy_idx is None
+                      else POLICIES.get_by_index(policy_idx))
             if options.read_affinity and options.sorting_method != 'affinity':
                 self.logger.warning(
                     _("sorting_method is set to '%(method)s', not 'affinity'; "
@@ -383,7 +428,8 @@ class Application(object):
                      admin_key=self.admin_key)
             return InfoController, d
 
-        version, account, container, obj = split_path(req.path, 1, 4, True)
+        version, account, container, obj = split_path(
+            wsgi_to_str(req.path), 1, 4, True)
         d = dict(version=version,
                  account_name=account,
                  container_name=container,
@@ -392,6 +438,8 @@ class Application(object):
             raise APIVersionError('Invalid path')
         if obj and container and account:
             info = get_container_info(req.environ, self)
+            if is_server_error(info.get('status')):
+                raise HTTPServiceUnavailable(request=req)
             policy_index = req.headers.get('X-Backend-Storage-Policy-Index',
                                            info['storage_policy'])
             policy = POLICIES.get_by_index(policy_index)
@@ -422,8 +470,6 @@ class Application(object):
         :param start_response: WSGI callable
         """
         try:
-            if self.memcache is None:
-                self.memcache = cache_from_env(env, True)
             req = self.update_request(Request(env))
             return self.handle_request(req)(env, start_response)
         except UnicodeError:
@@ -433,12 +479,19 @@ class Application(object):
         except (Exception, Timeout):
             start_response('500 Server Error',
                            [('Content-Type', 'text/plain')])
-            return ['Internal server error.\n']
+            return [b'Internal server error.\n']
 
     def update_request(self, req):
         if 'x-storage-token' in req.headers and \
                 'x-auth-token' not in req.headers:
             req.headers['x-auth-token'] = req.headers['x-storage-token']
+        te = req.headers.get('transfer-encoding', '').lower()
+        if te.rsplit(',', 1)[-1].strip() == 'chunked' and \
+                'content-length' in req.headers:
+            # RFC says if both are present, transfer-encoding wins.
+            # Definitely *don't* forward on the header the backend
+            # ought to ignore; that offers request-smuggling vectors.
+            del req.headers['content-length']
         return req
 
     def handle_request(self, req):
@@ -456,7 +509,8 @@ class Application(object):
                                       body='Invalid Content-Length')
 
             try:
-                if not check_utf8(req.path_info):
+                if not check_utf8(wsgi_to_str(req.path_info),
+                                  internal=req.allow_reserved_names):
                     self.logger.increment('errors')
                     return HTTPPreconditionFailed(
                         request=req, body='Invalid UTF8 or contains NULL')
@@ -496,9 +550,14 @@ class Application(object):
             controller.trans_id = req.environ['swift.trans_id']
             self.logger.client_ip = get_remote_client(req)
 
-            if req.method not in controller.allowed_methods:
+            allowed_methods = controller.allowed_methods
+            if config_true_value(req.headers.get(
+                    'X-Backend-Allow-Private-Methods', False)):
+                allowed_methods = set(allowed_methods).union(
+                    controller.private_methods)
+            if req.method not in allowed_methods:
                 return HTTPMethodNotAllowed(request=req, headers={
-                    'Allow': ', '.join(controller.allowed_methods)})
+                    'Allow': ', '.join(allowed_methods)})
             handler = getattr(controller, req.method)
 
             old_authorize = None
@@ -625,8 +684,10 @@ class Application(object):
         :param msg: error message
         """
         self._incr_node_errors(node)
+        if isinstance(msg, bytes):
+            msg = msg.decode('utf-8')
         self.logger.error(_('%(msg)s %(ip)s:%(port)s/%(device)s'),
-                          {'msg': msg.decode('utf-8'), 'ip': node['ip'],
+                          {'msg': msg, 'ip': node['ip'],
                           'port': node['port'], 'device': node['device']})
 
     def iter_nodes(self, ring, partition, node_iter=None, policy=None):
