@@ -14,7 +14,8 @@
 # limitations under the License.
 
 import base64
-import unittest2
+import binascii
+import unittest
 import os
 import boto
 
@@ -23,16 +24,17 @@ import boto
 from distutils.version import StrictVersion
 
 from hashlib import md5
-from itertools import izip, izip_longest
+from six.moves import zip, zip_longest
 
 import test.functional as tf
-from swift.common.middleware.s3api.etree import fromstring, tostring, Element, \
-    SubElement
+from swift.common.middleware.s3api.etree import fromstring, tostring, \
+    Element, SubElement
 from swift.common.middleware.s3api.utils import mktime
 
 from test.functional.s3api import S3ApiBase
 from test.functional.s3api.s3_test_client import Connection
-from test.functional.s3api.utils import get_error_code, get_error_msg
+from test.functional.s3api.utils import get_error_code, get_error_msg, \
+    calculate_md5
 
 
 def setUpModule():
@@ -52,11 +54,11 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.min_segment_size = int(tf.cluster_info['s3api'].get(
             'min_segment_size', 5242880))
 
-    def _gen_comp_xml(self, etags):
+    def _gen_comp_xml(self, etags, step=1):
         elem = Element('CompleteMultipartUpload')
         for i, etag in enumerate(etags):
             elem_part = SubElement(elem, 'Part')
-            SubElement(elem_part, 'PartNumber').text = str(i + 1)
+            SubElement(elem_part, 'PartNumber').text = str(i * step + 1)
             SubElement(elem_part, 'ETag').text = etag
         return tostring(elem)
 
@@ -66,8 +68,8 @@ class TestS3ApiMultiUpload(S3ApiBase):
             headers = [None] * len(keys)
         self.conn.make_request('PUT', bucket)
         query = 'uploads'
-        for key, key_headers in izip_longest(keys, headers):
-            for i in xrange(trials):
+        for key, key_headers in zip_longest(keys, headers):
+            for i in range(trials):
                 status, resp_headers, body = \
                     self.conn.make_request('POST', bucket, key,
                                            headers=key_headers, query=query)
@@ -75,16 +77,19 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
     def _upload_part(self, bucket, key, upload_id, content=None, part_num=1):
         query = 'partNumber=%s&uploadId=%s' % (part_num, upload_id)
-        content = content if content else 'a' * self.min_segment_size
-        status, headers, body = \
-            self.conn.make_request('PUT', bucket, key, body=content,
-                                   query=query)
+        content = content if content else b'a' * self.min_segment_size
+        with self.quiet_boto_logging():
+            status, headers, body = self.conn.make_request(
+                'PUT', bucket, key, body=content, query=query)
         return status, headers, body
 
     def _upload_part_copy(self, src_bucket, src_obj, dst_bucket, dst_key,
-                          upload_id, part_num=1, src_range=None):
+                          upload_id, part_num=1, src_range=None,
+                          src_version_id=None):
 
         src_path = '%s/%s' % (src_bucket, src_obj)
+        if src_version_id:
+            src_path += '?versionId=%s' % src_version_id
         query = 'partNumber=%s&uploadId=%s' % (part_num, upload_id)
         req_headers = {'X-Amz-Copy-Source': src_path}
         if src_range:
@@ -107,8 +112,9 @@ class TestS3ApiMultiUpload(S3ApiBase):
     def test_object_multi_upload(self):
         bucket = 'bucket'
         keys = ['obj1', 'obj2', 'obj3']
-        headers = [None,
-                   {'Content-MD5': base64.b64encode('a' * 16).strip()},
+        bad_content_md5 = base64.b64encode(b'a' * 16).strip().decode('ascii')
+        headers = [{'Content-Type': 'foo/bar', 'x-amz-meta-baz': 'quux'},
+                   {'Content-MD5': bad_content_md5},
                    {'Etag': 'nonsense'}]
         uploads = []
 
@@ -117,20 +123,20 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         # Initiate Multipart Upload
         for expected_key, (status, headers, body) in \
-                izip(keys, results_generator):
-            self.assertEqual(status, 200)
+                zip(keys, results_generator):
+            self.assertEqual(status, 200, body)
             self.assertCommonResponseHeaders(headers)
-            self.assertTrue('content-type' in headers)
+            self.assertIn('content-type', headers)
             self.assertEqual(headers['content-type'], 'application/xml')
-            self.assertTrue('content-length' in headers)
+            self.assertIn('content-length', headers)
             self.assertEqual(headers['content-length'], str(len(body)))
             elem = fromstring(body, 'InitiateMultipartUploadResult')
             self.assertEqual(elem.find('Bucket').text, bucket)
             key = elem.find('Key').text
             self.assertEqual(expected_key, key)
             upload_id = elem.find('UploadId').text
-            self.assertTrue(upload_id is not None)
-            self.assertTrue((key, upload_id) not in uploads)
+            self.assertIsNotNone(upload_id)
+            self.assertNotIn((key, upload_id), uploads)
             uploads.append((key, upload_id))
 
         self.assertEqual(len(uploads), len(keys))  # sanity
@@ -156,7 +162,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.assertEqual(elem.find('IsTruncated').text, 'false')
         self.assertEqual(len(elem.findall('Upload')), 3)
         for (expected_key, expected_upload_id), u in \
-                izip(uploads, elem.findall('Upload')):
+                zip(uploads, elem.findall('Upload')):
             key = u.find('Key').text
             upload_id = u.find('UploadId').text
             self.assertEqual(expected_key, key)
@@ -173,7 +179,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         # Upload Part
         key, upload_id = uploads[0]
-        content = 'a' * self.min_segment_size
+        content = b'a' * self.min_segment_size
         etag = md5(content).hexdigest()
         status, headers, body = \
             self._upload_part(bucket, key, upload_id, content)
@@ -189,7 +195,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         key, upload_id = uploads[1]
         src_bucket = 'bucket2'
         src_obj = 'obj3'
-        src_content = 'b' * self.min_segment_size
+        src_content = b'b' * self.min_segment_size
         etag = md5(src_content).hexdigest()
 
         # prepare src obj
@@ -265,7 +271,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         # etags will be used to generate xml for Complete Multipart Upload
         etags = []
         for (expected_etag, expected_date), p in \
-                izip(expected_parts_list, elem.findall('Part')):
+                zip(expected_parts_list, elem.findall('Part')):
             last_modified = p.find('LastModified').text
             self.assertTrue(last_modified is not None)
             # TODO: sanity check
@@ -280,6 +286,114 @@ class TestS3ApiMultiUpload(S3ApiBase):
             self.assertEqual(self.min_segment_size, int(p.find('Size').text))
             etags.append(p.find('ETag').text)
 
+        # Complete Multipart Upload
+        key, upload_id = uploads[0]
+        xml = self._gen_comp_xml(etags)
+        status, headers, body = \
+            self._complete_multi_upload(bucket, key, upload_id, xml)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertIn('content-type', headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        if 'content-length' in headers:
+            self.assertEqual(headers['content-length'], str(len(body)))
+        else:
+            self.assertIn('transfer-encoding', headers)
+            self.assertEqual(headers['transfer-encoding'], 'chunked')
+        lines = body.split(b'\n')
+        self.assertTrue(lines[0].startswith(b'<?xml'), body)
+        self.assertTrue(lines[0].endswith(b'?>'), body)
+        elem = fromstring(body, 'CompleteMultipartUploadResult')
+        self.assertEqual(
+            '%s/bucket/obj1' % tf.config['s3_storage_url'].rstrip('/'),
+            elem.find('Location').text)
+        self.assertEqual(elem.find('Bucket').text, bucket)
+        self.assertEqual(elem.find('Key').text, key)
+        concatted_etags = b''.join(
+            etag.strip('"').encode('ascii') for etag in etags)
+        exp_etag = '"%s-%s"' % (
+            md5(binascii.unhexlify(concatted_etags)).hexdigest(), len(etags))
+        etag = elem.find('ETag').text
+        self.assertEqual(etag, exp_etag)
+
+        exp_size = self.min_segment_size * len(etags)
+        status, headers, body = \
+            self.conn.make_request('HEAD', bucket, key)
+        self.assertEqual(status, 200)
+        self.assertEqual(headers['content-length'], str(exp_size))
+        self.assertEqual(headers['content-type'], 'foo/bar')
+        self.assertEqual(headers['x-amz-meta-baz'], 'quux')
+
+        swift_etag = '"%s"' % md5(concatted_etags).hexdigest()
+        # TODO: GET via swift api, check against swift_etag
+
+        # Should be safe to retry
+        status, headers, body = \
+            self._complete_multi_upload(bucket, key, upload_id, xml)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertIn('content-type', headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        if 'content-length' in headers:
+            self.assertEqual(headers['content-length'], str(len(body)))
+        else:
+            self.assertIn('transfer-encoding', headers)
+            self.assertEqual(headers['transfer-encoding'], 'chunked')
+        lines = body.split(b'\n')
+        self.assertTrue(lines[0].startswith(b'<?xml'), body)
+        self.assertTrue(lines[0].endswith(b'?>'), body)
+        elem = fromstring(body, 'CompleteMultipartUploadResult')
+        self.assertEqual(
+            '%s/bucket/obj1' % tf.config['s3_storage_url'].rstrip('/'),
+            elem.find('Location').text)
+        self.assertEqual(elem.find('Bucket').text, bucket)
+        self.assertEqual(elem.find('Key').text, key)
+        self.assertEqual(elem.find('ETag').text, exp_etag)
+
+        status, headers, body = \
+            self.conn.make_request('HEAD', bucket, key)
+        self.assertEqual(status, 200)
+        self.assertEqual(headers['content-length'], str(exp_size))
+        self.assertEqual(headers['content-type'], 'foo/bar')
+        self.assertEqual(headers['x-amz-meta-baz'], 'quux')
+
+        # Upload Part Copy -- MU as source
+        key, upload_id = uploads[1]
+        status, headers, body, resp_etag = \
+            self._upload_part_copy(bucket, keys[0], bucket,
+                                   key, upload_id, part_num=2)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertIn('content-type', headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        self.assertIn('content-length', headers)
+        self.assertEqual(headers['content-length'], str(len(body)))
+        self.assertNotIn('etag', headers)
+        elem = fromstring(body, 'CopyPartResult')
+
+        last_modified = elem.find('LastModified').text
+        self.assertIsNotNone(last_modified)
+
+        exp_content = b'a' * self.min_segment_size
+        etag = md5(exp_content).hexdigest()
+        self.assertEqual(resp_etag, etag)
+
+        # Also check that the etag is correct in part listings
+        query = 'uploadId=%s' % upload_id
+        status, headers, body = \
+            self.conn.make_request('GET', bucket, key, query=query)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertTrue('content-type' in headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        self.assertTrue('content-length' in headers)
+        self.assertEqual(headers['content-length'], str(len(body)))
+        elem = fromstring(body, 'ListPartsResult')
+        self.assertEqual(len(elem.findall('Part')), 2)
+        self.assertEqual(elem.findall('Part')[1].find('PartNumber').text, '2')
+        self.assertEqual(elem.findall('Part')[1].find('ETag').text,
+                         '"%s"' % etag)
+
         # Abort Multipart Uploads
         # note that uploads[1] has part data while uploads[2] does not
         for key, upload_id in uploads[1:]:
@@ -294,26 +408,55 @@ class TestS3ApiMultiUpload(S3ApiBase):
             self.assertTrue('content-length' in headers)
             self.assertEqual(headers['content-length'], '0')
 
-        # Complete Multipart Upload
-        key, upload_id = uploads[0]
-        xml = self._gen_comp_xml(etags)
-        status, headers, body = \
-            self._complete_multi_upload(bucket, key, upload_id, xml)
+        # Check object
+        def check_obj(req_headers, exp_status):
+            status, headers, body = \
+                self.conn.make_request('HEAD', bucket, keys[0], req_headers)
+            self.assertEqual(status, exp_status)
+            self.assertCommonResponseHeaders(headers)
+            self.assertIn('content-length', headers)
+            if exp_status == 412:
+                self.assertNotIn('etag', headers)
+                self.assertEqual(headers['content-length'], '0')
+            else:
+                self.assertIn('etag', headers)
+                self.assertEqual(headers['etag'], exp_etag)
+                if exp_status == 304:
+                    self.assertEqual(headers['content-length'], '0')
+                else:
+                    self.assertEqual(headers['content-length'], str(exp_size))
+
+        check_obj({}, 200)
+
+        # Sanity check conditionals
+        check_obj({'If-Match': 'some other thing'}, 412)
+        check_obj({'If-None-Match': 'some other thing'}, 200)
+
+        # More interesting conditional cases
+        check_obj({'If-Match': exp_etag}, 200)
+        check_obj({'If-Match': swift_etag}, 412)
+        check_obj({'If-None-Match': swift_etag}, 200)
+        check_obj({'If-None-Match': exp_etag}, 304)
+
+        # Check listings
+        status, headers, body = self.conn.make_request('GET', bucket)
         self.assertEqual(status, 200)
-        self.assertCommonResponseHeaders(headers)
-        self.assertTrue('content-type' in headers)
-        self.assertEqual(headers['content-type'], 'application/xml')
-        self.assertTrue('content-length' in headers)
-        self.assertEqual(headers['content-length'], str(len(body)))
-        elem = fromstring(body, 'CompleteMultipartUploadResult')
-        # TODO: use tf.config value
-        self.assertEqual(
-            'http://%s:%s/bucket/obj1' % (self.conn.host, self.conn.port),
-            elem.find('Location').text)
-        self.assertEqual(elem.find('Bucket').text, bucket)
-        self.assertEqual(elem.find('Key').text, key)
-        # TODO: confirm completed etag value
-        self.assertTrue(elem.find('ETag').text is not None)
+
+        elem = fromstring(body, 'ListBucketResult')
+        resp_objects = list(elem.findall('./Contents'))
+        self.assertEqual(len(resp_objects), 1)
+        o = resp_objects[0]
+        self.assertEqual(o.find('Key').text, keys[0])
+        self.assertIsNotNone(o.find('LastModified').text)
+        self.assertRegexpMatches(
+            o.find('LastModified').text,
+            r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z$')
+        self.assertEqual(o.find('ETag').text, exp_etag)
+        self.assertEqual(o.find('Size').text, str(exp_size))
+        self.assertIsNotNone(o.find('StorageClass').text)
+        self.assertEqual(o.find('Owner/ID').text, self.conn.user_id)
+        self.assertEqual(o.find('Owner/DisplayName').text,
+                         self.conn.user_id)
 
     def test_initiate_multi_upload_error(self):
         bucket = 'bucket'
@@ -321,7 +464,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.conn.make_request('PUT', bucket)
         query = 'uploads'
 
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('POST', bucket, key, query=query)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -335,7 +478,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.conn.make_request('PUT', bucket)
         query = 'uploads'
 
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('GET', bucket, query=query)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -355,7 +498,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         upload_id = elem.find('UploadId').text
 
         query = 'partNumber=%s&uploadId=%s' % (1, upload_id)
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('PUT', bucket, key, query=query)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -393,7 +536,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         upload_id = elem.find('UploadId').text
 
         query = 'partNumber=%s&uploadId=%s' % (1, upload_id)
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('PUT', bucket, key,
                                          headers={
@@ -434,7 +577,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         upload_id = elem.find('UploadId').text
 
         query = 'uploadId=%s' % upload_id
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
 
         status, headers, body = \
             auth_error_conn.make_request('GET', bucket, key, query=query)
@@ -461,7 +604,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self._upload_part(bucket, key, upload_id)
 
         query = 'uploadId=%s' % upload_id
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('DELETE', bucket, key, query=query)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -490,7 +633,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         upload_id = elem.find('UploadId').text
 
         etags = []
-        for i in xrange(1, 3):
+        for i in range(1, 3):
             query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
             status, headers, body = \
                 self.conn.make_request('PUT', bucket, keys[0], query=query)
@@ -505,7 +648,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.assertEqual(get_error_code(body), 'EntityTooSmall')
 
         # invalid credentials
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('POST', bucket, keys[0], body=xml,
                                          query=query)
@@ -564,7 +707,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         # multi parts with no body
         etags = []
-        for i in xrange(1, 3):
+        for i in range(1, 3):
             query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
             status, headers, body = \
                 self.conn.make_request('PUT', bucket, key, query=query)
@@ -579,7 +722,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         # multi parts with all parts less than min segment size
         etags = []
-        for i in xrange(1, 3):
+        for i in range(1, 3):
             query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
             status, headers, body = \
                 self.conn.make_request('PUT', bucket, key, query=query,
@@ -618,11 +761,11 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         etags = []
         body_size = [self.min_segment_size, self.min_segment_size - 1, 2]
-        for i in xrange(1, 3):
+        for i in range(1, 3):
             query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
             status, headers, body = \
                 self.conn.make_request('PUT', bucket, key, query=query,
-                                       body='A' * body_size[i])
+                                       body=b'A' * body_size[i])
             etags.append(headers['etag'])
             xml = self._gen_comp_xml(etags)
 
@@ -642,11 +785,11 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         etags = []
         body_size = [self.min_segment_size, self.min_segment_size, 2]
-        for i in xrange(1, 3):
+        for i in range(1, 3):
             query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
             status, headers, body = \
                 self.conn.make_request('PUT', bucket, key, query=query,
-                                       body='A' * body_size[i])
+                                       body=b'A' * body_size[i])
             etags.append(headers['etag'])
             xml = self._gen_comp_xml(etags)
 
@@ -667,14 +810,14 @@ class TestS3ApiMultiUpload(S3ApiBase):
         upload_id = elem.find('UploadId').text
 
         etags = []
-        for i in xrange(1, 4):
-            query = 'partNumber=%s&uploadId=%s' % (i, upload_id)
-            status, headers, body = \
-                self.conn.make_request('PUT', bucket, key,
-                                       body='A' * 1024 * 1024 * 5, query=query)
+        for i in range(1, 4):
+            query = 'partNumber=%s&uploadId=%s' % (2 * i - 1, upload_id)
+            status, headers, body = self.conn.make_request(
+                'PUT', bucket, key, body=b'A' * 1024 * 1024 * 5,
+                query=query)
             etags.append(headers['etag'])
         query = 'uploadId=%s' % upload_id
-        xml = self._gen_comp_xml(etags[:-1])
+        xml = self._gen_comp_xml(etags[:-1], step=2)
         status, headers, body = \
             self.conn.make_request('POST', bucket, key, body=xml,
                                    query=query)
@@ -690,7 +833,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
 
         # Initiate Multipart Upload
         for expected_key, (status, headers, body) in \
-                izip(keys, results_generator):
+                zip(keys, results_generator):
             self.assertEqual(status, 200)
             self.assertCommonResponseHeaders(headers)
             self.assertTrue('content-type' in headers)
@@ -712,7 +855,7 @@ class TestS3ApiMultiUpload(S3ApiBase):
         key, upload_id = uploads[0]
         src_bucket = 'bucket2'
         src_obj = 'obj4'
-        src_content = 'y' * (self.min_segment_size / 2) + 'z' * \
+        src_content = b'y' * (self.min_segment_size // 2) + b'z' * \
             self.min_segment_size
         src_range = 'bytes=0-%d' % (self.min_segment_size - 1)
         etag = md5(src_content[:self.min_segment_size]).hexdigest()
@@ -773,6 +916,135 @@ class TestS3ApiMultiUpload(S3ApiBase):
         self.assertTrue('content-length' in headers)
         self.assertEqual(headers['content-length'], '0')
 
+    def test_object_multi_upload_part_copy_version(self):
+        if 'object_versioning' not in tf.cluster_info:
+            self.skipTest('Object Versioning not enabled')
+        bucket = 'bucket'
+        keys = ['obj1']
+        uploads = []
+
+        results_generator = self._initiate_multi_uploads_result_generator(
+            bucket, keys)
+
+        # Initiate Multipart Upload
+        for expected_key, (status, headers, body) in \
+                zip(keys, results_generator):
+            self.assertEqual(status, 200)
+            self.assertCommonResponseHeaders(headers)
+            self.assertTrue('content-type' in headers)
+            self.assertEqual(headers['content-type'], 'application/xml')
+            self.assertTrue('content-length' in headers)
+            self.assertEqual(headers['content-length'], str(len(body)))
+            elem = fromstring(body, 'InitiateMultipartUploadResult')
+            self.assertEqual(elem.find('Bucket').text, bucket)
+            key = elem.find('Key').text
+            self.assertEqual(expected_key, key)
+            upload_id = elem.find('UploadId').text
+            self.assertTrue(upload_id is not None)
+            self.assertTrue((key, upload_id) not in uploads)
+            uploads.append((key, upload_id))
+
+        self.assertEqual(len(uploads), len(keys))  # sanity
+
+        key, upload_id = uploads[0]
+        src_bucket = 'bucket2'
+        src_obj = 'obj4'
+        src_content = b'y' * (self.min_segment_size // 2) + b'z' * \
+            self.min_segment_size
+        etags = [md5(src_content).hexdigest()]
+
+        # prepare null-version src obj
+        self.conn.make_request('PUT', src_bucket)
+        self.conn.make_request('PUT', src_bucket, src_obj, body=src_content)
+        _, headers, _ = self.conn.make_request('HEAD', src_bucket, src_obj)
+        self.assertCommonResponseHeaders(headers)
+
+        # Turn on versioning
+        elem = Element('VersioningConfiguration')
+        SubElement(elem, 'Status').text = 'Enabled'
+        xml = tostring(elem)
+        status, headers, body = self.conn.make_request(
+            'PUT', src_bucket, body=xml, query='versioning')
+        self.assertEqual(status, 200)
+
+        src_obj2 = 'obj5'
+        src_content2 = b'stub'
+        etags.append(md5(src_content2).hexdigest())
+
+        # prepare src obj w/ real version
+        self.conn.make_request('PUT', src_bucket, src_obj2, body=src_content2)
+        _, headers, _ = self.conn.make_request('HEAD', src_bucket, src_obj2)
+        self.assertCommonResponseHeaders(headers)
+        version_id2 = headers['x-amz-version-id']
+
+        status, headers, body, resp_etag = \
+            self._upload_part_copy(src_bucket, src_obj, bucket,
+                                   key, upload_id, 1,
+                                   src_version_id='null')
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertTrue('content-type' in headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        self.assertTrue('content-length' in headers)
+        self.assertEqual(headers['content-length'], str(len(body)))
+        self.assertTrue('etag' not in headers)
+        elem = fromstring(body, 'CopyPartResult')
+
+        last_modifieds = [elem.find('LastModified').text]
+        self.assertTrue(last_modifieds[0] is not None)
+
+        self.assertEqual(resp_etag, etags[0])
+
+        status, headers, body, resp_etag = \
+            self._upload_part_copy(src_bucket, src_obj2, bucket,
+                                   key, upload_id, 2,
+                                   src_version_id=version_id2)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        self.assertTrue('content-type' in headers)
+        self.assertEqual(headers['content-type'], 'application/xml')
+        self.assertTrue('content-length' in headers)
+        self.assertEqual(headers['content-length'], str(len(body)))
+        self.assertTrue('etag' not in headers)
+        elem = fromstring(body, 'CopyPartResult')
+
+        last_modifieds.append(elem.find('LastModified').text)
+        self.assertTrue(last_modifieds[1] is not None)
+
+        self.assertEqual(resp_etag, etags[1])
+
+        # Check last-modified timestamp
+        key, upload_id = uploads[0]
+        query = 'uploadId=%s' % upload_id
+        status, headers, body = \
+            self.conn.make_request('GET', bucket, key, query=query)
+
+        elem = fromstring(body, 'ListPartsResult')
+
+        # FIXME: COPY result drops milli/microseconds but GET doesn't
+        last_modified_gets = [p.find('LastModified').text
+                              for p in elem.iterfind('Part')]
+        self.assertEqual(
+            [lm.rsplit('.', 1)[0] for lm in last_modified_gets],
+            [lm.rsplit('.', 1)[0] for lm in last_modifieds])
+
+        # There should be *exactly* two parts in the result
+        self.assertEqual(2, len(last_modified_gets))
+
+        # Abort Multipart Upload
+        key, upload_id = uploads[0]
+        query = 'uploadId=%s' % upload_id
+        status, headers, body = \
+            self.conn.make_request('DELETE', bucket, key, query=query)
+
+        # sanity checks
+        self.assertEqual(status, 204)
+        self.assertCommonResponseHeaders(headers)
+        self.assertTrue('content-type' in headers)
+        self.assertEqual(headers['content-type'], 'text/html; charset=UTF-8')
+        self.assertTrue('content-length' in headers)
+        self.assertEqual(headers['content-length'], '0')
+
 
 class TestS3ApiMultiUploadSigV4(TestS3ApiMultiUpload):
     @classmethod
@@ -788,6 +1060,11 @@ class TestS3ApiMultiUploadSigV4(TestS3ApiMultiUpload):
 
     def test_object_multi_upload_part_copy_range(self):
         if StrictVersion(boto.__version__) < StrictVersion('3.0'):
+            # boto 2 doesn't sort headers properly; see
+            # https://github.com/boto/boto/pull/3032
+            # or https://github.com/boto/boto/pull/3176
+            # or https://github.com/boto/boto/pull/3751
+            # or https://github.com/boto/boto/pull/3824
             self.skipTest('This stuff got the issue of boto<=2.x')
 
     def test_delete_bucket_multi_upload_object_exisiting(self):
@@ -800,7 +1077,7 @@ class TestS3ApiMultiUploadSigV4(TestS3ApiMultiUpload):
 
         # Initiate Multipart Upload
         for expected_key, (status, _, body) in \
-                izip(keys, results_generator):
+                zip(keys, results_generator):
             self.assertEqual(status, 200)  # sanity
             elem = fromstring(body, 'InitiateMultipartUploadResult')
             key = elem.find('Key').text
@@ -814,7 +1091,7 @@ class TestS3ApiMultiUploadSigV4(TestS3ApiMultiUpload):
 
         # Upload Part
         key, upload_id = uploads[0]
-        content = 'a' * self.min_segment_size
+        content = b'a' * self.min_segment_size
         status, headers, body = \
             self._upload_part(bucket, key, upload_id, content)
         self.assertEqual(status, 200)
@@ -844,6 +1121,28 @@ class TestS3ApiMultiUploadSigV4(TestS3ApiMultiUpload):
         self.assertEqual(status, 200)  # sanity
         self.assertEqual(content, body)  # sanity
 
+        # Can delete it with DeleteMultipleObjects request
+        elem = Element('Delete')
+        SubElement(elem, 'Quiet').text = 'true'
+        obj_elem = SubElement(elem, 'Object')
+        SubElement(obj_elem, 'Key').text = key
+        body = tostring(elem, use_s3ns=False)
+
+        status, headers, body = self.conn.make_request(
+            'POST', bucket, body=body, query='delete',
+            headers={'Content-MD5': calculate_md5(body)})
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+
+        status, headers, body = \
+            self.conn.make_request('GET', bucket, key)
+        self.assertEqual(status, 404)  # sanity
+
+        # Now we can delete
+        status, headers, body = \
+            self.conn.make_request('DELETE', bucket)
+        self.assertEqual(status, 204)  # sanity
+
 
 if __name__ == '__main__':
-    unittest2.main()
+    unittest.main()
