@@ -13,7 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import unittest2
+import unittest
 import os
 import boto
 
@@ -25,7 +25,8 @@ import email.parser
 from email.utils import formatdate, parsedate
 from time import mktime
 from hashlib import md5
-from urllib import quote
+import six
+from six.moves.urllib.parse import quote
 
 import test.functional as tf
 
@@ -59,7 +60,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_object(self):
         obj = 'object name with %-sign'
-        content = 'abc123'
+        content = b'abc123'
         etag = md5(content).hexdigest()
 
         # PUT Object
@@ -146,7 +147,7 @@ class TestS3ApiObject(S3ApiBase):
         self.assertCommonResponseHeaders(headers)
 
     def test_put_object_error(self):
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('PUT', self.bucket, 'object')
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -165,7 +166,7 @@ class TestS3ApiObject(S3ApiBase):
         dst_obj = 'dst_object'
 
         headers = {'x-amz-copy-source': '/%s/%s' % (self.bucket, obj)}
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('PUT', dst_bucket, dst_obj, headers)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -196,7 +197,7 @@ class TestS3ApiObject(S3ApiBase):
         obj = 'object'
         self.conn.make_request('PUT', self.bucket, obj)
 
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('GET', self.bucket, obj)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -215,30 +216,30 @@ class TestS3ApiObject(S3ApiBase):
         obj = 'object'
         self.conn.make_request('PUT', self.bucket, obj)
 
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('HEAD', self.bucket, obj)
         self.assertEqual(status, 403)
-        self.assertEqual(body, '')  # sanity
+        self.assertEqual(body, b'')  # sanity
         self.assertEqual(headers['content-type'], 'application/xml')
 
         status, headers, body = \
             self.conn.make_request('HEAD', self.bucket, 'invalid')
         self.assertEqual(status, 404)
-        self.assertEqual(body, '')  # sanity
+        self.assertEqual(body, b'')  # sanity
         self.assertEqual(headers['content-type'], 'application/xml')
 
         status, headers, body = \
             self.conn.make_request('HEAD', 'invalid', obj)
         self.assertEqual(status, 404)
-        self.assertEqual(body, '')  # sanity
+        self.assertEqual(body, b'')  # sanity
         self.assertEqual(headers['content-type'], 'application/xml')
 
     def test_delete_object_error(self):
         obj = 'object'
         self.conn.make_request('PUT', self.bucket, obj)
 
-        auth_error_conn = Connection(aws_secret_key='invalid')
+        auth_error_conn = Connection(tf.config['s3_access_key'], 'invalid')
         status, headers, body = \
             auth_error_conn.make_request('DELETE', self.bucket, obj)
         self.assertEqual(get_error_code(body), 'SignatureDoesNotMatch')
@@ -265,7 +266,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_content_md5(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         headers = {'Content-MD5': calculate_md5(content)}
         status, headers, body = \
@@ -276,7 +277,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_content_type(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         headers = {'Content-Type': 'text/plain'}
         status, headers, body = \
@@ -290,7 +291,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_conditional_requests(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         headers = {'If-None-Match': '*'}
         status, headers, body = \
             self.conn.make_request('PUT', self.bucket, obj, headers, content)
@@ -318,7 +319,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_expect(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         headers = {'Expect': '100-continue'}
         status, headers, body = \
@@ -331,7 +332,7 @@ class TestS3ApiObject(S3ApiBase):
         if expected_headers is None:
             expected_headers = req_headers
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         status, headers, body = \
             self.conn.make_request('PUT', self.bucket, obj,
@@ -387,7 +388,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_storage_class(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         headers = {'X-Amz-Storage-Class': 'STANDARD'}
         status, headers, body = \
@@ -399,7 +400,7 @@ class TestS3ApiObject(S3ApiBase):
     def test_put_object_copy_source_params(self):
         obj = 'object'
         src_headers = {'X-Amz-Meta-Test': 'src'}
-        src_body = 'some content'
+        src_body = b'some content'
         dst_bucket = 'dst-bucket'
         dst_obj = 'dst_object'
         self.conn.make_request('PUT', self.bucket, obj, src_headers, src_body)
@@ -433,7 +434,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_put_object_copy_source(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         etag = md5(content).hexdigest()
         self.conn.make_request('PUT', self.bucket, obj, body=content)
 
@@ -487,6 +488,35 @@ class TestS3ApiObject(S3ApiBase):
             self.conn.make_request('HEAD', dst_bucket, dst_obj)
         self.assertEqual(headers['x-amz-meta-test'], 'dst')
 
+        headers = {'X-Amz-Copy-Source': '/%s/%s' % (self.bucket, obj),
+                   'X-Amz-Metadata-Directive': 'COPY',
+                   'X-Amz-Meta-Test': 'dst'}
+        status, headers, body = \
+            self.conn.make_request('PUT', dst_bucket, dst_obj, headers)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        status, headers, body = \
+            self.conn.make_request('HEAD', dst_bucket, dst_obj)
+        self.assertEqual(headers['x-amz-meta-test'], 'src')
+
+        headers = {'X-Amz-Copy-Source': '/%s/%s' % (self.bucket, obj),
+                   'X-Amz-Meta-Test2': 'dst',
+                   'X-Amz-Metadata-Directive': 'REPLACE'}
+        status, headers, body = \
+            self.conn.make_request('PUT', dst_bucket, dst_obj, headers)
+        self.assertEqual(status, 200)
+        self.assertCommonResponseHeaders(headers)
+        status, headers, body = \
+            self.conn.make_request('HEAD', dst_bucket, dst_obj)
+        self.assertNotIn('x-amz-meta-test', headers)
+        self.assertEqual(headers['x-amz-meta-test2'], 'dst')
+
+        headers = {'X-Amz-Copy-Source': '/%s/%s' % (self.bucket, obj),
+                   'X-Amz-Metadata-Directive': 'BAD'}
+        status, headers, body = \
+            self.conn.make_request('PUT', dst_bucket, dst_obj, headers)
+        self.assertEqual(status, 400)
+
     def test_put_object_copy_source_if_modified_since(self):
         obj = 'object'
         dst_bucket = 'dst-bucket'
@@ -619,8 +649,9 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_get_object_range(self):
         obj = 'object'
-        content = 'abcdefghij'
-        headers = {'x-amz-meta-test': 'swift'}
+        content = b'abcdefghij'
+        headers = {'x-amz-meta-test': 'swift',
+                   'content-type': 'application/octet-stream'}
         self.conn.make_request(
             'PUT', self.bucket, obj, headers=headers, body=content)
 
@@ -633,7 +664,8 @@ class TestS3ApiObject(S3ApiBase):
         self.assertEqual(headers['content-length'], '5')
         self.assertTrue('x-amz-meta-test' in headers)
         self.assertEqual('swift', headers['x-amz-meta-test'])
-        self.assertEqual(body, 'bcdef')
+        self.assertEqual(body, b'bcdef')
+        self.assertEqual('application/octet-stream', headers['content-type'])
 
         headers = {'Range': 'bytes=5-'}
         status, headers, body = \
@@ -644,7 +676,7 @@ class TestS3ApiObject(S3ApiBase):
         self.assertEqual(headers['content-length'], '5')
         self.assertTrue('x-amz-meta-test' in headers)
         self.assertEqual('swift', headers['x-amz-meta-test'])
-        self.assertEqual(body, 'fghij')
+        self.assertEqual(body, b'fghij')
 
         headers = {'Range': 'bytes=-5'}
         status, headers, body = \
@@ -655,7 +687,7 @@ class TestS3ApiObject(S3ApiBase):
         self.assertEqual(headers['content-length'], '5')
         self.assertTrue('x-amz-meta-test' in headers)
         self.assertEqual('swift', headers['x-amz-meta-test'])
-        self.assertEqual(body, 'fghij')
+        self.assertEqual(body, b'fghij')
 
         ranges = ['1-2', '4-5']
 
@@ -664,9 +696,9 @@ class TestS3ApiObject(S3ApiBase):
             self.conn.make_request('GET', self.bucket, obj, headers=headers)
         self.assertEqual(status, 206)
         self.assertCommonResponseHeaders(headers)
-        self.assertTrue('content-length' in headers)
+        self.assertIn('content-length', headers)
 
-        self.assertTrue('content-type' in headers)  # sanity
+        self.assertIn('content-type', headers)  # sanity
         content_type, boundary = headers['content-type'].split(';')
 
         self.assertEqual('multipart/byteranges', content_type)
@@ -675,10 +707,13 @@ class TestS3ApiObject(S3ApiBase):
 
         # TODO: Using swift.common.utils.multipart_byteranges_to_document_iters
         #       could be easy enough.
-        parser = email.parser.FeedParser()
+        if six.PY2:
+            parser = email.parser.FeedParser()
+        else:
+            parser = email.parser.BytesFeedParser()
         parser.feed(
-            "Content-Type: multipart/byterange; boundary=%s\r\n\r\n" %
-            boundary_str)
+            b"Content-Type: multipart/byterange; boundary=%s\r\n\r\n" %
+            boundary_str.encode('ascii'))
         parser.feed(body)
         message = parser.close()
 
@@ -698,7 +733,7 @@ class TestS3ApiObject(S3ApiBase):
             self.assertEqual(
                 expected_range, part.get('Content-Range'))
             # rest
-            payload = part.get_payload().strip()
+            payload = part.get_payload(decode=True).strip()
             self.assertEqual(content[start:end + 1], payload)
 
     def test_get_object_if_modified_since(self):
@@ -754,7 +789,7 @@ class TestS3ApiObject(S3ApiBase):
 
     def test_head_object_range(self):
         obj = 'object'
-        content = 'abcdefghij'
+        content = b'abcdefghij'
         self.conn.make_request('PUT', self.bucket, obj, body=content)
 
         headers = {'Range': 'bytes=1-5'}
@@ -840,35 +875,35 @@ class TestS3ApiObjectSigV4(TestS3ApiObject):
     def setUp(self):
         super(TestS3ApiObjectSigV4, self).setUp()
 
-    @unittest2.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
-                      'This stuff got the signing issue of boto<=2.x')
+    @unittest.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
+                     'This stuff got the signing issue of boto<=2.x')
     def test_put_object_metadata(self):
         super(TestS3ApiObjectSigV4, self).test_put_object_metadata()
 
-    @unittest2.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
-                      'This stuff got the signing issue of boto<=2.x')
+    @unittest.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
+                     'This stuff got the signing issue of boto<=2.x')
     def test_put_object_copy_source_if_modified_since(self):
         super(TestS3ApiObjectSigV4, self).\
             test_put_object_copy_source_if_modified_since()
 
-    @unittest2.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
-                      'This stuff got the signing issue of boto<=2.x')
+    @unittest.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
+                     'This stuff got the signing issue of boto<=2.x')
     def test_put_object_copy_source_if_unmodified_since(self):
         super(TestS3ApiObjectSigV4, self).\
             test_put_object_copy_source_if_unmodified_since()
 
-    @unittest2.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
-                      'This stuff got the signing issue of boto<=2.x')
+    @unittest.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
+                     'This stuff got the signing issue of boto<=2.x')
     def test_put_object_copy_source_if_match(self):
         super(TestS3ApiObjectSigV4,
               self).test_put_object_copy_source_if_match()
 
-    @unittest2.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
-                      'This stuff got the signing issue of boto<=2.x')
+    @unittest.skipIf(StrictVersion(boto.__version__) < StrictVersion('3.0'),
+                     'This stuff got the signing issue of boto<=2.x')
     def test_put_object_copy_source_if_none_match(self):
         super(TestS3ApiObjectSigV4,
               self).test_put_object_copy_source_if_none_match()
 
 
 if __name__ == '__main__':
-    unittest2.main()
+    unittest.main()
