@@ -16,16 +16,18 @@
 # limitations under the License.
 
 import datetime
+import hashlib
 import json
-import unittest2
-from unittest2 import SkipTest
+import unittest
 from uuid import uuid4
 import time
+from xml.dom import minidom
 
+import six
 from six.moves import range
 
 from test.functional import check_response, retry, requires_acls, \
-    requires_policies
+    requires_policies, SkipTest, requires_bulk
 import test.functional as tf
 
 
@@ -37,11 +39,14 @@ def tearDownModule():
     tf.teardown_package()
 
 
-class TestObject(unittest2.TestCase):
+class TestObject(unittest.TestCase):
 
     def setUp(self):
-        if tf.skip:
+        if tf.skip or tf.skip2:
             raise SkipTest
+
+        if tf.in_process:
+            tf.skip_if_no_xattrs()
         self.container = uuid4().hex
 
         self.containers = []
@@ -72,7 +77,7 @@ class TestObject(unittest2.TestCase):
             return check_response(conn)
         resp = retry(put, name, use_account=use_account)
         resp.read()
-        self.assertEqual(resp.status, 201)
+        self.assertIn(resp.status, (201, 202))
 
         # With keystoneauth we need the accounts to have had the project
         # domain id persisted as sysmeta prior to testing ACLs. This may
@@ -102,9 +107,12 @@ class TestObject(unittest2.TestCase):
 
         # delete an object
         def delete(url, token, parsed, conn, container, obj):
-            conn.request(
-                'DELETE', '/'.join([parsed.path, container, obj['name']]), '',
-                {'X-Auth-Token': token})
+            if six.PY2:
+                obj_name = obj['name'].encode('utf8')
+            else:
+                obj_name = obj['name']
+            path = '/'.join([parsed.path, container, obj_name])
+            conn.request('DELETE', path, '', {'X-Auth-Token': token})
             return check_response(conn)
 
         for container in self.containers:
@@ -173,7 +181,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 201)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {})
         # empty post
@@ -181,7 +189,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 202)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {})
 
@@ -194,7 +202,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 201)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Color': 'blUe',
@@ -206,7 +214,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 202)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Color': 'oraNge'
@@ -222,7 +230,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 201)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Color': 'Red',
@@ -238,7 +246,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 202)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Food': 'Burger',
@@ -253,7 +261,7 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 201)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Foo': 'B\xc3\xa2r',
@@ -266,13 +274,19 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 202)
         resp = retry(get)
-        self.assertEqual('', resp.read())
+        self.assertEqual(b'', resp.read())
         self.assertEqual(resp.status, 200)
         self.assertEqual(metadata(resp), {
             'X-Object-Meta-Foo': 'B\xc3\xa5z',
         })
 
     def test_if_none_match(self):
+        def delete(url, token, parsed, conn):
+            conn.request('DELETE', '%s/%s/%s' % (
+                parsed.path, self.container, 'if_none_match_test'), '',
+                {'X-Auth-Token': token})
+            return check_response(conn)
+
         def put(url, token, parsed, conn):
             conn.request('PUT', '%s/%s/%s' % (
                 parsed.path, self.container, 'if_none_match_test'), '',
@@ -287,6 +301,13 @@ class TestObject(unittest2.TestCase):
         resp.read()
         self.assertEqual(resp.status, 412)
 
+        resp = retry(delete)
+        resp.read()
+        self.assertEqual(resp.status, 204)
+        resp = retry(put)
+        resp.read()
+        self.assertEqual(resp.status, 201)
+
         def put(url, token, parsed, conn):
             conn.request('PUT', '%s/%s/%s' % (
                 parsed.path, self.container, 'if_none_match_test'), '',
@@ -314,8 +335,10 @@ class TestObject(unittest2.TestCase):
                               'Content-Length': '0'})
             return check_response(conn)
         ts_before = time.time()
+        time.sleep(0.05)
         resp = retry(put)
         body = resp.read()
+        time.sleep(0.05)
         ts_after = time.time()
         if resp.status == 400:
             # shunt_inbound_x_timestamp must be false
@@ -323,7 +346,7 @@ class TestObject(unittest2.TestCase):
                 'X-Timestamp should be a UNIX timestamp float value', body)
         else:
             self.assertEqual(resp.status, 201)
-            self.assertEqual(body, '')
+            self.assertEqual(body, b'')
             resp = retry(head)
             resp.read()
             self.assertGreater(float(resp.headers['x-timestamp']), ts_before)
@@ -345,8 +368,10 @@ class TestObject(unittest2.TestCase):
                               'Content-Length': '0'})
             return check_response(conn)
         ts_before = time.time()
+        time.sleep(0.05)
         resp = retry(put)
         body = resp.read()
+        time.sleep(0.05)
         ts_after = time.time()
         if resp.status == 400:
             # shunt_inbound_x_timestamp must be false
@@ -354,7 +379,7 @@ class TestObject(unittest2.TestCase):
                 'X-Timestamp should be a UNIX timestamp float value', body)
         else:
             self.assertEqual(resp.status, 201)
-            self.assertEqual(body, '')
+            self.assertEqual(body, b'')
             resp = retry(head)
             resp.read()
             self.assertGreater(float(resp.headers['x-timestamp']), ts_before)
@@ -366,7 +391,7 @@ class TestObject(unittest2.TestCase):
                                               'x_delete_after'),
                          '', {'X-Auth-Token': token,
                               'Content-Length': '0',
-                              'X-Delete-After': '1'})
+                              'X-Delete-After': '2'})
             return check_response(conn)
         resp = retry(put)
         resp.read()
@@ -387,7 +412,7 @@ class TestObject(unittest2.TestCase):
             resp = retry(get)
             resp.read()
             count += 1
-            time.sleep(1)
+            time.sleep(0.5)
 
         self.assertEqual(resp.status, 404)
 
@@ -450,7 +475,7 @@ class TestObject(unittest2.TestCase):
         resp = retry(put)
         body = resp.read()
         self.assertEqual(resp.status, 400)
-        self.assertEqual(body, 'Non-integer X-Delete-After')
+        self.assertEqual(body, b'Non-integer X-Delete-After')
 
     def test_non_integer_x_delete_at(self):
         def put(url, token, parsed, conn):
@@ -463,7 +488,7 @@ class TestObject(unittest2.TestCase):
         resp = retry(put)
         body = resp.read()
         self.assertEqual(resp.status, 400)
-        self.assertEqual(body, 'Non-integer X-Delete-At')
+        self.assertEqual(body, b'Non-integer X-Delete-At')
 
     def test_x_delete_at_in_the_past(self):
         def put(url, token, parsed, conn):
@@ -476,7 +501,30 @@ class TestObject(unittest2.TestCase):
         resp = retry(put)
         body = resp.read()
         self.assertEqual(resp.status, 400)
-        self.assertEqual(body, 'X-Delete-At in past')
+        self.assertEqual(body, b'X-Delete-At in past')
+
+    def test_x_delete_at_in_the_far_future(self):
+        def put(url, token, parsed, conn):
+            path = '%s/%s/%s' % (parsed.path, self.container,
+                                 'x_delete_at_in_the_far_future')
+            conn.request('PUT', path, '', {
+                'X-Auth-Token': token,
+                'Content-Length': '0',
+                'X-Delete-At': '1' * 100})
+            return check_response(conn)
+        resp = retry(put)
+        body = resp.read()
+        self.assertEqual(resp.status, 201, 'Got %s: %s' % (resp.status, body))
+
+        def head(url, token, parsed, conn):
+            path = '%s/%s/%s' % (parsed.path, self.container,
+                                 'x_delete_at_in_the_far_future')
+            conn.request('HEAD', path, '', {'X-Auth-Token': token})
+            return check_response(conn)
+        resp = retry(head)
+        body = resp.read()
+        self.assertEqual(resp.status, 200, 'Got %s: %s' % (resp.status, body))
+        self.assertEqual(resp.headers['x-delete-at'], '9' * 10)
 
     def test_copy_object(self):
         if tf.skip:
@@ -494,7 +542,7 @@ class TestObject(unittest2.TestCase):
         resp = retry(get_source)
         source_contents = resp.read()
         self.assertEqual(resp.status, 200)
-        self.assertEqual(source_contents, 'test')
+        self.assertEqual(source_contents, b'test')
 
         # copy source to dest with X-Copy-From
         def put(url, token, parsed, conn):
@@ -570,7 +618,7 @@ class TestObject(unittest2.TestCase):
         self.assertIn(resp.status, (204, 404))
 
     def test_copy_between_accounts(self):
-        if tf.skip:
+        if tf.skip2:
             raise SkipTest
 
         source = '%s/%s' % (self.container, self.obj)
@@ -585,7 +633,7 @@ class TestObject(unittest2.TestCase):
         resp = retry(get_source)
         source_contents = resp.read()
         self.assertEqual(resp.status, 200)
-        self.assertEqual(source_contents, 'test')
+        self.assertEqual(source_contents, b'test')
 
         acct = tf.parsed[0].path.split('/', 2)[2]
 
@@ -944,14 +992,16 @@ class TestObject(unittest2.TestCase):
         # can list objects
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertIn(self.obj, listing)
+        self.assertIn(self.obj, listing.split('\n'))
 
         # can get object
         resp = retry(get, self.obj, use_account=3)
         body = resp.read()
         self.assertEqual(resp.status, 200)
-        self.assertEqual(body, 'test')
+        self.assertEqual(body, b'test')
 
         # can not put an object
         obj_name = str(uuid4())
@@ -967,9 +1017,11 @@ class TestObject(unittest2.TestCase):
         # sanity with account1
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertNotIn(obj_name, listing)
-        self.assertIn(self.obj, listing)
+        self.assertNotIn(obj_name, listing.split('\n'))
+        self.assertIn(self.obj, listing.split('\n'))
 
     @requires_acls
     def test_read_write(self):
@@ -1025,14 +1077,16 @@ class TestObject(unittest2.TestCase):
         # can list objects
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertIn(self.obj, listing)
+        self.assertIn(self.obj, listing.split('\n'))
 
         # can get object
         resp = retry(get, self.obj, use_account=3)
         body = resp.read()
         self.assertEqual(resp.status, 200)
-        self.assertEqual(body, 'test')
+        self.assertEqual(body, b'test')
 
         # can put an object
         obj_name = str(uuid4())
@@ -1048,9 +1102,11 @@ class TestObject(unittest2.TestCase):
         # sanity with account1
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertIn(obj_name, listing)
-        self.assertNotIn(self.obj, listing)
+        self.assertIn(obj_name, listing.split('\n'))
+        self.assertNotIn(self.obj, listing.split('\n'))
 
     @requires_acls
     def test_admin(self):
@@ -1106,14 +1162,16 @@ class TestObject(unittest2.TestCase):
         # can list objects
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertIn(self.obj, listing)
+        self.assertIn(self.obj, listing.split('\n'))
 
         # can get object
         resp = retry(get, self.obj, use_account=3)
         body = resp.read()
         self.assertEqual(resp.status, 200)
-        self.assertEqual(body, 'test')
+        self.assertEqual(body, b'test')
 
         # can put an object
         obj_name = str(uuid4())
@@ -1129,17 +1187,19 @@ class TestObject(unittest2.TestCase):
         # sanity with account1
         resp = retry(get_listing, use_account=3)
         listing = resp.read()
+        if not six.PY2:
+            listing = listing.decode('utf8')
         self.assertEqual(resp.status, 200)
-        self.assertIn(obj_name, listing)
+        self.assertIn(obj_name, listing.split('\n'))
         self.assertNotIn(self.obj, listing)
 
     def test_manifest(self):
         if tf.skip:
             raise SkipTest
         # Data for the object segments
-        segments1 = ['one', 'two', 'three', 'four', 'five']
-        segments2 = ['six', 'seven', 'eight']
-        segments3 = ['nine', 'ten', 'eleven']
+        segments1 = [b'one', b'two', b'three', b'four', b'five']
+        segments2 = [b'six', b'seven', b'eight']
+        segments3 = [b'nine', b'ten', b'eleven']
 
         # Upload the first set of segments
         def put(url, token, parsed, conn, objnum):
@@ -1170,7 +1230,7 @@ class TestObject(unittest2.TestCase):
                 parsed.path, self.container), '', {'X-Auth-Token': token})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments1))
+        self.assertEqual(resp.read(), b''.join(segments1))
         self.assertEqual(resp.status, 200)
         self.assertEqual(resp.getheader('content-type'), 'text/jibberish')
 
@@ -1181,7 +1241,7 @@ class TestObject(unittest2.TestCase):
                     'X-Auth-Token': token, 'Range': 'bytes=3-'})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments1[1:]))
+        self.assertEqual(resp.read(), b''.join(segments1[1:]))
         self.assertEqual(resp.status, 206)
 
         # Get with a range in the middle of the second segment
@@ -1191,7 +1251,7 @@ class TestObject(unittest2.TestCase):
                     'X-Auth-Token': token, 'Range': 'bytes=5-'})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments1)[5:])
+        self.assertEqual(resp.read(), b''.join(segments1)[5:])
         self.assertEqual(resp.status, 206)
 
         # Get with a full start and stop range
@@ -1201,7 +1261,7 @@ class TestObject(unittest2.TestCase):
                     'X-Auth-Token': token, 'Range': 'bytes=5-10'})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments1)[5:11])
+        self.assertEqual(resp.read(), b''.join(segments1)[5:11])
         self.assertEqual(resp.status, 206)
 
         # Upload the second set of segments
@@ -1221,7 +1281,7 @@ class TestObject(unittest2.TestCase):
                 parsed.path, self.container), '', {'X-Auth-Token': token})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments1))
+        self.assertEqual(resp.read(), b''.join(segments1))
         self.assertEqual(resp.status, 200)
 
         # Update the manifest
@@ -1242,7 +1302,7 @@ class TestObject(unittest2.TestCase):
                 parsed.path, self.container), '', {'X-Auth-Token': token})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments2))
+        self.assertEqual(resp.read(), b''.join(segments2))
         self.assertEqual(resp.status, 200)
 
         if not tf.skip3:
@@ -1272,7 +1332,7 @@ class TestObject(unittest2.TestCase):
                     parsed.path, self.container), '', {'X-Auth-Token': token})
                 return check_response(conn)
             resp = retry(get, use_account=3)
-            self.assertEqual(resp.read(), ''.join(segments2))
+            self.assertEqual(resp.read(), b''.join(segments2))
             self.assertEqual(resp.status, 200)
 
         # Create another container for the third set of segments
@@ -1315,7 +1375,7 @@ class TestObject(unittest2.TestCase):
                 parsed.path, self.container), '', {'X-Auth-Token': token})
             return check_response(conn)
         resp = retry(get)
-        self.assertEqual(resp.read(), ''.join(segments3))
+        self.assertEqual(resp.read(), b''.join(segments3))
         self.assertEqual(resp.status, 200)
 
         if not tf.skip3:
@@ -1348,7 +1408,7 @@ class TestObject(unittest2.TestCase):
                     parsed.path, self.container), '', {'X-Auth-Token': token})
                 return check_response(conn)
             resp = retry(get, use_account=3)
-            self.assertEqual(resp.read(), ''.join(segments3))
+            self.assertEqual(resp.read(), b''.join(segments3))
             self.assertEqual(resp.status, 200)
 
         # Delete the manifest
@@ -1460,7 +1520,7 @@ class TestObject(unittest2.TestCase):
         if (tf.web_front_end == 'apache2'):
             self.assertEqual(resp.status, 404)
         else:
-            self.assertEqual(resp.read(), 'Invalid UTF8 or contains NULL')
+            self.assertEqual(resp.read(), b'Invalid UTF8 or contains NULL')
             self.assertEqual(resp.status, 412)
 
     def test_cors(self):
@@ -1625,8 +1685,102 @@ class TestObject(unittest2.TestCase):
         for c, o, body in validate_requests:
             resp = retry(get_obj, c, o)
             self.assertEqual(resp.status, 200)
+            if not six.PY2:
+                body = body.encode('utf8')
             self.assertEqual(body, resp.read())
 
+    @requires_bulk
+    def test_bulk_delete(self):
+
+        def bulk_delete(url, token, parsed, conn):
+            # try to bulk delete the object that was created during test setup
+            conn.request('DELETE', '%s/%s/%s?bulk-delete' % (
+                parsed.path, self.container, self.obj),
+                '%s/%s' % (self.container, self.obj),
+                {'X-Auth-Token': token,
+                 'Accept': 'application/xml',
+                 'Expect': '100-continue',
+                 'Content-Type': 'text/plain'})
+            return check_response(conn)
+        resp = retry(bulk_delete)
+        self.assertEqual(resp.status, 200)
+        body = resp.read()
+        tree = minidom.parseString(body)
+        self.assertEqual(tree.documentElement.tagName, 'delete')
+
+        errors = tree.getElementsByTagName('errors')
+        self.assertEqual(len(errors), 1)
+        errors = [c.data if c.nodeType == c.TEXT_NODE else c.childNodes[0].data
+                  for c in errors[0].childNodes
+                  if c.nodeType != c.TEXT_NODE or c.data.strip()]
+        self.assertEqual(errors, [])
+
+        final_status = tree.getElementsByTagName('response_status')
+        self.assertEqual(len(final_status), 1)
+        self.assertEqual(len(final_status[0].childNodes), 1)
+        self.assertEqual(final_status[0].childNodes[0].data, '200 OK')
+
+    def test_etag_quoter(self):
+        if tf.skip:
+            raise SkipTest
+        if 'etag_quoter' not in tf.cluster_info:
+            raise SkipTest("etag-quoter middleware is not enabled")
+
+        def do_head(expect_quoted=None):
+            def head(url, token, parsed, conn):
+                conn.request('HEAD', '%s/%s/%s' % (
+                    parsed.path, self.container, self.obj), '',
+                    {'X-Auth-Token': token})
+                return check_response(conn)
+
+            resp = retry(head)
+            resp.read()
+            self.assertEqual(resp.status, 200)
+
+            if expect_quoted is None:
+                expect_quoted = tf.cluster_info.get('etag_quoter', {}).get(
+                    'enable_by_default', False)
+
+            expected_etag = hashlib.md5(b'test').hexdigest()
+            if expect_quoted:
+                expected_etag = '"%s"' % expected_etag
+            self.assertEqual(resp.headers['etag'], expected_etag)
+
+        def _post(enable_flag, container_path):
+            def post(url, token, parsed, conn):
+                if container_path:
+                    path = '%s/%s' % (parsed.path, self.container)
+                    hdr = 'X-Container-Rfc-Compliant-Etags'
+                else:
+                    path = parsed.path
+                    hdr = 'X-Account-Rfc-Compliant-Etags'
+                headers = {hdr: enable_flag, 'X-Auth-Token': token}
+                conn.request('POST', path, '', headers)
+                return check_response(conn)
+
+            resp = retry(post)
+            resp.read()
+            self.assertEqual(resp.status, 204)
+
+        def post_account(enable_flag):
+            return _post(enable_flag, False)
+
+        def post_container(enable_flag):
+            return _post(enable_flag, True)
+
+        do_head()
+        post_container('t')
+        do_head(expect_quoted=True)
+        try:
+            post_account('t')
+            post_container('')
+            do_head(expect_quoted=True)
+            post_container('f')
+            do_head(expect_quoted=False)
+        finally:
+            # Don't leave a dirty account
+            post_account('')
+
 
 if __name__ == '__main__':
-    unittest2.main()
+    unittest.main()
