@@ -19,11 +19,11 @@ from __future__ import print_function
 import os
 import copy
 import logging
-from six.moves import range
-from six import BytesIO
+import logging.handlers
 import sys
 from contextlib import contextmanager, closing
 from collections import defaultdict, Iterable
+from hashlib import md5
 import itertools
 from numbers import Number
 from tempfile import NamedTemporaryFile
@@ -38,34 +38,33 @@ import json
 import random
 import errno
 import xattr
+from io import BytesIO
 
+import six
+import six.moves.cPickle as pickle
+from six.moves import range
+from six.moves.http_client import HTTPException
 
+from swift.common import storage_policy, swob, utils
+from swift.common.storage_policy import (StoragePolicy, ECStoragePolicy,
+                                         VALID_EC_TYPES)
 from swift.common.utils import Timestamp, NOTICE
 from test import get_config
-from swift.common import utils
 from swift.common.header_key_dict import HeaderKeyDict
 from swift.common.ring import Ring, RingData, RingBuilder
 from swift.obj import server
-from hashlib import md5
-import logging.handlers
 
-from six.moves.http_client import HTTPException
-from swift.common import storage_policy
-from swift.common.storage_policy import (StoragePolicy, ECStoragePolicy,
-                                         VALID_EC_TYPES)
-from swift.common import swob
 import functools
-import six.moves.cPickle as pickle
 from gzip import GzipFile
 import mock as mocklib
 import inspect
 import unittest
-import unittest2
 
 
-class SkipTest(unittest2.SkipTest, unittest.SkipTest):
+class SkipTest(unittest.SkipTest):
     pass
 
+
 EMPTY_ETAG = md5().hexdigest()
 
 # try not to import this module from swift
@@ -216,11 +215,13 @@ class PatchPolicies(object):
 class FakeRing(Ring):
 
     def __init__(self, replicas=3, max_more_nodes=0, part_power=0,
-                 base_port=1000):
+                 base_port=1000, separate_replication=False):
+        self.serialized_path = '/foo/bar/object.ring.gz'
         self._base_port = base_port
         self.max_more_nodes = max_more_nodes
         self._part_shift = 32 - part_power
         self._init_device_char()
+        self.separate_replication = separate_replication
         # 9 total nodes (6 more past the initial 3) is the cap, no matter if
         # this is set higher, or R^2 for R replicas
         self.set_replicas(replicas)
@@ -256,11 +257,16 @@ class FakeRing(Ring):
         for x in range(self.replicas):
             ip = '10.0.0.%s' % x
             port = self._base_port + x
+            if self.separate_replication:
+                repl_ip = '10.0.1.%s' % x
+                repl_port = port + 100
+            else:
+                repl_ip, repl_port = ip, port
             dev = {
                 'ip': ip,
-                'replication_ip': ip,
+                'replication_ip': repl_ip,
                 'port': port,
-                'replication_port': port,
+                'replication_port': repl_port,
                 'device': self.device_char,
                 'zone': x % 3,
                 'region': x % 2,
@@ -276,15 +282,24 @@ class FakeRing(Ring):
         return [dict(node, index=i) for i, node in enumerate(list(self._devs))]
 
     def get_more_nodes(self, part):
+        index_counter = itertools.count()
         for x in range(self.replicas, (self.replicas + self.max_more_nodes)):
-            yield {'ip': '10.0.0.%s' % x,
-                   'replication_ip': '10.0.0.%s' % x,
-                   'port': self._base_port + x,
-                   'replication_port': self._base_port + x,
+            ip = '10.0.0.%s' % x
+            port = self._base_port + x
+            if self.separate_replication:
+                repl_ip = '10.0.1.%s' % x
+                repl_port = port + 100
+            else:
+                repl_ip, repl_port = ip, port
+            yield {'ip': ip,
+                   'replication_ip': repl_ip,
+                   'port': port,
+                   'replication_port': repl_port,
                    'device': 'sda',
                    'zone': x % 3,
                    'region': x % 2,
-                   'id': x}
+                   'id': x,
+                   'handoff_index': next(index_counter)}
 
 
 def write_fake_ring(path, *devs):
@@ -293,7 +308,7 @@ def write_fake_ring(path, *devs):
     """
     dev1 = {'id': 0, 'zone': 0, 'device': 'sda1', 'ip': '127.0.0.1',
             'port': 6200}
-    dev2 = {'id': 0, 'zone': 0, 'device': 'sdb1', 'ip': '127.0.0.1',
+    dev2 = {'id': 1, 'zone': 0, 'device': 'sdb1', 'ip': '127.0.0.1',
             'port': 6200}
 
     dev1_updates, dev2_updates = devs or ({}, {})
@@ -348,6 +363,9 @@ class FabricatedRing(Ring):
         self._part_shift = 32 - part_power
         self._reload()
 
+    def has_changed(self):
+        return False
+
     def _reload(self, *args, **kwargs):
         self._rtime = time.time() * 2
         if hasattr(self, '_replica2part2dev_id'):
@@ -372,6 +390,7 @@ class FabricatedRing(Ring):
         for p in range(2 ** self.part_power):
             for r in range(self.replicas):
                 self._replica2part2dev_id[r][p] = next(dev_ids)
+        self._update_bookkeeping()
 
 
 class FakeMemcache(object):
@@ -406,17 +425,17 @@ class FakeMemcache(object):
 
 
 def readuntil2crlfs(fd):
-    rv = ''
-    lc = ''
+    rv = b''
+    lc = b''
     crlfs = 0
     while crlfs < 2:
         c = fd.read(1)
         if not c:
             raise ValueError("didn't get two CRLFs; just got %r" % rv)
         rv = rv + c
-        if c == '\r' and lc != '\n':
+        if c == b'\r' and lc != b'\n':
             crlfs = 0
-        if lc == '\r' and c == '\n':
+        if lc == b'\r' and c == b'\n':
             crlfs += 1
         lc = c
     return rv
@@ -637,6 +656,9 @@ class FakeLogger(logging.Logger, object):
     def handleError(self, record):
         pass
 
+    def isEnabledFor(self, level):
+        return True
+
 
 class DebugSwiftLogFormatter(utils.SwiftLogFormatter):
 
@@ -652,10 +674,13 @@ class DebugLogger(FakeLogger):
         FakeLogger.__init__(self, *args, **kwargs)
         self.formatter = DebugSwiftLogFormatter(
             "%(server)s %(levelname)s: %(message)s")
+        self.records = defaultdict(list)
 
     def handle(self, record):
         self._handle(record)
-        print(self.formatter.format(record))
+        formatted = self.formatter.format(record)
+        print(formatted)
+        self.records[record.levelname].append(formatted)
 
 
 class DebugLogAdapter(utils.LogAdapter):
@@ -867,7 +892,10 @@ def fake_http_connect(*code_iter, **kwargs):
 
     class FakeConn(object):
 
-        def __init__(self, status, etag=None, body='', timestamp='1',
+        SLOW_READS = 4
+        SLOW_WRITES = 4
+
+        def __init__(self, status, etag=None, body=b'', timestamp=-1,
                      headers=None, expect_headers=None, connection_id=None,
                      give_send=None, give_expect=None):
             if not isinstance(status, FakeStatus):
@@ -882,7 +910,15 @@ def fake_http_connect(*code_iter, **kwargs):
             self.body = body
             self.headers = headers or {}
             self.expect_headers = expect_headers or {}
-            self.timestamp = timestamp
+            if timestamp == -1:
+                # -1 is reserved to mean "magic default"
+                if status.status != 404:
+                    self.timestamp = '1'
+                else:
+                    self.timestamp = '0'
+            else:
+                # tests may specify int, string, Timestamp or None
+                self.timestamp = timestamp
             self.connection_id = connection_id
             self.give_send = give_send
             self.give_expect = give_expect
@@ -892,6 +928,12 @@ def fake_http_connect(*code_iter, **kwargs):
                     self._next_sleep = kwargs['slow'].pop(0)
                 except IndexError:
                     self._next_sleep = None
+
+            # if we're going to be slow, we need a body to send slowly
+            am_slow, _junk = self.get_slow()
+            if am_slow and len(self.body) < self.SLOW_READS:
+                self.body += b" " * (self.SLOW_READS - len(self.body))
+
             # be nice to trixy bits with node_iter's
             eventlet.sleep()
 
@@ -922,11 +964,12 @@ def fake_http_connect(*code_iter, **kwargs):
         def getheaders(self):
             etag = self.etag
             if not etag:
-                if isinstance(self.body, str):
+                if isinstance(self.body, bytes):
                     etag = '"' + md5(self.body).hexdigest() + '"'
                 else:
                     etag = '"68b329da9893e34099c7d8ad5cb9c940"'
 
+            am_slow, _junk = self.get_slow()
             headers = HeaderKeyDict({
                 'content-length': len(self.body),
                 'content-type': 'x-application/test',
@@ -949,9 +992,6 @@ def fake_http_connect(*code_iter, **kwargs):
                     headers['x-container-timestamp'] = '1'
             except StopIteration:
                 pass
-            am_slow, value = self.get_slow()
-            if am_slow:
-                headers['content-length'] = '4'
             headers.update(self.headers)
             return headers.items()
 
@@ -968,12 +1008,16 @@ def fake_http_connect(*code_iter, **kwargs):
         def read(self, amt=None):
             am_slow, value = self.get_slow()
             if am_slow:
-                if self.sent < 4:
+                if self.sent < self.SLOW_READS:
+                    slowly_read_byte = self.body[self.sent:self.sent + 1]
                     self.sent += 1
                     eventlet.sleep(value)
-                    return ' '
-            rv = self.body[:amt]
-            self.body = self.body[amt:]
+                    return slowly_read_byte
+            if amt is None:
+                rv = self.body[self.sent:]
+            else:
+                rv = self.body[self.sent:self.sent + amt]
+            self.sent += len(rv)
             return rv
 
         def send(self, data=None):
@@ -981,17 +1025,22 @@ def fake_http_connect(*code_iter, **kwargs):
                 self.give_send(self, data)
             am_slow, value = self.get_slow()
             if am_slow:
-                if self.received < 4:
+                if self.received < self.SLOW_WRITES:
                     self.received += 1
                     eventlet.sleep(value)
 
         def getheader(self, name, default=None):
             return HeaderKeyDict(self.getheaders()).get(name, default)
 
+        def nuke_from_orbit(self):
+            # wrapped connections from buffered_http have this helper
+            self.close()
+
         def close(self):
             self.closed = True
 
-    timestamps_iter = iter(kwargs.get('timestamps') or ['1'] * len(code_iter))
+    # unless tests provide timestamps we use the "magic default"
+    timestamps_iter = iter(kwargs.get('timestamps') or [-1] * len(code_iter))
     etag_iter = iter(kwargs.get('etags') or [None] * len(code_iter))
     if isinstance(kwargs.get('headers'), (list, tuple)):
         headers_iter = iter(kwargs['headers'])
@@ -1043,16 +1092,19 @@ def fake_http_connect(*code_iter, **kwargs):
         expect_headers = next(expect_headers_iter)
         timestamp = next(timestamps_iter)
 
-        if status <= 0:
+        if isinstance(status, int) and status <= 0:
             raise HTTPException()
         if body_iter is None:
-            body = static_body or ''
+            body = static_body or b''
         else:
             body = next(body_iter)
-        return FakeConn(status, etag, body=body, timestamp=timestamp,
+        conn = FakeConn(status, etag, body=body, timestamp=timestamp,
                         headers=headers, expect_headers=expect_headers,
                         connection_id=i, give_send=kwargs.get('give_send'),
                         give_expect=kwargs.get('give_expect'))
+        if 'capture_connections' in kwargs:
+            kwargs['capture_connections'].append(conn)
+        return conn
 
     connect.unexpected_requests = unexpected_requests
     connect.code_iter = code_iter
@@ -1063,8 +1115,11 @@ def fake_http_connect(*code_iter, **kwargs):
 @contextmanager
 def mocked_http_conn(*args, **kwargs):
     requests = []
+    responses = []
 
     def capture_requests(ip, port, method, path, headers, qs, ssl):
+        if six.PY2 and not isinstance(ip, bytes):
+            ip = ip.encode('ascii')
         req = {
             'ip': ip,
             'port': port,
@@ -1076,8 +1131,10 @@ def mocked_http_conn(*args, **kwargs):
         }
         requests.append(req)
     kwargs.setdefault('give_connect', capture_requests)
+    kwargs['capture_connections'] = responses
     fake_conn = fake_http_connect(*args, **kwargs)
     fake_conn.requests = requests
+    fake_conn.responses = responses
     with mocklib.patch('swift.common.bufferedhttp.http_connect_raw',
                        new=fake_conn):
         yield fake_conn
@@ -1129,18 +1186,9 @@ def requires_o_tmpfile_support_in_tmp(func):
     return wrapper
 
 
-def requires_o_tmpfile_support(func):
-    @functools.wraps(func)
-    def wrapper(*args, **kwargs):
-        if not utils.o_tmpfile_supported():
-            raise SkipTest('Requires O_TMPFILE support')
-        return func(*args, **kwargs)
-    return wrapper
-
-
 class StubResponse(object):
 
-    def __init__(self, status, body='', headers=None, frag_index=None):
+    def __init__(self, status, body=b'', headers=None, frag_index=None):
         self.status = status
         self.body = body
         self.readable = BytesIO(body)
@@ -1150,6 +1198,10 @@ class StubResponse(object):
         fake_reason = ('Fake', 'This response is a lie.')
         self.reason = swob.RESPONSE_REASONS.get(status, fake_reason)[0]
 
+    def nuke_from_orbit(self):
+        if hasattr(self, 'swift_conn'):
+            self.swift_conn.close()
+
     def getheader(self, header_name, default=None):
         return self.headers.get(header_name, default)
 
@@ -1161,6 +1213,14 @@ class StubResponse(object):
     def read(self, amt=0):
         return self.readable.read(amt)
 
+    def __repr__(self):
+        info = ['Status: %s' % self.status]
+        if self.headers:
+            info.append('Headers: %r' % dict(self.headers))
+        if self.body:
+            info.append('Body: %r' % self.body)
+        return '<StubResponse %s>' % ', '.join(info)
+
 
 def encode_frag_archive_bodies(policy, body):
     """
@@ -1187,7 +1247,7 @@ def encode_frag_archive_bodies(policy, body):
         fragment_payloads.append(fragments)
 
     # join up the fragment payloads per node
-    ec_archive_bodies = [''.join(frags)
+    ec_archive_bodies = [b''.join(frags)
                          for frags in zip(*fragment_payloads)]
     return ec_archive_bodies
 
@@ -1195,7 +1255,7 @@ def encode_frag_archive_bodies(policy, body):
 def make_ec_object_stub(test_body, policy, timestamp):
     segment_size = policy.ec_segment_size
     test_body = test_body or (
-        'test' * segment_size)[:-random.randint(1, 1000)]
+        b'test' * segment_size)[:-random.randint(1, 1000)]
     timestamp = timestamp or utils.Timestamp.now()
     etag = md5(test_body).hexdigest()
     ec_archive_bodies = encode_frag_archive_bodies(policy, test_body)
@@ -1235,12 +1295,11 @@ def fake_ec_node_response(node_frags, policy):
     call_count = {}  # maps node index to get_response call count for node
 
     def _build_node_map(req, policy):
-        node_key = lambda n: (n['ip'], n['port'])
         part = utils.split_path(req['path'], 5, 5, True)[1]
         all_nodes.extend(policy.object_ring.get_part_nodes(part))
         all_nodes.extend(policy.object_ring.get_more_nodes(part))
         for i, node in enumerate(all_nodes):
-            node_map[node_key(node)] = i
+            node_map[(node['ip'], node['port'])] = i
             call_count[i] = 0
 
     # normalize node_frags to a list of fragments for each node even
@@ -1346,7 +1405,8 @@ def xattr_supported_check():
         fd, tmppath = mkstemp()
         xattr.setxattr(fd, 'user.swift.testing_key', big_val)
     except IOError as e:
-        if errno.errorcode.get(e.errno) in ('ENOSPC', 'ENOTSUP', 'EOPNOTSUPP'):
+        if errno.errorcode.get(e.errno) in ('ENOSPC', 'ENOTSUP', 'EOPNOTSUPP',
+                                            'ERANGE'):
             # filesystem does not support xattr of this size
             return False
         raise
@@ -1415,3 +1475,13 @@ def attach_fake_replication_rpc(rpc, replicate_hook=None, errors=None):
             return resp
 
     return FakeReplConnection
+
+
+def group_by_byte(contents):
+    # This looks a little funny, but iterating through a byte string on py3
+    # yields a sequence of ints, not a sequence of single-byte byte strings
+    # as it did on py2.
+    byte_iter = (contents[i:i + 1] for i in range(len(contents)))
+    return [
+        (char, sum(1 for _ in grp))
+        for char, grp in itertools.groupby(byte_iter)]
