@@ -20,53 +20,18 @@ import shutil
 import tempfile
 import unittest
 
-from logging import DEBUG
 from mock import patch, call, DEFAULT
-import six
 import eventlet
 
 from swift.account import reaper
 from swift.account.backend import DATADIR
 from swift.common.exceptions import ClientException
-from swift.common.utils import normalize_timestamp
+from swift.common.utils import normalize_timestamp, Timestamp
 
 from test import unit
 from swift.common.storage_policy import StoragePolicy, POLICIES
 
 
-class FakeLogger(object):
-    def __init__(self, *args, **kwargs):
-        self.inc = {'return_codes.4': 0,
-                    'return_codes.2': 0,
-                    'objects_failures': 0,
-                    'objects_deleted': 0,
-                    'objects_remaining': 0,
-                    'objects_possibly_remaining': 0,
-                    'containers_failures': 0,
-                    'containers_deleted': 0,
-                    'containers_remaining': 0,
-                    'containers_possibly_remaining': 0}
-        self.exp = []
-
-    def info(self, msg, *args):
-        self.msg = msg
-
-    def error(self, msg, *args):
-        self.msg = msg
-
-    def timing_since(*args, **kwargs):
-        pass
-
-    def getEffectiveLevel(self):
-        return DEBUG
-
-    def exception(self, *args):
-        self.exp.append(args)
-
-    def increment(self, key):
-        self.inc[key] += 1
-
-
 class FakeBroker(object):
     def __init__(self):
         self.info = {}
@@ -76,7 +41,7 @@ class FakeBroker(object):
 
 
 class FakeAccountBroker(object):
-    def __init__(self, containers):
+    def __init__(self, containers, logger):
         self.containers = containers
         self.containers_yielded = []
 
@@ -85,9 +50,18 @@ class FakeAccountBroker(object):
                 'delete_timestamp': time.time() - 10}
         return info
 
-    def list_containers_iter(self, *args):
+    def list_containers_iter(self, limit, marker, *args, **kwargs):
+        if not kwargs.pop('allow_reserved'):
+            raise RuntimeError('Expected allow_reserved to be True!')
+        if kwargs:
+            raise RuntimeError('Got unexpected keyword arguments: %r' % (
+                kwargs, ))
         for cont in self.containers:
-            yield cont, None, None, None
+            if cont > marker:
+                yield cont, None, None, None, None
+            limit -= 1
+            if limit <= 0:
+                break
 
     def is_status_deleted(self):
         return True
@@ -196,11 +170,11 @@ class TestReaper(unittest.TestCase):
             raise self.myexp
         if self.timeout:
             raise eventlet.Timeout()
-        objects = [{'name': 'o1'},
-                   {'name': 'o2'},
-                   {'name': six.text_type('o3')},
-                   {'name': ''}]
-        return None, objects
+        objects = [{'name': u'o1'},
+                   {'name': u'o2'},
+                   {'name': u'o3'},
+                   {'name': u'o4'}]
+        return None, [o for o in objects if o['name'] > kwargs['marker']]
 
     def fake_container_ring(self):
         return FakeRing()
@@ -333,11 +307,11 @@ class TestReaper(unittest.TestCase):
         for policy in POLICIES:
             r.reset_stats()
             with patch(mock_path) as fake_direct_delete:
-                with patch('swift.account.reaper.time') as mock_time:
-                    mock_time.return_value = 1429117638.86767
+                with patch('swift.common.utils.Timestamp.now') as mock_now:
+                    mock_now.return_value = Timestamp(1429117638.86767)
                     r.reap_object('a', 'c', 'partition', cont_nodes, 'o',
                                   policy.idx)
-                    mock_time.assert_called_once_with()
+                    mock_now.assert_called_once_with()
                     for i, call_args in enumerate(
                             fake_direct_delete.call_args_list):
                         cnode = cont_nodes[i % len(cont_nodes)]
@@ -348,7 +322,8 @@ class TestReaper(unittest.TestCase):
                             'X-Container-Partition': 'partition',
                             'X-Container-Device': device,
                             'X-Backend-Storage-Policy-Index': policy.idx,
-                            'X-Timestamp': '1429117638.86767'
+                            'X-Timestamp': '1429117638.86767',
+                            'x-backend-use-replication-network': 'true',
                         }
                         ring = r.get_object_ring(policy.idx)
                         expected = call(dict(ring.devs[i], index=i), 0,
@@ -439,8 +414,9 @@ class TestReaper(unittest.TestCase):
                 return headers, obj_list
 
             mocks['direct_get_container'].side_effect = fake_get_container
-            with patch('swift.account.reaper.time') as mock_time:
-                mock_time.side_effect = [1429117638.86767, 1429117639.67676]
+            with patch('swift.common.utils.Timestamp.now') as mock_now:
+                mock_now.side_effect = [Timestamp(1429117638.86767),
+                                        Timestamp(1429117639.67676)]
                 r.reap_container('a', 'partition', acc_nodes, 'c')
 
             # verify calls to direct_delete_object
@@ -467,7 +443,8 @@ class TestReaper(unittest.TestCase):
                     'X-Account-Partition': 'partition',
                     'X-Account-Device': device,
                     'X-Account-Override-Deleted': 'yes',
-                    'X-Timestamp': '1429117639.67676'
+                    'X-Timestamp': '1429117639.67676',
+                    'x-backend-use-replication-network': 'true',
                 }
                 ring = r.get_object_ring(policy.idx)
                 expected = call(dict(ring.devs[i], index=i), 0, 'a', 'c',
@@ -588,8 +565,8 @@ class TestReaper(unittest.TestCase):
             self.r.stats_return_codes.get(2, 0) + 1
 
     def test_reap_account(self):
-        containers = ('c1', 'c2', 'c3', '')
-        broker = FakeAccountBroker(containers)
+        containers = ('c1', 'c2', 'c3', 'c4')
+        broker = FakeAccountBroker(containers, unit.FakeLogger())
         self.called_amount = 0
         self.r = r = self.init_reaper({}, fakelogger=True)
         r.start_time = time.time()
@@ -625,7 +602,7 @@ class TestReaper(unittest.TestCase):
         self.assertEqual(len(self.r.account_ring.devs), 3)
 
     def test_reap_account_no_container(self):
-        broker = FakeAccountBroker(tuple())
+        broker = FakeAccountBroker(tuple(), unit.FakeLogger())
         self.r = r = self.init_reaper({}, fakelogger=True)
         self.called_amount = 0
         r.start_time = time.time()
@@ -740,16 +717,21 @@ class TestReaper(unittest.TestCase):
         devices = self.prepare_data_dir()
         self.called_amount = 0
         conf = {'devices': devices}
-        r = self.init_reaper(conf, myips=['10.10.10.2'])
+        r = self.init_reaper(conf, myips=['10.10.10.2'], fakelogger=True)
 
         container_reaped = [0]
 
-        def fake_list_containers_iter(self, *args):
+        def fake_list_containers_iter(self, *args, **kwargs):
+            if not kwargs.pop('allow_reserved'):
+                raise RuntimeError('Expected allow_reserved to be True!')
+            if kwargs:
+                raise RuntimeError('Got unexpected keyword arguments: %r' % (
+                    kwargs, ))
             for container in self.containers:
                 if container in self.containers_yielded:
                     continue
 
-                yield container, None, None, None
+                yield container, None, None, None, None
                 self.containers_yielded.append(container)
 
         def fake_reap_container(self, account, account_partition,
@@ -757,6 +739,7 @@ class TestReaper(unittest.TestCase):
             container_reaped[0] += 1
 
         fake_ring = FakeRing()
+        fake_logger = unit.FakeLogger()
         with patch('swift.account.reaper.AccountBroker',
                    FakeAccountBroker), \
                 patch(
@@ -765,27 +748,32 @@ class TestReaper(unittest.TestCase):
                 patch('swift.account.reaper.AccountReaper.reap_container',
                       fake_reap_container):
 
-            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'])
+            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'],
+                                            fake_logger)
             r.reap_account(fake_broker, 10, fake_ring.nodes, 0)
             self.assertEqual(container_reaped[0], 0)
 
-            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'])
+            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'],
+                                            fake_logger)
             container_reaped[0] = 0
             r.reap_account(fake_broker, 10, fake_ring.nodes, 1)
             self.assertEqual(container_reaped[0], 1)
 
             container_reaped[0] = 0
-            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'])
+            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'],
+                                            fake_logger)
             r.reap_account(fake_broker, 10, fake_ring.nodes, 2)
             self.assertEqual(container_reaped[0], 0)
 
             container_reaped[0] = 0
-            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'])
+            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'],
+                                            fake_logger)
             r.reap_account(fake_broker, 10, fake_ring.nodes, 3)
             self.assertEqual(container_reaped[0], 3)
 
             container_reaped[0] = 0
-            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'])
+            fake_broker = FakeAccountBroker(['c', 'd', 'e', 'f', 'g'],
+                                            fake_logger)
             r.reap_account(fake_broker, 10, fake_ring.nodes, 4)
             self.assertEqual(container_reaped[0], 1)
 
@@ -805,16 +793,14 @@ class TestReaper(unittest.TestCase):
         devices = prepare_data_dir()
         r = init_reaper(devices)
 
-        with patch('swift.account.reaper.ismount', lambda x: True):
-            with patch(
-                    'swift.account.reaper.AccountReaper.reap_device') as foo:
-                r.run_once()
+        with patch('swift.account.reaper.AccountReaper.reap_device') as foo, \
+                unit.mock_check_drive(ismount=True):
+            r.run_once()
         self.assertEqual(foo.called, 1)
 
-        with patch('swift.account.reaper.ismount', lambda x: False):
-            with patch(
-                    'swift.account.reaper.AccountReaper.reap_device') as foo:
-                r.run_once()
+        with patch('swift.account.reaper.AccountReaper.reap_device') as foo, \
+                unit.mock_check_drive(ismount=False):
+            r.run_once()
         self.assertFalse(foo.called)
 
         with patch('swift.account.reaper.AccountReaper.reap_device') as foo:
@@ -842,12 +828,10 @@ class TestReaper(unittest.TestCase):
         r = init_reaper()
         with patch('swift.account.reaper.sleep', fake_sleep):
             with patch('swift.account.reaper.random.random', fake_random):
-                try:
+                with self.assertRaises(Exception) as raised:
                     r.run_forever()
-                except Exception as err:
-                    pass
         self.assertEqual(self.val, 1)
-        self.assertEqual(str(err), 'exit')
+        self.assertEqual(str(raised.exception), 'exit')
 
 
 if __name__ == '__main__':
