@@ -13,9 +13,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from datetime import datetime, timedelta
+import hashlib
 from mock import patch, MagicMock
 import unittest
 
+from io import BytesIO
+
 from swift.common import swob
 from swift.common.swob import Request, HTTPNoContent
 from swift.common.middleware.s3api.utils import mktime
@@ -24,7 +28,7 @@ from swift.common.middleware.s3api.subresource import ACL, User, Owner, \
     Grant, encode_acl
 from test.unit.common.middleware.s3api.test_s3api import S3ApiTestCase
 from swift.common.middleware.s3api.s3request import S3Request, \
-    S3AclRequest, SigV4Request, SIGV4_X_AMZ_DATE_FORMAT
+    S3AclRequest, SigV4Request, SIGV4_X_AMZ_DATE_FORMAT, HashingInput
 from swift.common.middleware.s3api.s3response import InvalidArgument, \
     NoSuchBucket, InternalError, \
     AccessDenied, SignatureDoesNotMatch, RequestTimeTooSkewed
@@ -140,7 +144,7 @@ class TestRequest(S3ApiTestCase):
     def test_get_response_without_match_ACL_MAP(self):
         with self.assertRaises(Exception) as e:
             self._test_get_response('POST', req_klass=S3AclRequest)
-        self.assertEqual(e.exception.message,
+        self.assertEqual(e.exception.args[0],
                          'No permission to be checked exists')
 
     def test_get_response_without_duplication_HEAD_request(self):
@@ -184,7 +188,7 @@ class TestRequest(S3ApiTestCase):
         self.assertEqual(mock_get_resp.call_count, 2)
         args, kargs = mock_get_resp.call_args_list[0]
         get_resp_obj = args[3]
-        self.assertTrue(get_resp_obj is '')
+        self.assertEqual(get_resp_obj, '')
         self.assertEqual(m_check_permission.call_count, 1)
         args, kargs = m_check_permission.call_args
         permission = args[1]
@@ -212,8 +216,8 @@ class TestRequest(S3ApiTestCase):
         s3req = create_s3request_with_param('max-keys', '1' * 30)
         with self.assertRaises(InvalidArgument) as result:
             s3req.get_validated_param('max-keys', 1)
-        self.assertTrue(
-            'not an integer or within integer range' in result.exception.body)
+        self.assertIn(
+            b'not an integer or within integer range', result.exception.body)
         self.assertEqual(
             result.exception.headers['content-type'], 'application/xml')
 
@@ -221,8 +225,8 @@ class TestRequest(S3ApiTestCase):
         s3req = create_s3request_with_param('max-keys', '-1')
         with self.assertRaises(InvalidArgument) as result:
             s3req.get_validated_param('max-keys', 1)
-        self.assertTrue(
-            'must be an integer between 0 and' in result.exception.body)
+        self.assertIn(
+            b'must be an integer between 0 and', result.exception.body)
         self.assertEqual(
             result.exception.headers['content-type'], 'application/xml')
 
@@ -230,8 +234,8 @@ class TestRequest(S3ApiTestCase):
         s3req = create_s3request_with_param('max-keys', 'invalid')
         with self.assertRaises(InvalidArgument) as result:
             s3req.get_validated_param('max-keys', 1)
-        self.assertTrue(
-            'not an integer or within integer range' in result.exception.body)
+        self.assertIn(
+            b'not an integer or within integer range', result.exception.body)
         self.assertEqual(
             result.exception.headers['content-type'], 'application/xml')
 
@@ -246,11 +250,11 @@ class TestRequest(S3ApiTestCase):
             m_swift_resp.return_value = FakeSwiftResponse()
             s3_req = S3AclRequest(req.environ, MagicMock())
             self.assertNotIn('s3api.auth_details', s3_req.environ)
-            self.assertNotIn('HTTP_AUTHORIZATION', s3_req.environ)
-            self.assertNotIn('Authorization', s3_req.headers)
-            self.assertEqual(s3_req.token, 'token')
 
     def test_to_swift_req_Authorization_not_exist_in_swreq(self):
+        # the difference from
+        # test_authenticate_delete_Authorization_from_s3req_headers above is
+        # this method asserts *to_swift_req* method.
         container = 'bucket'
         obj = 'obj'
         method = 'GET'
@@ -263,11 +267,12 @@ class TestRequest(S3ApiTestCase):
 
             m_swift_resp.return_value = FakeSwiftResponse()
             s3_req = S3AclRequest(req.environ, MagicMock())
+            # Yes, we *want* to assert this
             sw_req = s3_req.to_swift_req(method, container, obj)
+            # So since the result of S3AclRequest init tests and with this
+            # result to_swift_req doesn't add Authorization header and token
             self.assertNotIn('s3api.auth_details', sw_req.environ)
-            self.assertNotIn('HTTP_AUTHORIZATION', sw_req.environ)
-            self.assertNotIn('Authorization', sw_req.headers)
-            self.assertEqual(sw_req.headers['X-Auth-Token'], 'token')
+            self.assertNotIn('X-Auth-Token', sw_req.headers)
 
     def test_to_swift_req_subrequest_proxy_access_log(self):
         container = 'bucket'
@@ -303,9 +308,14 @@ class TestRequest(S3ApiTestCase):
             self.assertTrue(sw_req.environ['swift.proxy_access_log_made'])
 
     def test_get_container_info(self):
+        s3api_acl = '{"Owner":"owner","Grant":'\
+            '[{"Grantee":"owner","Permission":"FULL_CONTROL"}]}'
         self.swift.register('HEAD', '/v1/AUTH_test/bucket', HTTPNoContent,
                             {'x-container-read': 'foo',
-                             'X-container-object-count': 5,
+                             'X-container-object-count': '5',
+                             'x-container-sysmeta-versions-location':
+                                'bucket2',
+                             'x-container-sysmeta-s3api-acl': s3api_acl,
                              'X-container-meta-foo': 'bar'}, None)
         req = Request.blank('/bucket', environ={'REQUEST_METHOD': 'GET'},
                             headers={'Authorization': 'AWS test:tester:hmac',
@@ -316,13 +326,16 @@ class TestRequest(S3ApiTestCase):
         self.assertTrue('status' in info)  # sanity
         self.assertEqual(204, info['status'])  # sanity
         self.assertEqual('foo', info['read_acl'])  # sanity
-        self.assertEqual('5', info['object_count'])  # sanity
+        self.assertEqual(5, info['object_count'])  # sanity
+        self.assertEqual(
+            'bucket2', info['sysmeta']['versions-location'])  # sanity
+        self.assertEqual(s3api_acl, info['sysmeta']['s3api-acl'])  # sanity
         self.assertEqual({'foo': 'bar'}, info['meta'])  # sanity
         with patch(
                 'swift.common.middleware.s3api.s3request.get_container_info',
                 return_value={'status': 204}) as mock_info:
             # Then all calls goes to get_container_info
-            for x in xrange(10):
+            for x in range(10):
                 info = s3_req.get_container_info(self.swift)
                 self.assertTrue('status' in info)  # sanity
                 self.assertEqual(204, info['status'])  # sanity
@@ -344,7 +357,7 @@ class TestRequest(S3ApiTestCase):
                             headers={'Authorization': 'AWS test:tester:hmac'})
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '403')
-        self.assertEqual(body, '')
+        self.assertEqual(body, b'')
 
     def test_date_header_expired(self):
         self.swift.register('HEAD', '/v1/AUTH_test/nojunk', swob.HTTPNotFound,
@@ -356,7 +369,7 @@ class TestRequest(S3ApiTestCase):
 
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '403')
-        self.assertEqual(body, '')
+        self.assertEqual(body, b'')
 
     def test_date_header_with_x_amz_date_valid(self):
         self.swift.register('HEAD', '/v1/AUTH_test/nojunk', swob.HTTPNotFound,
@@ -369,7 +382,7 @@ class TestRequest(S3ApiTestCase):
 
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '404')
-        self.assertEqual(body, '')
+        self.assertEqual(body, b'')
 
     def test_date_header_with_x_amz_date_expired(self):
         self.swift.register('HEAD', '/v1/AUTH_test/nojunk', swob.HTTPNotFound,
@@ -383,7 +396,7 @@ class TestRequest(S3ApiTestCase):
 
         status, headers, body = self.call_s3api(req)
         self.assertEqual(status.split()[0], '403')
-        self.assertEqual(body, '')
+        self.assertEqual(body, b'')
 
     def _test_request_timestamp_sigv4(self, date_header):
         # signature v4 here
@@ -400,9 +413,11 @@ class TestRequest(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request, '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=%s,'
-                'Signature=X' % ';'.join(sorted(['host', included_header])),
+                'Signature=X' % (
+                    self.get_v4_amz_date_header().split('T', 1)[0],
+                    ';'.join(sorted(['host', included_header]))),
             'X-Amz-Content-SHA256': '0123456789'}
 
         headers.update(date_header)
@@ -419,7 +434,7 @@ class TestRequest(S3ApiTestCase):
 
     def test_request_timestamp_sigv4(self):
         access_denied_message = \
-            'AWS authentication requires a valid Date or x-amz-date header'
+            b'AWS authentication requires a valid Date or x-amz-date header'
 
         # normal X-Amz-Date header
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header()}
@@ -434,7 +449,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv4(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # mangled Date header
@@ -442,7 +457,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv4(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # Negative timestamp
@@ -450,7 +465,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv4(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # far-past Date header
@@ -458,17 +473,24 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv4(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
+        # near-future X-Amz-Date header
+        date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
+            datetime.utcnow() + timedelta(days=1)
+        )}
+        with self.assertRaises(RequestTimeTooSkewed) as cm:
+            self._test_request_timestamp_sigv4(date_header)
+
         # far-future Date header
         date_header = {'Date': 'Tue, 07 Jul 9999 21:53:04 GMT'}
         with self.assertRaises(RequestTimeTooSkewed) as cm:
             self._test_request_timestamp_sigv4(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
-        self.assertIn('The difference between the request time and the '
-                      'current time is too large.', cm.exception.body)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
+        self.assertIn(b'The difference between the request time and the '
+                      b'current time is too large.', cm.exception.body)
 
     def _test_request_timestamp_sigv2(self, date_header):
         # signature v4 here
@@ -490,7 +512,7 @@ class TestRequest(S3ApiTestCase):
 
     def test_request_timestamp_sigv2(self):
         access_denied_message = \
-            'AWS authentication requires a valid Date or x-amz-date header'
+            b'AWS authentication requires a valid Date or x-amz-date header'
 
         # In v2 format, normal X-Amz-Date header is same
         date_header = {'X-Amz-Date': self.get_date_header()}
@@ -505,7 +527,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv2(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # mangled Date header
@@ -513,7 +535,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv2(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # Negative timestamp
@@ -521,7 +543,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv2(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # far-past Date header
@@ -529,7 +551,7 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(AccessDenied) as cm:
             self._test_request_timestamp_sigv2(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
         self.assertIn(access_denied_message, cm.exception.body)
 
         # far-future Date header
@@ -537,9 +559,9 @@ class TestRequest(S3ApiTestCase):
         with self.assertRaises(RequestTimeTooSkewed) as cm:
             self._test_request_timestamp_sigv2(date_header)
 
-        self.assertEqual('403 Forbidden', cm.exception.message)
-        self.assertIn('The difference between the request time and the '
-                      'current time is too large.', cm.exception.body)
+        self.assertEqual('403 Forbidden', cm.exception.args[0])
+        self.assertIn(b'The difference between the request time and the '
+                      b'current time is too large.', cm.exception.body)
 
     def test_headers_to_sign_sigv4(self):
         environ = {
@@ -550,9 +572,9 @@ class TestRequest(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request, '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=X',
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
             'X-Amz-Content-SHA256': '0123456789',
             'Date': self.get_date_header(),
             'X-Amz-Date': x_amz_date}
@@ -570,9 +592,9 @@ class TestRequest(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request, '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256,'
-                'Signature=X',
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
             'X-Amz-Content-SHA256': '0123456789',
             'Date': self.get_date_header()}
 
@@ -589,9 +611,9 @@ class TestRequest(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request, '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=X',
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
             'X-Amz-Content-SHA256': '0123456789',
             'Date': self.get_date_header()}
 
@@ -653,9 +675,9 @@ class TestRequest(S3ApiTestCase):
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
-                'Credential=test/20130524/US/s3/aws4_request, '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=X',
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
             'X-Amz-Content-SHA256': '0123456789',
             'Date': self.get_date_header(),
             'X-Amz-Date': x_amz_date}
@@ -666,14 +688,14 @@ class TestRequest(S3ApiTestCase):
         sigv4_req = SigV4Request(req.environ)
         uri = sigv4_req._canonical_uri()
 
-        self.assertEqual(uri, '/')
+        self.assertEqual(uri, b'/')
         self.assertEqual(req.environ['PATH_INFO'], '/')
 
         req = Request.blank('/obj1', environ=environ, headers=headers)
         sigv4_req = SigV4Request(req.environ)
         uri = sigv4_req._canonical_uri()
 
-        self.assertEqual(uri, '/obj1')
+        self.assertEqual(uri, b'/obj1')
         self.assertEqual(req.environ['PATH_INFO'], '/obj1')
 
         environ = {
@@ -686,7 +708,7 @@ class TestRequest(S3ApiTestCase):
         sigv4_req = SigV4Request(req.environ)
         uri = sigv4_req._canonical_uri()
 
-        self.assertEqual(uri, '/')
+        self.assertEqual(uri, b'/')
         self.assertEqual(req.environ['PATH_INFO'], '/')
 
         req = Request.blank('/bucket/obj1',
@@ -695,11 +717,11 @@ class TestRequest(S3ApiTestCase):
         sigv4_req = SigV4Request(req.environ)
         uri = sigv4_req._canonical_uri()
 
-        self.assertEqual(uri, '/bucket/obj1')
+        self.assertEqual(uri, b'/bucket/obj1')
         self.assertEqual(req.environ['PATH_INFO'], '/bucket/obj1')
 
-    @patch.object(S3Request, '_validate_headers', lambda *a: None)
-    def test_check_signature_sigv2(self):
+    @patch.object(S3Request, '_validate_dates', lambda *a: None)
+    def _test_check_signature_sigv2(self, secret):
         # See https://web.archive.org/web/20151226025049/http://
         # docs.aws.amazon.com//AmazonS3/latest/dev/RESTAuthentication.html
         req = Request.blank('/photos/puppy.jpg', headers={
@@ -709,16 +731,15 @@ class TestRequest(S3ApiTestCase):
                               'bWq2s1WEIj+Ydj0vQ697zp+IXMU='),
         })
         sigv2_req = S3Request(req.environ, storage_domain='s3.amazonaws.com')
-        expected_sts = '\n'.join([
-            'GET',
-            '',
-            '',
-            'Tue, 27 Mar 2007 19:36:42 +0000',
-            '/johnsmith/photos/puppy.jpg',
+        expected_sts = b'\n'.join([
+            b'GET',
+            b'',
+            b'',
+            b'Tue, 27 Mar 2007 19:36:42 +0000',
+            b'/johnsmith/photos/puppy.jpg',
         ])
         self.assertEqual(expected_sts, sigv2_req._string_to_sign())
-        self.assertTrue(sigv2_req.check_signature(
-            'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'))
+        self.assertTrue(sigv2_req.check_signature(secret))
 
         req = Request.blank('/photos/puppy.jpg', method='PUT', headers={
             'Content-Type': 'image/jpeg',
@@ -729,16 +750,15 @@ class TestRequest(S3ApiTestCase):
                               'MyyxeRY7whkBe+bq8fHCL/2kKUg='),
         })
         sigv2_req = S3Request(req.environ, storage_domain='s3.amazonaws.com')
-        expected_sts = '\n'.join([
-            'PUT',
-            '',
-            'image/jpeg',
-            'Tue, 27 Mar 2007 21:15:45 +0000',
-            '/johnsmith/photos/puppy.jpg',
+        expected_sts = b'\n'.join([
+            b'PUT',
+            b'',
+            b'image/jpeg',
+            b'Tue, 27 Mar 2007 21:15:45 +0000',
+            b'/johnsmith/photos/puppy.jpg',
         ])
         self.assertEqual(expected_sts, sigv2_req._string_to_sign())
-        self.assertTrue(sigv2_req.check_signature(
-            'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'))
+        self.assertTrue(sigv2_req.check_signature(secret))
 
         req = Request.blank(
             '/?prefix=photos&max-keys=50&marker=puppy',
@@ -750,16 +770,126 @@ class TestRequest(S3ApiTestCase):
                                   'htDYFYduRNen8P9ZfE/s9SuKy0U='),
             })
         sigv2_req = S3Request(req.environ, storage_domain='s3.amazonaws.com')
-        expected_sts = '\n'.join([
-            'GET',
-            '',
-            '',
-            'Tue, 27 Mar 2007 19:42:41 +0000',
-            '/johnsmith/',
+        expected_sts = b'\n'.join([
+            b'GET',
+            b'',
+            b'',
+            b'Tue, 27 Mar 2007 19:42:41 +0000',
+            b'/johnsmith/',
         ])
         self.assertEqual(expected_sts, sigv2_req._string_to_sign())
-        self.assertTrue(sigv2_req.check_signature(
-            'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'))
+        self.assertTrue(sigv2_req.check_signature(secret))
+
+    def test_check_signature_sigv2(self):
+        self._test_check_signature_sigv2(
+            'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY')
+
+    def test_check_signature_sigv2_unicode_string(self):
+        self._test_check_signature_sigv2(
+            u'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY')
+
+    @patch.object(S3Request, '_validate_dates', lambda *a: None)
+    def test_check_signature_multi_bytes_secret_failure(self):
+        # Test v2 check_signature with multi bytes invalid secret
+        req = Request.blank('/photos/puppy.jpg', headers={
+            'Host': 'johnsmith.s3.amazonaws.com',
+            'Date': 'Tue, 27 Mar 2007 19:36:42 +0000',
+            'Authorization': ('AWS AKIAIOSFODNN7EXAMPLE:'
+                              'bWq2s1WEIj+Ydj0vQ697zp+IXMU='),
+        })
+        sigv2_req = S3Request(req.environ, storage_domain='s3.amazonaws.com')
+        # This is a failure case with utf-8 non-ascii multi-bytes charactor
+        # but we expect to return just False instead of exceptions
+        self.assertFalse(sigv2_req.check_signature(
+            u'\u30c9\u30e9\u30b4\u30f3'))
+
+        # Test v4 check_signature with multi bytes invalid secret
+        amz_date_header = self.get_v4_amz_date_header()
+        req = Request.blank('/photos/puppy.jpg', headers={
+            'Authorization':
+                'AWS4-HMAC-SHA256 '
+                'Credential=test/%s/us-east-1/s3/aws4_request, '
+                'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
+                'Signature=X' % amz_date_header.split('T', 1)[0],
+            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Date': amz_date_header
+        })
+        sigv4_req = SigV4Request(
+            req.environ, storage_domain='s3.amazonaws.com')
+        self.assertFalse(sigv4_req.check_signature(
+            u'\u30c9\u30e9\u30b4\u30f3'))
+
+
+class TestHashingInput(S3ApiTestCase):
+    def test_good(self):
+        raw = b'123456789'
+        wrapped = HashingInput(BytesIO(raw), 9, hashlib.md5,
+                               hashlib.md5(raw).hexdigest())
+        self.assertEqual(b'1234', wrapped.read(4))
+        self.assertEqual(b'56', wrapped.read(2))
+        # trying to read past the end gets us whatever's left
+        self.assertEqual(b'789', wrapped.read(4))
+        # can continue trying to read -- but it'll be empty
+        self.assertEqual(b'', wrapped.read(2))
+
+        self.assertFalse(wrapped._input.closed)
+        wrapped.close()
+        self.assertTrue(wrapped._input.closed)
+
+    def test_empty(self):
+        wrapped = HashingInput(BytesIO(b''), 0, hashlib.sha256,
+                               hashlib.sha256(b'').hexdigest())
+        self.assertEqual(b'', wrapped.read(4))
+        self.assertEqual(b'', wrapped.read(2))
+
+        self.assertFalse(wrapped._input.closed)
+        wrapped.close()
+        self.assertTrue(wrapped._input.closed)
+
+    def test_too_long(self):
+        raw = b'123456789'
+        wrapped = HashingInput(BytesIO(raw), 8, hashlib.md5,
+                               hashlib.md5(raw).hexdigest())
+        self.assertEqual(b'1234', wrapped.read(4))
+        self.assertEqual(b'56', wrapped.read(2))
+        # even though the hash matches, there was more data than we expected
+        with self.assertRaises(swob.HTTPException) as raised:
+            wrapped.read(3)
+        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
+        # the error causes us to close the input
+        self.assertTrue(wrapped._input.closed)
+
+    def test_too_short(self):
+        raw = b'123456789'
+        wrapped = HashingInput(BytesIO(raw), 10, hashlib.md5,
+                               hashlib.md5(raw).hexdigest())
+        self.assertEqual(b'1234', wrapped.read(4))
+        self.assertEqual(b'56', wrapped.read(2))
+        # even though the hash matches, there was more data than we expected
+        with self.assertRaises(swob.HTTPException) as raised:
+            wrapped.read(4)
+        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
+        self.assertTrue(wrapped._input.closed)
+
+    def test_bad_hash(self):
+        raw = b'123456789'
+        wrapped = HashingInput(BytesIO(raw), 9, hashlib.sha256,
+                               hashlib.md5(raw).hexdigest())
+        self.assertEqual(b'1234', wrapped.read(4))
+        self.assertEqual(b'5678', wrapped.read(4))
+        with self.assertRaises(swob.HTTPException) as raised:
+            wrapped.read(4)
+        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
+        self.assertTrue(wrapped._input.closed)
+
+    def test_empty_bad_hash(self):
+        wrapped = HashingInput(BytesIO(b''), 0, hashlib.sha256, 'nope')
+        with self.assertRaises(swob.HTTPException) as raised:
+            wrapped.read(3)
+        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
+        # the error causes us to close the input
+        self.assertTrue(wrapped._input.closed)
+
 
 if __name__ == '__main__':
     unittest.main()
