@@ -196,11 +196,11 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
         self.middleware(req.environ, self.start_fake_response)
         self.assertEqual(self.response_status, 200)
 
-    def _assert_authorized(self, req, expect_token=True,
-                           account_path='/v1/AUTH_TENANT_ID/'):
+    def _assert_authorized(self, req, account_path='/v1/AUTH_TENANT_ID/'):
         self.assertTrue(
             req.path.startswith(account_path),
             '%r does not start with %r' % (req.path, account_path))
+        self.assertNotIn('X-Auth-Token', req.headers)
         expected_headers = {
             'X-Identity-Status': 'Confirmed',
             'X-Roles': 'swift-user,_member_',
@@ -210,12 +210,8 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
             'X-Tenant-Name': 'TENANT_NAME',
             'X-Project-Id': 'TENANT_ID',
             'X-Project-Name': 'TENANT_NAME',
-            'X-Auth-Token': 'TOKEN_ID',
         }
         for header, value in expected_headers.items():
-            if header == 'X-Auth-Token' and not expect_token:
-                self.assertNotIn(header, req.headers)
-                continue
             self.assertIn(header, req.headers)
             self.assertEqual(value, req.headers[header])
             # WSGI wants native strings for headers
@@ -253,7 +249,7 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
             'string_to_sign': u'token',
         }
         req.get_response(self.middleware)
-        self._assert_authorized(req, expect_token=False)
+        self._assert_authorized(req)
 
     def test_authorized_bytes(self):
         req = Request.blank('/v1/AUTH_cfa/c/o')
@@ -372,6 +368,15 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
         middleware = s3token.filter_factory(config)(self.app)
         self.assertIs('false_ind', middleware._verify)
 
+    def test_reseller_prefix(self):
+        def do_test(conf, expected):
+            conf.update(self.conf)
+            middleware = s3token.filter_factory(conf)(self.app)
+            self.assertEqual(expected, middleware._reseller_prefix)
+        do_test({}, 'AUTH_')
+        do_test({'reseller_prefix': 'KEY_'}, 'KEY_')
+        do_test({'reseller_prefix': 'KEY'}, 'KEY_')
+
     def test_auth_uris(self):
         for conf, expected in [
                 ({'auth_uri': 'https://example.com/v2.0'},
@@ -456,12 +461,12 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
             with self.assertRaises(ConfigFileError) as cm:
                 s3token.filter_factory({'auth_uri': auth_uri})(self.app)
             self.assertEqual('Invalid auth_uri; must include scheme and host',
-                             cm.exception.message)
+                             cm.exception.args[0])
         with self.assertRaises(ConfigFileError) as cm:
             s3token.filter_factory({
                 'auth_uri': 'nonhttp://example.com'})(self.app)
         self.assertEqual('Invalid auth_uri; scheme must be http or https',
-                         cm.exception.message)
+                         cm.exception.args[0])
         for auth_uri in [
                 'http://user@example.com/',
                 'http://example.com/?with=query',
@@ -469,7 +474,7 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
             with self.assertRaises(ConfigFileError) as cm:
                 s3token.filter_factory({'auth_uri': auth_uri})(self.app)
             self.assertEqual('Invalid auth_uri; must not include username, '
-                             'query, or fragment', cm.exception.message)
+                             'query, or fragment', cm.exception.args[0])
 
     def test_unicode_path(self):
         url = u'/v1/AUTH_cfa/c/euro\u20ac'.encode('utf8')
@@ -482,6 +487,119 @@ class S3TokenMiddlewareTestGood(S3TokenMiddlewareTestBase):
         req.get_response(self.middleware)
         self._assert_authorized(req)
 
+    def test_authorize_with_access_key(self):
+        req = Request.blank('/v1/accesskey/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+        }
+        req.get_response(self.middleware)
+        self._assert_authorized(req, account_path='/v1/')
+        self.assertEqual(req.environ['PATH_INFO'], '/v1/AUTH_TENANT_ID/c/o')
+
+    def test_authorize_with_access_key_and_unquote_chars(self):
+        req = Request.blank('/v1/access%key=/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+        }
+        req.get_response(self.middleware)
+        self._assert_authorized(req, account_path='/v1/')
+        self.assertEqual(req.environ['PATH_INFO'], '/v1/AUTH_TENANT_ID/c/o')
+
+    @mock.patch('swift.common.middleware.s3api.s3token.cache_from_env')
+    @mock.patch('keystoneclient.v3.client.Client')
+    @mock.patch.object(requests, 'post')
+    def test_secret_is_cached(self, MOCK_REQUEST, MOCK_KEYSTONE,
+                              MOCK_CACHE_FROM_ENV):
+        self.middleware = s3token.filter_factory({
+            'auth_uri': 'http://example.com',
+            'secret_cache_duration': '20',
+            'auth_type': 'v3password',
+            'auth_url': 'http://example.com:5000/v3',
+            'username': 'swift',
+            'password': 'secret',
+            'project_name': 'service',
+            'user_domain_name': 'default',
+            'project_domain_name': 'default',
+        })(FakeApp())
+        self.assertEqual(20, self.middleware._secret_cache_duration)
+
+        cache = MOCK_CACHE_FROM_ENV.return_value
+
+        fake_cache_response = ({}, {'id': 'tenant_id'}, 'secret')
+        cache.get.return_value = fake_cache_response
+
+        MOCK_REQUEST.return_value = TestResponse({
+            'status_code': 201,
+            'text': json.dumps(GOOD_RESPONSE_V2)})
+
+        req = Request.blank('/v1/AUTH_cfa/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+            'check_signature': lambda x: True
+        }
+        req.get_response(self.middleware)
+        # Ensure we don't request auth from keystone
+        self.assertFalse(MOCK_REQUEST.called)
+
+    @mock.patch('swift.common.middleware.s3api.s3token.cache_from_env')
+    @mock.patch('keystoneclient.v3.client.Client')
+    @mock.patch.object(requests, 'post')
+    def test_secret_sets_cache(self, MOCK_REQUEST, MOCK_KEYSTONE,
+                               MOCK_CACHE_FROM_ENV):
+        self.middleware = s3token.filter_factory({
+            'auth_uri': 'http://example.com',
+            'secret_cache_duration': '20',
+            'auth_type': 'v3password',
+            'auth_url': 'http://example.com:5000/v3',
+            'username': 'swift',
+            'password': 'secret',
+            'project_name': 'service',
+            'user_domain_name': 'default',
+            'project_domain_name': 'default',
+        })(FakeApp())
+        self.assertEqual(20, self.middleware._secret_cache_duration)
+
+        cache = MOCK_CACHE_FROM_ENV.return_value
+        cache.get.return_value = None
+
+        keystone_client = MOCK_KEYSTONE.return_value
+        keystone_client.ec2.get.return_value = mock.Mock(secret='secret')
+
+        MOCK_REQUEST.return_value = TestResponse({
+            'status_code': 201,
+            'text': json.dumps(GOOD_RESPONSE_V2).encode('ascii')})
+
+        req = Request.blank('/v1/AUTH_cfa/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+            'check_signature': lambda x: True
+        }
+        req.get_response(self.middleware)
+        expected_headers = {
+            'X-Identity-Status': u'Confirmed',
+            'X-Roles': u'swift-user,_member_',
+            'X-User-Id': u'USER_ID',
+            'X-User-Name': u'S3_USER',
+            'X-Tenant-Id': u'TENANT_ID',
+            'X-Tenant-Name': u'TENANT_NAME',
+            'X-Project-Id': u'TENANT_ID',
+            'X-Project-Name': u'TENANT_NAME',
+        }
+
+        self.assertTrue(MOCK_REQUEST.called)
+        tenant = GOOD_RESPONSE_V2['access']['token']['tenant']
+        expected_cache = (expected_headers, tenant, 'secret')
+        cache.set.assert_called_once_with('s3secret/access', expected_cache,
+                                          time=20)
+
 
 class S3TokenMiddlewareTestBad(S3TokenMiddlewareTestBase):
     def test_unauthorized_token(self):
@@ -819,3 +937,25 @@ class S3TokenMiddlewareTestV3(S3TokenMiddlewareTestBase):
         self._test_bad_reply_missing_parts('token', 'project', 'domain')
         self._test_bad_reply_missing_parts('token', 'project')
         self._test_bad_reply_missing_parts('token', 'roles')
+
+    def test_authorize_with_access_key(self):
+        req = Request.blank('/v1/accesskey/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+        }
+        req.get_response(self.middleware)
+        self._assert_authorized(req, account_path='/v1/')
+        self.assertEqual(req.environ['PATH_INFO'], '/v1/AUTH_PROJECT_ID/c/o')
+
+    def test_authorize_with_access_key_and_unquote_chars(self):
+        req = Request.blank('/v1/ab%c=/c/o')
+        req.environ['s3api.auth_details'] = {
+            'access_key': u'access',
+            'signature': u'signature',
+            'string_to_sign': u'token',
+        }
+        req.get_response(self.middleware)
+        self._assert_authorized(req, account_path='/v1/')
+        self.assertEqual(req.environ['PATH_INFO'], '/v1/AUTH_PROJECT_ID/c/o')
