@@ -14,20 +14,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import time
 import mock
-import shutil
-import tempfile
 import unittest
 from hashlib import md5
 from six.moves import urllib
-from textwrap import dedent
 
 from swift.common import swob
 from swift.common.middleware import copy
 from swift.common.storage_policy import POLICIES
 from swift.common.swob import Request, HTTPException
-from test.unit import patch_policies, debug_logger, FakeMemcache, FakeRing
+from swift.common.utils import closing_if_possible
+from test.unit import patch_policies, debug_logger, FakeRing
 from test.unit.common.middleware.helpers import FakeSwift
 from test.unit.proxy.controllers.test_obj import set_http_connect, \
     PatchedObjControllerApp
@@ -92,11 +89,12 @@ class TestCopyConstraints(unittest.TestCase):
 class TestServerSideCopyMiddleware(unittest.TestCase):
     def setUp(self):
         self.app = FakeSwift()
-        self.ssc = copy.filter_factory({
-            'object_post_as_copy': 'yes',
-        })(self.app)
+        self.ssc = copy.filter_factory({})(self.app)
         self.ssc.logger = self.app.logger
 
+    def tearDown(self):
+        self.assertEqual(self.app.unclosed_requests, {})
+
     def call_app(self, req, app=None, expect_exception=False):
         if app is None:
             app = self.app
@@ -119,11 +117,13 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
             headers[0] = h
 
         body_iter = app(req.environ, start_response)
-        body = ''
+        body = b''
         caught_exc = None
         try:
-            for chunk in body_iter:
-                body += chunk
+            # appease the close-checker
+            with closing_if_possible(body_iter):
+                for chunk in body_iter:
+                    body += chunk
         except Exception as exc:
             if expect_exception:
                 caught_exc = exc
@@ -150,65 +150,15 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         self.assertEqual(len(self.authorized), 1)
         self.assertRequestEqual(req, self.authorized[0])
 
-    def test_object_delete_pass_through(self):
-        self.app.register('DELETE', '/v1/a/c/o', swob.HTTPOk, {})
-        req = Request.blank('/v1/a/c/o', method='DELETE')
-        status, headers, body = self.call_ssc(req)
-        self.assertEqual(status, '200 OK')
-        self.assertEqual(len(self.authorized), 1)
-        self.assertRequestEqual(req, self.authorized[0])
-
-    def test_POST_as_COPY_simple(self):
-        self.app.register('GET', '/v1/a/c/o', swob.HTTPOk, {}, 'passed')
-        self.app.register('PUT', '/v1/a/c/o', swob.HTTPAccepted, {})
-        req = Request.blank('/v1/a/c/o', method='POST')
-        status, headers, body = self.call_ssc(req)
-        self.assertEqual(status, '202 Accepted')
-        self.assertEqual(len(self.authorized), 1)
-        self.assertRequestEqual(req, self.authorized[0])
-
-    def test_POST_as_COPY_201_return_202(self):
-        self.app.register('GET', '/v1/a/c/o', swob.HTTPOk, {}, 'passed')
-        self.app.register('PUT', '/v1/a/c/o', swob.HTTPCreated, {})
-        req = Request.blank('/v1/a/c/o', method='POST')
-        status, headers, body = self.call_ssc(req)
-        self.assertEqual(status, '202 Accepted')
-        self.assertEqual(len(self.authorized), 1)
-        self.assertRequestEqual(req, self.authorized[0])
-
-    def test_POST_delete_at(self):
-        self.app.register('GET', '/v1/a/c/o', swob.HTTPOk, {}, 'passed')
-        self.app.register('PUT', '/v1/a/c/o', swob.HTTPAccepted, {})
-        t = str(int(time.time() + 100))
-        req = Request.blank('/v1/a/c/o', method='POST',
-                            headers={'Content-Type': 'foo/bar',
-                                     'X-Delete-At': t})
-        status, headers, body = self.call_ssc(req)
-        self.assertEqual(status, '202 Accepted')
-
-        calls = self.app.calls_with_headers
-        method, path, req_headers = calls[1]
-        self.assertEqual('PUT', method)
-        self.assertTrue('X-Delete-At' in req_headers)
-        self.assertEqual(req_headers['X-Delete-At'], str(t))
-        self.assertEqual(len(self.authorized), 1)
-        self.assertRequestEqual(req, self.authorized[0])
-
-    def test_POST_as_COPY_static_large_object(self):
-        self.app.register('GET', '/v1/a/c/o', swob.HTTPOk,
-                          {'X-Static-Large-Object': True}, 'passed')
-        self.app.register('PUT', '/v1/a/c/o', swob.HTTPAccepted, {})
-        req = Request.blank('/v1/a/c/o', method='POST',
-                            headers={})
-        status, headers, body = self.call_ssc(req)
-        self.assertEqual(status, '202 Accepted')
-
-        calls = self.app.calls_with_headers
-        method, path, req_headers = calls[1]
-        self.assertEqual('PUT', method)
-        self.assertNotIn('X-Static-Large-Object', req_headers)
-        self.assertEqual(len(self.authorized), 1)
-        self.assertRequestEqual(req, self.authorized[0])
+    def test_object_pass_through_methods(self):
+        for method in ['DELETE', 'GET', 'HEAD', 'REPLICATE']:
+            self.app.register(method, '/v1/a/c/o', swob.HTTPOk, {})
+            req = Request.blank('/v1/a/c/o', method=method)
+            status, headers, body = self.call_ssc(req)
+            self.assertEqual(status, '200 OK')
+            self.assertEqual(len(self.authorized), 1)
+            self.assertRequestEqual(req, self.authorized[0])
+            self.assertNotIn('swift.orig_req_method', req.environ)
 
     def test_basic_put_with_x_copy_from(self):
         self.app.register('GET', '/v1/a/c/o', swob.HTTPOk, {}, 'passed')
@@ -224,6 +174,10 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         self.assertEqual('/v1/a/c/o', self.authorized[0].path)
         self.assertEqual('PUT', self.authorized[1].method)
         self.assertEqual('/v1/a/c/o2', self.authorized[1].path)
+        self.assertEqual(self.app.swift_sources[0], 'SSC')
+        self.assertEqual(self.app.swift_sources[1], 'SSC')
+        # For basic test cases, assert orig_req_method behavior
+        self.assertNotIn('swift.orig_req_method', req.environ)
 
     def test_static_large_object_manifest(self):
         self.app.register('GET', '/v1/a/c/o', swob.HTTPOk,
@@ -381,8 +335,31 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         self.assertEqual('PUT', self.authorized[1].method)
         self.assertEqual('/v1/a/c/o', self.authorized[1].path)
 
+    def test_copy_with_unicode(self):
+        self.app.register('GET', '/v1/a/c/\xF0\x9F\x8C\xB4', swob.HTTPOk,
+                          {}, 'passed')
+        self.app.register('PUT', '/v1/a/c/\xE2\x98\x83', swob.HTTPCreated, {})
+        # Just for fun, let's have a mix of properly encoded and not
+        req = Request.blank('/v1/a/c/%F0\x9F%8C%B4',
+                            environ={'REQUEST_METHOD': 'COPY'},
+                            headers={'Content-Length': '0',
+                                     'Destination': 'c/%E2\x98%83'})
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(status, '201 Created')
+        calls = self.app.calls_with_headers
+        method, path, req_headers = calls[0]
+        self.assertEqual('GET', method)
+        self.assertEqual('/v1/a/c/\xF0\x9F\x8C\xB4', path)
+        self.assertIn(('X-Copied-From', 'c/%F0%9F%8C%B4'), headers)
+
+        self.assertEqual(len(self.authorized), 2)
+        self.assertEqual('GET', self.authorized[0].method)
+        self.assertEqual('/v1/a/c/%F0%9F%8C%B4', self.authorized[0].path)
+        self.assertEqual('PUT', self.authorized[1].method)
+        self.assertEqual('/v1/a/c/%E2%98%83', self.authorized[1].path)
+
     def test_copy_with_spaces_in_x_copy_from_and_account(self):
-        self.app.register('GET', '/v1/a/c/o o2', swob.HTTPOk, {}, 'passed')
+        self.app.register('GET', '/v1/a/c/o o2', swob.HTTPOk, {}, b'passed')
         self.app.register('PUT', '/v1/a1/c1/o', swob.HTTPCreated, {})
         # space in soure path
         req = Request.blank('/v1/a1/c1/o', environ={'REQUEST_METHOD': 'PUT'},
@@ -489,24 +466,16 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         req = Request.blank('/v1/a/c/o', environ={'REQUEST_METHOD': 'PUT'},
                             headers={'Content-Length': '0',
                                      'X-Copy-From': '/c'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_copy_with_no_object_in_x_copy_from_and_account(self):
         req = Request.blank('/v1/a/c/o', environ={'REQUEST_METHOD': 'PUT'},
                             headers={'Content-Length': '0',
                                      'X-Copy-From': '/c',
                                      'X-Copy-From-Account': 'a'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_copy_with_bad_x_copy_from_account(self):
         req = Request.blank('/v1/a/c/o',
@@ -514,12 +483,8 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
                             headers={'Content-Length': '0',
                                      'X-Copy-From': '/c/o',
                                      'X-Copy-From-Account': '/i/am/bad'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_copy_server_error_reading_source(self):
         self.app.register('GET', '/v1/a/c/o', swob.HTTPServiceUnavailable, {})
@@ -641,6 +606,8 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
             ('PUT', '/v1/a/c/o-copy')])
         self.assertIn('etag', self.app.headers[1])
         self.assertEqual(self.app.headers[1]['etag'], 'is sent')
+        # For basic test cases, assert orig_req_method behavior
+        self.assertEqual(req.environ['swift.orig_req_method'], 'COPY')
 
     def test_basic_DLO(self):
         self.app.register('GET', '/v1/a/c/o', swob.HTTPOk, {
@@ -699,7 +666,7 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         get_resp_headers['etag'] = 'source etag'
         self.app.register(
             'GET', '/v1/a/c/o', swob.HTTPOk,
-            headers=get_resp_headers, body='passed')
+            headers=get_resp_headers, body=b'passed')
 
         def verify_headers(expected_headers, unexpected_headers,
                            actual_headers):
@@ -960,36 +927,27 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         req = Request.blank('/v1/a/c/o',
                             environ={'REQUEST_METHOD': 'COPY'},
                             headers={'Destination': 'c_o'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_COPY_account_no_object_in_destination(self):
         req = Request.blank('/v1/a/c/o',
                             environ={'REQUEST_METHOD': 'COPY'},
                             headers={'Destination': 'c_o',
                                      'Destination-Account': 'a1'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_COPY_account_bad_destination_account(self):
         req = Request.blank('/v1/a/c/o',
                             environ={'REQUEST_METHOD': 'COPY'},
                             headers={'Destination': '/c/o',
                                      'Destination-Account': '/i/am/bad'})
-        try:
-            status, headers, body = self.call_ssc(req)
-        except HTTPException as resp:
-            self.assertEqual("412 Precondition Failed", str(resp))
-        else:
-            self.fail("Expecting HTTPException.")
+        status, headers, body = self.call_ssc(req)
+
+        self.assertEqual(status, '412 Precondition Failed')
 
     def test_COPY_server_error_reading_source(self):
         self.app.register('GET', '/v1/a/c/o', swob.HTTPServiceUnavailable, {})
@@ -1178,6 +1136,8 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         self.assertEqual(len(self.authorized), 1)
         self.assertEqual('OPTIONS', self.authorized[0].method)
         self.assertEqual('/v1/a/c/o', self.authorized[0].path)
+        # For basic test cases, assert orig_req_method behavior
+        self.assertNotIn('swift.orig_req_method', req.environ)
 
     def test_COPY_in_OPTIONS_response_CORS(self):
         self.app.register('OPTIONS', '/v1/a/c/o', swob.HTTPOk,
@@ -1198,68 +1158,147 @@ class TestServerSideCopyMiddleware(unittest.TestCase):
         self.assertEqual('OPTIONS', self.authorized[0].method)
         self.assertEqual('/v1/a/c/o', self.authorized[0].path)
 
+    def _test_COPY_source_headers(self, extra_put_headers):
+        # helper method to perform a COPY with some metadata headers that
+        # should always be sent to the destination
+        put_headers = {'Destination': '/c1/o',
+                       'X-Object-Meta-Test2': 'added',
+                       'X-Object-Sysmeta-Test2': 'added',
+                       'X-Object-Transient-Sysmeta-Test2': 'added'}
+        put_headers.update(extra_put_headers)
+        get_resp_headers = {
+            'X-Timestamp': '1234567890.12345',
+            'X-Backend-Timestamp': '1234567890.12345',
+            'Content-Type': 'text/original',
+            'Content-Encoding': 'gzip',
+            'Content-Disposition': 'attachment; filename=myfile',
+            'X-Object-Meta-Test': 'original',
+            'X-Object-Sysmeta-Test': 'original',
+            'X-Object-Transient-Sysmeta-Test': 'original',
+            'X-Foo': 'Bar'}
+        self.app.register(
+            'GET', '/v1/a/c/o', swob.HTTPOk, headers=get_resp_headers)
+        self.app.register('PUT', '/v1/a/c1/o', swob.HTTPCreated, {})
+        req = Request.blank('/v1/a/c/o', method='COPY', headers=put_headers)
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(status, '201 Created')
+        calls = self.app.calls_with_headers
+        self.assertEqual(2, len(calls))
+        method, path, req_headers = calls[1]
+        self.assertEqual('PUT', method)
+        # these headers should always be applied to the destination
+        self.assertEqual('added', req_headers.get('X-Object-Meta-Test2'))
+        self.assertEqual('added', req_headers.get('X-Object-Sysmeta-Test2'))
+        self.assertEqual('added',
+                         req_headers.get('X-Object-Transient-Sysmeta-Test2'))
+        return req_headers
 
-class TestServerSideCopyConfiguration(unittest.TestCase):
+    def test_COPY_source_headers_no_updates(self):
+        # copy should preserve existing metadata if not updated
+        req_headers = self._test_COPY_source_headers({})
+        self.assertEqual('text/original', req_headers.get('Content-Type'))
+        self.assertEqual('gzip', req_headers.get('Content-Encoding'))
+        self.assertEqual('attachment; filename=myfile',
+                         req_headers.get('Content-Disposition'))
+        self.assertEqual('original', req_headers.get('X-Object-Meta-Test'))
+        self.assertEqual('original', req_headers.get('X-Object-Sysmeta-Test'))
+        self.assertEqual('original',
+                         req_headers.get('X-Object-Transient-Sysmeta-Test'))
+        self.assertEqual('Bar', req_headers.get('X-Foo'))
+        self.assertNotIn('X-Timestamp', req_headers)
+        self.assertNotIn('X-Backend-Timestamp', req_headers)
 
-    def setUp(self):
-        self.tmpdir = tempfile.mkdtemp()
+    def test_COPY_source_headers_with_updates(self):
+        # copy should apply any updated values to existing metadata
+        put_headers = {
+            'Content-Type': 'text/not_original',
+            'Content-Encoding': 'not_gzip',
+            'Content-Disposition': 'attachment; filename=notmyfile',
+            'X-Object-Meta-Test': 'not_original',
+            'X-Object-Sysmeta-Test': 'not_original',
+            'X-Object-Transient-Sysmeta-Test': 'not_original',
+            'X-Foo': 'Not Bar'}
+        req_headers = self._test_COPY_source_headers(put_headers)
+        self.assertEqual('text/not_original', req_headers.get('Content-Type'))
+        self.assertEqual('not_gzip', req_headers.get('Content-Encoding'))
+        self.assertEqual('attachment; filename=notmyfile',
+                         req_headers.get('Content-Disposition'))
+        self.assertEqual('not_original', req_headers.get('X-Object-Meta-Test'))
+        self.assertEqual('not_original',
+                         req_headers.get('X-Object-Sysmeta-Test'))
+        self.assertEqual('not_original',
+                         req_headers.get('X-Object-Transient-Sysmeta-Test'))
+        self.assertEqual('Not Bar', req_headers.get('X-Foo'))
+        self.assertNotIn('X-Timestamp', req_headers)
+        self.assertNotIn('X-Backend-Timestamp', req_headers)
 
-    def tearDown(self):
-        shutil.rmtree(self.tmpdir)
+    def test_COPY_x_fresh_metadata_no_updates(self):
+        # existing user metadata should not be copied, sysmeta is copied
+        put_headers = {
+            'X-Fresh-Metadata': 'true',
+            'X-Extra': 'Fresh'}
+        req_headers = self._test_COPY_source_headers(put_headers)
+        self.assertEqual('text/original', req_headers.get('Content-Type'))
+        self.assertEqual('Fresh', req_headers.get('X-Extra'))
+        self.assertEqual('original',
+                         req_headers.get('X-Object-Sysmeta-Test'))
+        self.assertIn('X-Fresh-Metadata', req_headers)
+        self.assertNotIn('X-Object-Meta-Test', req_headers)
+        self.assertNotIn('X-Object-Transient-Sysmeta-Test', req_headers)
+        self.assertNotIn('X-Timestamp', req_headers)
+        self.assertNotIn('X-Backend-Timestamp', req_headers)
+        self.assertNotIn('Content-Encoding', req_headers)
+        self.assertNotIn('Content-Disposition', req_headers)
+        self.assertNotIn('X-Foo', req_headers)
 
-    def test_reading_proxy_conf_when_no_middleware_conf_present(self):
-        proxy_conf = dedent("""
-        [DEFAULT]
-        bind_ip = 10.4.5.6
+    def test_COPY_x_fresh_metadata_with_updates(self):
+        # existing user metadata should not be copied, new metadata replaces it
+        put_headers = {
+            'X-Fresh-Metadata': 'true',
+            'Content-Type': 'text/not_original',
+            'Content-Encoding': 'not_gzip',
+            'Content-Disposition': 'attachment; filename=notmyfile',
+            'X-Object-Meta-Test': 'not_original',
+            'X-Object-Sysmeta-Test': 'not_original',
+            'X-Object-Transient-Sysmeta-Test': 'not_original',
+            'X-Foo': 'Not Bar',
+            'X-Extra': 'Fresh'}
+        req_headers = self._test_COPY_source_headers(put_headers)
+        self.assertEqual('Fresh', req_headers.get('X-Extra'))
+        self.assertEqual('text/not_original', req_headers.get('Content-Type'))
+        self.assertEqual('not_gzip', req_headers.get('Content-Encoding'))
+        self.assertEqual('attachment; filename=notmyfile',
+                         req_headers.get('Content-Disposition'))
+        self.assertEqual('not_original', req_headers.get('X-Object-Meta-Test'))
+        self.assertEqual('not_original',
+                         req_headers.get('X-Object-Sysmeta-Test'))
+        self.assertEqual('not_original',
+                         req_headers.get('X-Object-Transient-Sysmeta-Test'))
+        self.assertEqual('Not Bar', req_headers.get('X-Foo'))
 
-        [pipeline:main]
-        pipeline = catch_errors copy ye-olde-proxy-server
+    def test_COPY_with_single_range(self):
+        # verify that source etag is not copied when copying a range
+        self.app.register('GET', '/v1/a/c/o', swob.HTTPOk,
+                          {'etag': 'bogus etag'}, "abcdefghijklmnop")
+        self.app.register('PUT', '/v1/a/c1/o', swob.HTTPCreated, {})
+        req = swob.Request.blank(
+            '/v1/a/c/o', method='COPY',
+            headers={'Destination': 'c1/o',
+                     'Range': 'bytes=5-10'})
 
-        [filter:copy]
-        use = egg:swift#copy
+        status, headers, body = self.call_ssc(req)
 
-        [app:ye-olde-proxy-server]
-        use = egg:swift#proxy
-        object_post_as_copy = no
-        """)
-
-        conffile = tempfile.NamedTemporaryFile()
-        conffile.write(proxy_conf)
-        conffile.flush()
-
-        ssc = copy.filter_factory({
-            '__file__': conffile.name
-        })("no app here")
-
-        self.assertEqual(ssc.object_post_as_copy, False)
-
-    def test_middleware_conf_precedence(self):
-        proxy_conf = dedent("""
-        [DEFAULT]
-        bind_ip = 10.4.5.6
-
-        [pipeline:main]
-        pipeline = catch_errors copy ye-olde-proxy-server
-
-        [filter:copy]
-        use = egg:swift#copy
-        object_post_as_copy = no
-
-        [app:ye-olde-proxy-server]
-        use = egg:swift#proxy
-        object_post_as_copy = yes
-        """)
-
-        conffile = tempfile.NamedTemporaryFile()
-        conffile.write(proxy_conf)
-        conffile.flush()
-
-        ssc = copy.filter_factory({
-            'object_post_as_copy': 'no',
-            '__file__': conffile.name
-        })("no app here")
-
-        self.assertEqual(ssc.object_post_as_copy, False)
+        self.assertEqual(status, '201 Created')
+        calls = self.app.calls_with_headers
+        self.assertEqual(2, len(calls))
+        method, path, req_headers = calls[1]
+        self.assertEqual('PUT', method)
+        self.assertEqual('/v1/a/c1/o', path)
+        self.assertNotIn('etag', (h.lower() for h in req_headers))
+        self.assertEqual('6', req_headers['content-length'])
+        req = swob.Request.blank('/v1/a/c1/o', method='GET')
+        status, headers, body = self.call_ssc(req)
+        self.assertEqual(b'fghijk', body)
 
 
 @patch_policies(with_ec_default=True)
@@ -1277,23 +1316,21 @@ class TestServerSideCopyMiddlewareWithEC(unittest.TestCase):
         self.logger = debug_logger('proxy-server')
         self.logger.thread_locals = ('txn1', '127.0.0.2')
         self.app = PatchedObjControllerApp(
-            None, FakeMemcache(), account_ring=FakeRing(),
+            None, account_ring=FakeRing(),
             container_ring=FakeRing(), logger=self.logger)
-        self.ssc = copy.filter_factory({
-            'object_post_as_copy': 'yes',
-        })(self.app)
+        self.ssc = copy.filter_factory({})(self.app)
         self.ssc.logger = self.app.logger
         self.policy = POLICIES.default
         self.app.container_info = dict(self.container_info)
 
-    def test_COPY_with_ranges(self):
+    def test_COPY_with_single_range(self):
         req = swob.Request.blank(
             '/v1/a/c/o', method='COPY',
             headers={'Destination': 'c1/o',
                      'Range': 'bytes=5-10'})
         # turn a real body into fragments
         segment_size = self.policy.ec_segment_size
-        real_body = ('asdf' * segment_size)[:-10]
+        real_body = (b'asdf' * segment_size)[:-10]
 
         # split it up into chunks
         chunks = [real_body[x:x + segment_size]
@@ -1304,27 +1341,41 @@ class TestServerSideCopyMiddlewareWithEC(unittest.TestCase):
         fragment_payloads = []
         fragment_payloads.append(fragments)
 
-        node_fragments = zip(*fragment_payloads)
+        node_fragments = list(zip(*fragment_payloads))
         self.assertEqual(len(node_fragments),
                          self.policy.object_ring.replicas)  # sanity
         headers = {'X-Object-Sysmeta-Ec-Content-Length': str(len(real_body))}
-        responses = [(200, ''.join(node_fragments[i]), headers)
+        responses = [(200, b''.join(node_fragments[i]), headers)
                      for i in range(POLICIES.default.ec_ndata)]
-        responses += [(201, '', {})] * self.policy.object_ring.replicas
+        responses += [(201, b'', {})] * self.policy.object_ring.replicas
         status_codes, body_iter, headers = zip(*responses)
         expect_headers = {
             'X-Obj-Metadata-Footer': 'yes',
             'X-Obj-Multiphase-Commit': 'yes'
         }
+
+        put_hdrs = []
+
+        def capture_conn(host, port, dev, part, method, path, *args, **kwargs):
+            if method == 'PUT':
+                put_hdrs.append(args[0])
+
         with set_http_connect(*status_codes, body_iter=body_iter,
-                              headers=headers, expect_headers=expect_headers):
+                              headers=headers, expect_headers=expect_headers,
+                              give_connect=capture_conn):
             resp = req.get_response(self.ssc)
+
         self.assertEqual(resp.status_int, 201)
+        expected_puts = POLICIES.default.ec_ndata + POLICIES.default.ec_nparity
+        self.assertEqual(expected_puts, len(put_hdrs))
+        for hdrs in put_hdrs:
+            # etag should not be copied from source
+            self.assertNotIn('etag', (h.lower() for h in hdrs))
 
     def test_COPY_with_invalid_ranges(self):
         # real body size is segment_size - 10 (just 1 segment)
         segment_size = self.policy.ec_segment_size
-        real_body = ('a' * segment_size)[:-10]
+        real_body = (b'a' * segment_size)[:-10]
 
         # range is out of real body but in segment size
         self._test_invalid_ranges('COPY', real_body,
@@ -1344,7 +1395,7 @@ class TestServerSideCopyMiddlewareWithEC(unittest.TestCase):
         fragments = self.policy.pyeclib_driver.encode(real_body)
         fragment_payloads = [fragments]
 
-        node_fragments = zip(*fragment_payloads)
+        node_fragments = list(zip(*fragment_payloads))
         self.assertEqual(len(node_fragments),
                          self.policy.object_ring.replicas)  # sanity
         headers = {'X-Object-Sysmeta-Ec-Content-Length': str(len(real_body)),
@@ -1354,11 +1405,12 @@ class TestServerSideCopyMiddlewareWithEC(unittest.TestCase):
         title, exp = swob.RESPONSE_REASONS[416]
         range_not_satisfiable_body = \
             '<html><h1>%s</h1><p>%s</p></html>' % (title, exp)
+        range_not_satisfiable_body = range_not_satisfiable_body.encode('ascii')
         if start >= segment_size:
             responses = [(416, range_not_satisfiable_body, headers)
                          for i in range(POLICIES.default.ec_ndata)]
         else:
-            responses = [(200, ''.join(node_fragments[i]), headers)
+            responses = [(200, b''.join(node_fragments[i]), headers)
                          for i in range(POLICIES.default.ec_ndata)]
         status_codes, body_iter, headers = zip(*responses)
         expect_headers = {
