@@ -72,9 +72,18 @@ class FakeMemcache(object):
 
 
 class FakeApp(object):
+    skip_handled_check = False
 
     def __call__(self, env, start_response):
-        return ['204 No Content']
+        assert self.skip_handled_check or env.get('swift.ratelimit.handled')
+        start_response('200 OK', [])
+        return [b'Some Content']
+
+
+class FakeReq(object):
+    def __init__(self, method, env=None):
+        self.method = method
+        self.environ = env or {}
 
 
 def start_response(*args):
@@ -159,36 +168,29 @@ class TestRateLimit(unittest.TestCase):
             {'object_count': '5'}
         the_app = ratelimit.filter_factory(conf_dict)(FakeApp())
         the_app.memcache_client = fake_memcache
-        req = lambda: None
-        req.environ = {'swift.cache': fake_memcache, 'PATH_INFO': '/v1/a/c/o'}
+        environ = {'swift.cache': fake_memcache, 'PATH_INFO': '/v1/a/c/o'}
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
                         lambda *args, **kwargs: {}):
-            req.method = 'DELETE'
             self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-                req, 'a', None, None)), 0)
-            req.method = 'PUT'
+                FakeReq('DELETE', environ), 'a', None, None)), 0)
             self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-                req, 'a', 'c', None)), 1)
-            req.method = 'DELETE'
+                FakeReq('PUT', environ), 'a', 'c', None)), 1)
             self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-                req, 'a', 'c', None)), 1)
-            req.method = 'GET'
+                FakeReq('DELETE', environ), 'a', 'c', None)), 1)
             self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-                req, 'a', 'c', 'o')), 0)
-            req.method = 'PUT'
+                FakeReq('GET', environ), 'a', 'c', 'o')), 0)
             self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-                req, 'a', 'c', 'o')), 1)
+                FakeReq('PUT', environ), 'a', 'c', 'o')), 1)
 
-        req.method = 'PUT'
         self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-            req, 'a', 'c', None, global_ratelimit=10)), 2)
+            FakeReq('PUT', environ), 'a', 'c', None, global_ratelimit=10)), 2)
         self.assertEqual(the_app.get_ratelimitable_key_tuples(
-            req, 'a', 'c', None, global_ratelimit=10)[1],
+            FakeReq('PUT', environ), 'a', 'c', None, global_ratelimit=10)[1],
             ('ratelimit/global-write/a', 10))
 
-        req.method = 'PUT'
         self.assertEqual(len(the_app.get_ratelimitable_key_tuples(
-            req, 'a', 'c', None, global_ratelimit='notafloat')), 1)
+            FakeReq('PUT', environ), 'a', 'c', None,
+            global_ratelimit='notafloat')), 1)
 
     def test_memcached_container_info_dict(self):
         mdict = headers_to_container_info({'x-container-object-count': '45'})
@@ -203,9 +205,8 @@ class TestRateLimit(unittest.TestCase):
             {'container_size': 5}
         the_app = ratelimit.filter_factory(conf_dict)(FakeApp())
         the_app.memcache_client = fake_memcache
-        req = lambda: None
-        req.method = 'PUT'
-        req.environ = {'PATH_INFO': '/v1/a/c/o', 'swift.cache': fake_memcache}
+        req = FakeReq('PUT', {
+            'PATH_INFO': '/v1/a/c/o', 'swift.cache': fake_memcache})
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
                         lambda *args, **kwargs: {}):
             tuples = the_app.get_ratelimitable_key_tuples(req, 'a', 'c', 'o')
@@ -223,11 +224,11 @@ class TestRateLimit(unittest.TestCase):
                     lambda *args, **kwargs: {}):
                 for meth, exp_time in [('DELETE', 9.8), ('GET', 0),
                                        ('POST', 0), ('PUT', 9.8)]:
-                    req = Request.blank('/v/a%s/c' % meth)
+                    req = Request.blank('/v1/a%s/c' % meth)
                     req.method = meth
                     req.environ['swift.cache'] = FakeMemcache()
-                    make_app_call = lambda: self.test_ratelimit(req.environ,
-                                                                start_response)
+                    make_app_call = lambda: self.test_ratelimit(
+                        req.environ.copy(), start_response)
                     begin = time.time()
                     self._run(make_app_call, num_calls, current_rate,
                               check_time=bool(exp_time))
@@ -239,11 +240,11 @@ class TestRateLimit(unittest.TestCase):
         num_calls = 50
         conf_dict = {'account_ratelimit': current_rate}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
-        req = Request.blank('/v/a/c')
+        req = Request.blank('/v1/a/c')
         req.method = 'PUT'
         req.environ['swift.cache'] = FakeMemcache()
         req.environ['swift.cache'].init_incr_return_neg = True
-        make_app_call = lambda: self.test_ratelimit(req.environ,
+        make_app_call = lambda: self.test_ratelimit(req.environ.copy(),
                                                     start_response)
         begin = time.time()
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
@@ -259,15 +260,15 @@ class TestRateLimit(unittest.TestCase):
                      'account_whitelist': 'a',
                      'account_blacklist': 'b'}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
-        req = Request.blank('/')
         with mock.patch.object(self.test_ratelimit,
                                'memcache_client', FakeMemcache()):
             self.assertEqual(
-                self.test_ratelimit.handle_ratelimit(req, 'a', 'c', 'o'),
+                self.test_ratelimit.handle_ratelimit(
+                    Request.blank('/'), 'a', 'c', 'o'),
                 None)
             self.assertEqual(
                 self.test_ratelimit.handle_ratelimit(
-                    req, 'b', 'c', 'o').status_int,
+                    Request.blank('/'), 'b', 'c', 'o').status_int,
                 497)
 
     def test_ratelimit_whitelist_sysmeta(self):
@@ -278,7 +279,7 @@ class TestRateLimit(unittest.TestCase):
                      'account_whitelist': 'a',
                      'account_blacklist': 'b'}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
-        req = Request.blank('/v/a/c')
+        req = Request.blank('/v1/a/c')
         req.environ['swift.cache'] = FakeMemcache()
 
         class rate_caller(threading.Thread):
@@ -306,7 +307,7 @@ class TestRateLimit(unittest.TestCase):
                 thread.join()
             the_498s = [
                 t for t in threads
-                if ''.join(t.result).startswith('Slow down')]
+                if b''.join(t.result).startswith(b'Slow down')]
             self.assertEqual(len(the_498s), 0)
             self.assertEqual(time_ticker, 0)
 
@@ -320,7 +321,7 @@ class TestRateLimit(unittest.TestCase):
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
         self.test_ratelimit.logger = FakeLogger()
         self.test_ratelimit.BLACK_LIST_SLEEP = 0
-        req = Request.blank('/v/b/c')
+        req = Request.blank('/v1/b/c')
         req.environ['swift.cache'] = FakeMemcache()
 
         class rate_caller(threading.Thread):
@@ -330,7 +331,7 @@ class TestRateLimit(unittest.TestCase):
                 self.parent = parent
 
             def run(self):
-                self.result = self.parent.test_ratelimit(req.environ,
+                self.result = self.parent.test_ratelimit(req.environ.copy(),
                                                          start_response)
 
         def get_fake_ratelimit(*args, **kwargs):
@@ -348,7 +349,7 @@ class TestRateLimit(unittest.TestCase):
                 thread.join()
             the_497s = [
                 t for t in threads
-                if ''.join(t.result).startswith('Your account')]
+                if b''.join(t.result).startswith(b'Your account')]
             self.assertEqual(len(the_497s), 5)
             self.assertEqual(time_ticker, 0)
 
@@ -361,7 +362,7 @@ class TestRateLimit(unittest.TestCase):
                      'max_sleep_time_seconds': 1}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
         self.test_ratelimit.log_sleep_time_seconds = .00001
-        req = Request.blank('/v/a/c')
+        req = Request.blank('/v1/a/c')
         req.method = 'PUT'
         req.environ['swift.cache'] = FakeMemcache()
 
@@ -369,18 +370,18 @@ class TestRateLimit(unittest.TestCase):
         # simulates 4 requests coming in at same time, then sleeping
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
                         lambda *args, **kwargs: {}):
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], '204 No Content')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Some Content')
 
     def test_ratelimit_max_rate_double_container(self):
         global time_ticker
@@ -391,7 +392,7 @@ class TestRateLimit(unittest.TestCase):
                      'max_sleep_time_seconds': 1}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
         self.test_ratelimit.log_sleep_time_seconds = .00001
-        req = Request.blank('/v/a/c/o')
+        req = Request.blank('/v1/a/c/o')
         req.method = 'PUT'
         req.environ['swift.cache'] = FakeMemcache()
         req.environ['swift.cache'].set(
@@ -402,18 +403,18 @@ class TestRateLimit(unittest.TestCase):
         # simulates 4 requests coming in at same time, then sleeping
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
                         lambda *args, **kwargs: {}):
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], '204 No Content')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Some Content')
 
     def test_ratelimit_max_rate_double_container_listing(self):
         global time_ticker
@@ -424,7 +425,7 @@ class TestRateLimit(unittest.TestCase):
                      'max_sleep_time_seconds': 1}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
         self.test_ratelimit.log_sleep_time_seconds = .00001
-        req = Request.blank('/v/a/c')
+        req = Request.blank('/v1/a/c')
         req.method = 'GET'
         req.environ['swift.cache'] = FakeMemcache()
         req.environ['swift.cache'].set(
@@ -435,18 +436,18 @@ class TestRateLimit(unittest.TestCase):
                         lambda *args, **kwargs: {}):
             time_override = [0, 0, 0, 0, None]
             # simulates 4 requests coming in at same time, then sleeping
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
+            r = self.test_ratelimit(req.environ.copy(), start_response)
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], 'Slow down')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Slow down')
             mock_sleep(.1)
-            r = self.test_ratelimit(req.environ, start_response)
-            self.assertEqual(r[0], '204 No Content')
+            r = self.test_ratelimit(req.environ.copy(), start_response)
+            self.assertEqual(r[0], b'Some Content')
             mc = self.test_ratelimit.memcache_client
             try:
                 self.test_ratelimit.memcache_client = None
@@ -464,9 +465,6 @@ class TestRateLimit(unittest.TestCase):
 
         the_app = ratelimit.filter_factory(conf_dict)(FakeApp())
         the_app.memcache_client = fake_memcache
-        req = lambda: None
-        req.method = 'PUT'
-        req.environ = {}
 
         class rate_caller(threading.Thread):
 
@@ -476,8 +474,8 @@ class TestRateLimit(unittest.TestCase):
 
             def run(self):
                 for j in range(num_calls):
-                    self.result = the_app.handle_ratelimit(req, self.myname,
-                                                           'c', None)
+                    self.result = the_app.handle_ratelimit(
+                        FakeReq('PUT'), self.myname, 'c', None)
 
         with mock.patch('swift.common.middleware.ratelimit.get_account_info',
                         lambda *args, **kwargs: {}):
@@ -511,14 +509,38 @@ class TestRateLimit(unittest.TestCase):
             def __call__(self, *args, **kwargs):
                 pass
         resp = rate_mid.__call__(env, a_callable())
-        self.assertTrue('fake_app' == resp[0])
+        self.assertEqual('fake_app', resp[0])
+
+    def test_call_non_swift_api_path(self):
+        env = {'REQUEST_METHOD': 'GET',
+               'SCRIPT_NAME': '',
+               'PATH_INFO': '/ive/got/a/lovely/bunch/of/coconuts',
+               'SERVER_NAME': '127.0.0.1',
+               'SERVER_PORT': '80',
+               'swift.cache': FakeMemcache(),
+               'SERVER_PROTOCOL': 'HTTP/1.0'}
+
+        app = lambda *args, **kwargs: ['some response']
+        rate_mid = ratelimit.filter_factory({})(app)
+
+        class a_callable(object):
+
+            def __call__(self, *args, **kwargs):
+                pass
+
+        with mock.patch('swift.common.middleware.ratelimit.get_account_info',
+                        side_effect=Exception("you shouldn't call this")):
+            resp = rate_mid(env, a_callable())
+        self.assertEqual(resp[0], 'some response')
 
     def test_no_memcache(self):
         current_rate = 13
         num_calls = 5
         conf_dict = {'account_ratelimit': current_rate}
-        self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
-        req = Request.blank('/v/a')
+        fake_app = FakeApp()
+        fake_app.skip_handled_check = True
+        self.test_ratelimit = ratelimit.filter_factory(conf_dict)(fake_app)
+        req = Request.blank('/v1/a')
         req.environ['swift.cache'] = None
         make_app_call = lambda: self.test_ratelimit(req.environ,
                                                     start_response)
@@ -527,12 +549,30 @@ class TestRateLimit(unittest.TestCase):
         time_took = time.time() - begin
         self.assertEqual(round(time_took, 1), 0)  # no memcache, no limiting
 
+    def test_already_handled(self):
+        current_rate = 13
+        num_calls = 5
+        conf_dict = {'container_listing_ratelimit_0': current_rate}
+        self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
+        fake_cache = FakeMemcache()
+        fake_cache.set(
+            get_cache_key('a', 'c'),
+            {'object_count': 1})
+        req = Request.blank('/v1/a/c', environ={'swift.cache': fake_cache})
+        req.environ['swift.ratelimit.handled'] = True
+        make_app_call = lambda: self.test_ratelimit(req.environ,
+                                                    start_response)
+        begin = time.time()
+        self._run(make_app_call, num_calls, current_rate, check_time=False)
+        time_took = time.time() - begin
+        self.assertEqual(round(time_took, 1), 0)  # no memcache, no limiting
+
     def test_restarting_memcache(self):
         current_rate = 2
         num_calls = 5
         conf_dict = {'account_ratelimit': current_rate}
         self.test_ratelimit = ratelimit.filter_factory(conf_dict)(FakeApp())
-        req = Request.blank('/v/a/c')
+        req = Request.blank('/v1/a/c')
         req.method = 'PUT'
         req.environ['swift.cache'] = FakeMemcache()
         req.environ['swift.cache'].error_on_incr = True
@@ -553,11 +593,11 @@ class TestSwiftInfo(unittest.TestCase):
 
     def test_registered_defaults(self):
 
-        def check_key_is_absnet(key):
+        def check_key_is_absent(key):
             try:
                 swift_info[key]
             except KeyError as err:
-                if key not in err:
+                if key not in str(err):
                     raise
 
         test_limits = {'account_ratelimit': 1,
@@ -571,7 +611,7 @@ class TestSwiftInfo(unittest.TestCase):
 
         ratelimit.filter_factory(test_limits)('have to pass in an app')
         swift_info = utils.get_swift_info()
-        self.assertTrue('ratelimit' in swift_info)
+        self.assertIn('ratelimit', swift_info)
         self.assertEqual(swift_info['ratelimit']
                          ['account_ratelimit'], 1.0)
         self.assertEqual(swift_info['ratelimit']
@@ -605,7 +645,7 @@ class TestSwiftInfo(unittest.TestCase):
         for key in ['log_sleep_time_seconds', 'clock_accuracy',
                     'rate_buffer_seconds', 'ratelimit_whitelis',
                     'ratelimit_blacklist']:
-            check_key_is_absnet(key)
+            check_key_is_absent(key)
 
 
 if __name__ == '__main__':
