@@ -16,6 +16,7 @@
 import array
 import collections
 import six.moves.cPickle as pickle
+import hashlib
 import os
 import unittest
 import stat
@@ -63,6 +64,8 @@ class TestRingData(unittest.TestCase):
                          rd_got._replica2part2dev_id)
         self.assertEqual(rd_expected.devs, rd_got.devs)
         self.assertEqual(rd_expected._part_shift, rd_got._part_shift)
+        self.assertEqual(rd_expected.next_part_power, rd_got.next_part_power)
+        self.assertEqual(rd_expected.version, rd_got.version)
 
     def test_attrs(self):
         r2p2d = [[0, 1, 0, 1], [0, 1, 0, 1]]
@@ -230,11 +233,22 @@ class TestRing(TestRingBase):
         self.assertEqual(self.ring.devs, self.intended_devs)
         self.assertEqual(self.ring.reload_time, self.intended_reload_time)
         self.assertEqual(self.ring.serialized_path, self.testgz)
+        self.assertIsNone(self.ring.version)
+
+        with open(self.testgz, 'rb') as fp:
+            expected_md5 = hashlib.md5()
+            expected_size = 0
+            for chunk in iter(lambda: fp.read(2 ** 16), b''):
+                expected_md5.update(chunk)
+                expected_size += len(chunk)
+        self.assertEqual(self.ring.md5, expected_md5.hexdigest())
+        self.assertEqual(self.ring.size, expected_size)
+
         # test invalid endcap
         with mock.patch.object(utils, 'HASH_PATH_SUFFIX', b''), \
                 mock.patch.object(utils, 'HASH_PATH_PREFIX', b''), \
                 mock.patch.object(utils, 'SWIFT_CONF_FILE', ''):
-            self.assertRaises(SystemExit, ring.Ring, self.testdir, 'whatever')
+            self.assertRaises(IOError, ring.Ring, self.testdir, 'whatever')
 
     def test_replica_count(self):
         self.assertEqual(self.ring.replica_count, 3)
@@ -545,7 +559,8 @@ class TestRing(TestRingBase):
                                 'ip': '1.2.%d.%d' % (zone, server),
                                 'port': 1234 + device,
                                 'zone': zone, 'region': 0,
-                                'weight': 1.0})
+                                'weight': 1.0,
+                                'device': "d%s" % device})
                     next_dev_id += 1
         rb.rebalance(seed=2)
         rb.get_ring().save(self.testgz)
@@ -568,6 +583,10 @@ class TestRing(TestRingBase):
         self.assertEqual(len(devs), len(exp_handoffs))
         dev_ids = [d['id'] for d in devs]
         self.assertEqual(dev_ids, exp_handoffs)
+        # We mark handoffs so code consuming extra nodes can reason about how
+        # far they've gone
+        for i, d in enumerate(devs):
+            self.assertEqual(d['handoff_index'], i)
 
         # The first 6 replicas plus the 3 primary nodes should cover all 9
         # zones in this test
@@ -586,7 +605,8 @@ class TestRing(TestRingBase):
         server = 0
         rb.add_dev({'id': next_dev_id,
                     'ip': '1.2.%d.%d' % (zone, server),
-                    'port': 1234, 'zone': zone, 'region': 0, 'weight': 1.0})
+                    'port': 1234, 'zone': zone, 'region': 0, 'weight': 1.0,
+                    'device': 'xd0'})
         next_dev_id += 1
         rb.pretend_min_part_hours_passed()
         num_parts_changed, _balance, _removed_dev = rb.rebalance(seed=2)
@@ -818,7 +838,8 @@ class TestRing(TestRingBase):
                         # 108.0 is the weight of all devices created prior to
                         # this test in region 0; this way all regions have
                         # equal combined weight
-                        'zone': 1, 'region': region, 'weight': 108.0})
+                        'zone': 1, 'region': region, 'weight': 108.0,
+                        'device': 'sdx'})
             next_dev_id += 1
         rb.pretend_min_part_hours_passed()
         rb.rebalance(seed=1)
@@ -896,6 +917,7 @@ class TestRing(TestRingBase):
         rb.rebalance()
         rb.get_ring().save(self.testgz)
         r = ring.Ring(self.testdir, ring_name='whatever')
+        self.assertEqual(r.version, rb.version)
 
         class CountingRingTable(object):
 
