@@ -29,6 +29,7 @@ import json
 
 import mock
 from mock import patch, call
+import six
 from six.moves import reload_module
 
 from swift.container.backend import DATADIR
@@ -159,6 +160,8 @@ def _mock_process(*args):
 
 class ReplHttp(object):
     def __init__(self, response=None, set_status=200):
+        if isinstance(response, six.text_type):
+            response = response.encode('ascii')
         self.response = response
         self.set_status = set_status
     replicated = False
@@ -321,6 +324,11 @@ class TestDBReplicator(unittest.TestCase):
         # later config should be extended to assert more config options
         replicator = TestReplicator({'node_timeout': '3.5'})
         self.assertEqual(replicator.node_timeout, 3.5)
+        self.assertEqual(replicator.databases_per_second, 50.0)
+
+        replicator = TestReplicator({'databases_per_second': '0.1'})
+        self.assertEqual(replicator.node_timeout, 10)
+        self.assertEqual(replicator.databases_per_second, 0.1)
 
     def test_repl_connection(self):
         node = {'replication_ip': '127.0.0.1', 'replication_port': 80,
@@ -688,7 +696,7 @@ class TestDBReplicator(unittest.TestCase):
         class FakeResponse(object):
             def __init__(self, status, rinfo):
                 self._status = status
-                self.data = json.dumps(rinfo)
+                self.data = json.dumps(rinfo).encode('ascii')
 
             @property
             def status(self):
@@ -1120,7 +1128,7 @@ class TestDBReplicator(unittest.TestCase):
                                           mount_check=False)
         with unit.mock_check_drive(isdir=True):
             response = rpc.dispatch(('a',), 'arg')
-        self.assertEqual('Invalid object type', response.body)
+        self.assertEqual(b'Invalid object type', response.body)
         self.assertEqual(400, response.status_int)
 
     def test_dispatch_drive_not_mounted(self):
@@ -1819,7 +1827,7 @@ class TestHandoffsOnly(unittest.TestCase):
         # part 2 does not belong on sdq
         self._make_fake_db('sdq', 2, 'bcbcbcbc15d3835053d568c57e2c83b5')
 
-    def cleanUp(self):
+    def tearDown(self):
         rmtree(self.root, ignore_errors=True)
 
     def test_scary_warnings(self):
