@@ -13,6 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import io
 import json
 import unittest
 import os
@@ -22,7 +23,6 @@ import time
 import pickle
 
 import mock
-import six
 from six.moves import urllib
 
 from swift.common import direct_client
@@ -72,7 +72,7 @@ class FakeConn(object):
         return self.resp_headers.items()
 
     def read(self, amt=None):
-        if isinstance(self.body, six.StringIO):
+        if isinstance(self.body, io.BytesIO):
             return self.body.read(amt)
         elif amt is None:
             return self.body
@@ -95,11 +95,6 @@ def mocked_http_conn(*args, **kwargs):
         yield fake_conn
 
 
-@contextmanager
-def noop_timeout(duration):
-    yield
-
-
 @patch_policies
 class TestDirectClient(unittest.TestCase):
 
@@ -122,52 +117,91 @@ class TestDirectClient(unittest.TestCase):
                 self.account, self.container, self.obj))
         self.user_agent = 'direct-client %s' % os.getpid()
 
-        patcher = mock.patch.object(direct_client, 'Timeout', noop_timeout)
+        class FakeTimeout(BaseException):
+            def __enter__(self):
+                return self
+
+            def __exit__(self, typ, value, tb):
+                pass
+
+        patcher = mock.patch.object(direct_client, 'Timeout', FakeTimeout)
         patcher.start()
         self.addCleanup(patcher.stop)
 
     def test_gen_headers(self):
         stub_user_agent = 'direct-client %s' % os.getpid()
 
-        headers = direct_client.gen_headers()
-        self.assertEqual(headers['user-agent'], stub_user_agent)
-        self.assertEqual(1, len(headers))
+        headers = direct_client.gen_headers(add_ts=False)
+        self.assertEqual(dict(headers), {
+            'User-Agent': stub_user_agent,
+            'X-Backend-Allow-Reserved-Names': 'true',
+        })
 
-        now = time.time()
-        headers = direct_client.gen_headers(add_ts=True)
-        self.assertEqual(headers['user-agent'], stub_user_agent)
-        self.assertTrue(now - 1 < Timestamp(headers['x-timestamp']) < now + 1)
-        self.assertEqual(headers['x-timestamp'],
-                         Timestamp(headers['x-timestamp']).internal)
-        self.assertEqual(2, len(headers))
+        with mock.patch('swift.common.utils.Timestamp.now',
+                        return_value=Timestamp('123.45')):
+            headers = direct_client.gen_headers()
+        self.assertEqual(dict(headers), {
+            'User-Agent': stub_user_agent,
+            'X-Backend-Allow-Reserved-Names': 'true',
+            'X-Timestamp': '0000000123.45000',
+        })
 
-        headers = direct_client.gen_headers(hdrs_in={'foo-bar': '47'})
-        self.assertEqual(headers['user-agent'], stub_user_agent)
-        self.assertEqual(headers['foo-bar'], '47')
-        self.assertEqual(2, len(headers))
+        headers = direct_client.gen_headers(hdrs_in={'x-timestamp': '15'})
+        self.assertEqual(dict(headers), {
+            'User-Agent': stub_user_agent,
+            'X-Backend-Allow-Reserved-Names': 'true',
+            'X-Timestamp': '15',
+        })
 
-        headers = direct_client.gen_headers(hdrs_in={'user-agent': '47'})
-        self.assertEqual(headers['user-agent'], stub_user_agent)
-        self.assertEqual(1, len(headers))
+        with mock.patch('swift.common.utils.Timestamp.now',
+                        return_value=Timestamp('12345.6789')):
+            headers = direct_client.gen_headers(hdrs_in={'foo-bar': '63'})
+        self.assertEqual(dict(headers), {
+            'User-Agent': stub_user_agent,
+            'Foo-Bar': '63',
+            'X-Backend-Allow-Reserved-Names': 'true',
+            'X-Timestamp': '0000012345.67890',
+        })
+
+        hdrs_in = {'foo-bar': '55'}
+        headers = direct_client.gen_headers(hdrs_in, add_ts=False)
+        self.assertEqual(dict(headers), {
+            'User-Agent': stub_user_agent,
+            'Foo-Bar': '55',
+            'X-Backend-Allow-Reserved-Names': 'true',
+        })
+
+        with mock.patch('swift.common.utils.Timestamp.now',
+                        return_value=Timestamp('12345')):
+            headers = direct_client.gen_headers(hdrs_in={'user-agent': '32'})
+        self.assertEqual(dict(headers), {
+            'User-Agent': '32',
+            'X-Backend-Allow-Reserved-Names': 'true',
+            'X-Timestamp': '0000012345.00000',
+        })
+
+        hdrs_in = {'user-agent': '47'}
+        headers = direct_client.gen_headers(hdrs_in, add_ts=False)
+        self.assertEqual(dict(headers), {
+            'User-Agent': '47',
+            'X-Backend-Allow-Reserved-Names': 'true',
+        })
 
         for policy in POLICIES:
             for add_ts in (True, False):
-                now = time.time()
-                headers = direct_client.gen_headers(
-                    {'X-Backend-Storage-Policy-Index': policy.idx},
-                    add_ts=add_ts)
-                self.assertEqual(headers['user-agent'], stub_user_agent)
-                self.assertEqual(headers['X-Backend-Storage-Policy-Index'],
-                                 str(policy.idx))
-                expected_header_count = 2
+                with mock.patch('swift.common.utils.Timestamp.now',
+                                return_value=Timestamp('123456789')):
+                    headers = direct_client.gen_headers(
+                        {'X-Backend-Storage-Policy-Index': policy.idx},
+                        add_ts=add_ts)
+                expected = {
+                    'User-Agent': stub_user_agent,
+                    'X-Backend-Storage-Policy-Index': str(policy.idx),
+                    'X-Backend-Allow-Reserved-Names': 'true',
+                }
                 if add_ts:
-                    expected_header_count += 1
-                    self.assertEqual(
-                        headers['x-timestamp'],
-                        Timestamp(headers['x-timestamp']).internal)
-                    self.assertTrue(
-                        now - 1 < Timestamp(headers['x-timestamp']) < now + 1)
-                self.assertEqual(expected_header_count, len(headers))
+                    expected['X-Timestamp'] = '0123456789.00000'
+                self.assertEqual(dict(headers), expected)
 
     def test_direct_get_account(self):
         def do_test(req_params):
@@ -178,7 +212,7 @@ class TestDirectClient(unittest.TestCase):
                 'X-Timestamp': '1234567890',
                 'X-PUT-Timestamp': '1234567890'})
 
-            body = '[{"count": 1, "bytes": 20971520, "name": "c1"}]'
+            body = b'[{"count": 1, "bytes": 20971520, "name": "c1"}]'
 
             with mocked_http_conn(200, stub_headers, body) as conn:
                 resp_headers, resp = direct_client.direct_get_account(
@@ -216,28 +250,23 @@ class TestDirectClient(unittest.TestCase):
         stub_headers = {'X-Trans-Id': 'txb5f59485c578460f8be9e-0053478d09'}
         body = 'a server error has occurred'
         with mocked_http_conn(500, stub_headers, body):
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_get_account(self.node, self.part,
                                                  self.account)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
-        self.assertEqual(err.http_status, 500)
+        self.assertEqual(raised.exception.http_status, 500)
         expected_err_msg_parts = (
             'Account server %s:%s' % (self.node['ip'], self.node['port']),
             'GET %r' % self.account_path,
             'status 500',
         )
         for item in expected_err_msg_parts:
-            self.assertTrue(
-                item in str(err), '%r was not in "%s"' % (item, err))
-        self.assertEqual(err.http_host, self.node['ip'])
-        self.assertEqual(err.http_port, self.node['port'])
-        self.assertEqual(err.http_device, self.node['device'])
-        self.assertEqual(err.http_status, 500)
-        self.assertEqual(err.http_reason, 'Internal Error')
-        self.assertEqual(err.http_headers, stub_headers)
+            self.assertIn(item, str(raised.exception))
+        self.assertEqual(raised.exception.http_host, self.node['ip'])
+        self.assertEqual(raised.exception.http_port, self.node['port'])
+        self.assertEqual(raised.exception.http_device, self.node['device'])
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertEqual(raised.exception.http_reason, 'Internal Error')
+        self.assertEqual(raised.exception.http_headers, stub_headers)
 
     def test_direct_get_account_no_content_does_not_parse_body(self):
         headers = {
@@ -258,17 +287,13 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_get_account_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_get_account(
                     self.node, self.part, self.account)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.method, 'GET')
             self.assertEqual(conn.path, self.account_path)
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('GET' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('GET' in str(raised.exception))
 
     def test_direct_delete_account(self):
         part = '0'
@@ -291,22 +316,45 @@ class TestDirectClient(unittest.TestCase):
             self.assertIn('X-Timestamp', headers)
             self.assertIn('User-Agent', headers)
 
+    def test_direct_delete_account_replication_net(self):
+        part = '0'
+        account = 'a'
+
+        mock_path = 'swift.common.bufferedhttp.http_connect_raw'
+        with mock.patch(mock_path) as fake_connect:
+            fake_connect.return_value.getresponse.return_value.status = 200
+            direct_client.direct_delete_account(
+                self.node, part, account,
+                headers={'X-Backend-Use-Replication-Network': 't'})
+            args, kwargs = fake_connect.call_args
+            ip = args[0]
+            self.assertEqual(self.node['replication_ip'], ip)
+            self.assertNotEqual(self.node['ip'], ip)
+            port = args[1]
+            self.assertEqual(self.node['replication_port'], port)
+            self.assertNotEqual(self.node['port'], port)
+            method = args[2]
+            self.assertEqual('DELETE', method)
+            path = args[3]
+            self.assertEqual('/sda/0/a', path)
+            headers = args[4]
+            self.assertIn('X-Timestamp', headers)
+            self.assertIn('User-Agent', headers)
+
     def test_direct_delete_account_failure(self):
         part = '0'
         account = 'a'
 
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_delete_account(self.node, part, account)
-            except ClientException as err:
-                pass
             self.assertEqual(self.node['ip'], conn.host)
             self.assertEqual(self.node['port'], conn.port)
             self.assertEqual('DELETE', conn.method)
             self.assertEqual('/sda/0/a', conn.path)
             self.assertIn('X-Timestamp', conn.req_headers)
             self.assertIn('User-Agent', conn.req_headers)
-            self.assertEqual(err.http_status, 500)
+            self.assertEqual(raised.exception.http_status, 500)
 
     def test_direct_head_container(self):
         headers = HeaderKeyDict(key='value')
@@ -323,17 +371,31 @@ class TestDirectClient(unittest.TestCase):
                          self.user_agent)
         self.assertEqual(headers, resp)
 
+    def test_direct_head_container_replication_net(self):
+        headers = HeaderKeyDict(key='value')
+
+        with mocked_http_conn(200, headers) as conn:
+            resp = direct_client.direct_head_container(
+                self.node, self.part, self.account, self.container,
+                headers={'X-Backend-Use-Replication-Network': 'on'})
+            self.assertEqual(conn.host, self.node['replication_ip'])
+            self.assertEqual(conn.port, self.node['replication_port'])
+            self.assertNotEqual(conn.host, self.node['ip'])
+            self.assertNotEqual(conn.port, self.node['port'])
+            self.assertEqual(conn.method, 'HEAD')
+            self.assertEqual(conn.path, self.container_path)
+
+        self.assertEqual(conn.req_headers['user-agent'],
+                         self.user_agent)
+        self.assertEqual(headers, resp)
+
     def test_direct_head_container_error(self):
         headers = HeaderKeyDict(key='value')
 
         with mocked_http_conn(503, headers) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_head_container(
                     self.node, self.part, self.account, self.container)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             # check request
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
@@ -341,9 +403,9 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.container_path)
 
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
-        self.assertEqual(err.http_status, 503)
-        self.assertEqual(err.http_headers, headers)
-        self.assertTrue('HEAD' in str(err))
+        self.assertEqual(raised.exception.http_status, 503)
+        self.assertEqual(raised.exception.http_headers, headers)
+        self.assertTrue('HEAD' in str(raised.exception))
 
     def test_direct_head_container_deleted(self):
         important_timestamp = Timestamp.now().internal
@@ -351,27 +413,23 @@ class TestDirectClient(unittest.TestCase):
                                  important_timestamp})
 
         with mocked_http_conn(404, headers) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_head_container(
                     self.node, self.part, self.account, self.container)
-            except Exception as err:
-                self.assertTrue(isinstance(err, ClientException))
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.container_path)
 
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
-        self.assertEqual(err.http_status, 404)
-        self.assertEqual(err.http_headers, headers)
+        self.assertEqual(raised.exception.http_status, 404)
+        self.assertEqual(raised.exception.http_headers, headers)
 
     def test_direct_get_container(self):
         def do_test(req_params):
             headers = HeaderKeyDict({'key': 'value'})
-            body = ('[{"hash": "8f4e3", "last_modified": "317260", '
-                    '"bytes": 209}]')
+            body = (b'[{"hash": "8f4e3", "last_modified": "317260", '
+                    b'"bytes": 209}]')
 
             with mocked_http_conn(200, headers, body) as conn:
                 resp_headers, resp = direct_client.direct_get_container(
@@ -426,6 +484,18 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
 
+    def test_direct_delete_container_replication_net(self):
+        with mocked_http_conn(200) as conn:
+            direct_client.direct_delete_container(
+                self.node, self.part, self.account, self.container,
+                headers={'X-Backend-Use-Replication-Network': '1'})
+            self.assertEqual(conn.host, self.node['replication_ip'])
+            self.assertEqual(conn.port, self.node['replication_port'])
+            self.assertNotEqual(conn.host, self.node['ip'])
+            self.assertNotEqual(conn.port, self.node['port'])
+            self.assertEqual(conn.method, 'DELETE')
+            self.assertEqual(conn.path, self.container_path)
+
     def test_direct_delete_container_with_timestamp(self):
         # ensure timestamp is different from any that might be auto-generated
         timestamp = Timestamp(time.time() - 100)
@@ -443,24 +513,20 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_delete_container_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_delete_container(
                     self.node, self.part, self.account, self.container)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
 
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('DELETE' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('DELETE' in str(raised.exception))
 
     def test_direct_put_container(self):
-        body = 'Let us begin with a quick introduction'
+        body = b'Let us begin with a quick introduction'
         headers = {'x-foo': 'bar', 'Content-Length': str(len(body)),
                    'Content-Type': 'application/json',
                    'User-Agent': 'my UA'}
@@ -484,7 +550,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertIsNone(rv)
 
     def test_direct_put_container_chunked(self):
-        body = 'Let us begin with a quick introduction'
+        body = b'Let us begin with a quick introduction'
         headers = {'x-foo': 'bar', 'Content-Type': 'application/json'}
 
         with mocked_http_conn(204) as conn:
@@ -501,7 +567,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertTrue('x-timestamp' in conn.req_headers)
             self.assertEqual('bar', conn.req_headers.get('x-foo'))
             self.assertNotIn('Content-Length', conn.req_headers)
-            expected_sent = '%0x\r\n%s\r\n0\r\n\r\n' % (len(body), body)
+            expected_sent = b'%0x\r\n%s\r\n0\r\n\r\n' % (len(body), body)
             self.assertEqual(md5(expected_sent).hexdigest(),
                              conn.etag.hexdigest())
         self.assertIsNone(rv)
@@ -538,22 +604,18 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_put_container_object_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_put_container_object(
                     self.node, self.part, self.account, self.container,
                     self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
 
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('PUT' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('PUT' in str(raised.exception))
 
     def test_direct_delete_container_object(self):
         with mocked_http_conn(204) as conn:
@@ -568,22 +630,18 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_delete_container_obj_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_delete_container_object(
                     self.node, self.part, self.account, self.container,
                     self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
 
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('DELETE' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('DELETE' in str(raised.exception))
 
     def test_direct_head_object(self):
         headers = HeaderKeyDict({'x-foo': 'bar'})
@@ -599,50 +657,43 @@ class TestDirectClient(unittest.TestCase):
 
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
         self.assertEqual('bar', conn.req_headers.get('x-foo'))
-        self.assertNotIn('x-timestamp', conn.req_headers)
+        self.assertIn('x-timestamp', conn.req_headers)
         self.assertEqual(headers, resp)
 
     def test_direct_head_object_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_head_object(
                     self.node, self.part, self.account, self.container,
                     self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('HEAD' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('HEAD' in str(raised.exception))
 
     def test_direct_head_object_not_found(self):
         important_timestamp = Timestamp.now().internal
         stub_headers = {'X-Backend-Important-Timestamp': important_timestamp}
         with mocked_http_conn(404, headers=stub_headers) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_head_object(
                     self.node, self.part, self.account, self.container,
                     self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(err.http_status, 404)
-        self.assertEqual(err.http_headers['x-backend-important-timestamp'],
-                         important_timestamp)
+        self.assertEqual(raised.exception.http_status, 404)
+        self.assertEqual(
+            raised.exception.http_headers['x-backend-important-timestamp'],
+            important_timestamp)
 
     def test_direct_get_object(self):
-        contents = six.StringIO('123456')
+        contents = io.BytesIO(b'123456')
 
         with mocked_http_conn(200, body=contents) as conn:
             resp_header, obj_body = direct_client.direct_get_object(
@@ -655,41 +706,30 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_get_object_error(self):
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_get_object(
                     self.node, self.part,
                     self.account, self.container, self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'GET')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('GET' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('GET' in str(raised.exception))
 
     def test_direct_get_object_chunks(self):
-        contents = six.StringIO('123456')
-        downloaded = b''
+        contents = io.BytesIO(b'123456')
 
         with mocked_http_conn(200, body=contents) as conn:
             resp_header, obj_body = direct_client.direct_get_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 resp_chunk_size=2)
-            while obj_body:
-                try:
-                    chunk = obj_body.next()
-                except StopIteration:
-                    break
-                downloaded += chunk
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual('GET', conn.method)
             self.assertEqual(self.obj_path, conn.path)
-        self.assertEqual('123456', downloaded)
+            self.assertEqual([b'12', b'34', b'56'], list(obj_body))
 
     def test_direct_post_object(self):
         headers = {'Key': 'value'}
@@ -712,14 +752,10 @@ class TestDirectClient(unittest.TestCase):
         headers = {'Key': 'value'}
 
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_post_object(
                     self.node, self.part, self.account, self.container,
                     self.obj, headers)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'POST')
@@ -729,8 +765,8 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
             self.assertTrue('x-timestamp' in conn.req_headers)
 
-        self.assertEqual(err.http_status, 500)
-        self.assertTrue('POST' in str(err))
+        self.assertEqual(raised.exception.http_status, 500)
+        self.assertTrue('POST' in str(raised.exception))
 
     def test_direct_delete_object(self):
         with mocked_http_conn(200) as conn:
@@ -759,18 +795,14 @@ class TestDirectClient(unittest.TestCase):
 
     def test_direct_delete_object_error(self):
         with mocked_http_conn(503) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_delete_object(
                     self.node, self.part, self.account, self.container,
                     self.obj)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
-        self.assertEqual(err.http_status, 503)
-        self.assertTrue('DELETE' in str(err))
+        self.assertEqual(raised.exception.http_status, 503)
+        self.assertTrue('DELETE' in str(raised.exception))
 
     def test_direct_get_suffix_hashes(self):
         data = {'a83': 'c130a2c17ed45102aada0f4eee69494ff'}
@@ -789,12 +821,12 @@ class TestDirectClient(unittest.TestCase):
             with self.assertRaises(DirectClientException) as cm:
                 direct_client.direct_get_suffix_hashes(
                     self.node, self.part, ['a83', 'b52'])
-        self.assertIn('REPLICATE', cm.exception.message)
+        self.assertIn('REPLICATE', cm.exception.args[0])
         self.assertIn(quote('/%s/%s/a83-b52'
                             % (self.node['device'], self.part)),
-                      cm.exception.message)
-        self.assertIn(self.node['replication_ip'], cm.exception.message)
-        self.assertIn(self.node['replication_port'], cm.exception.message)
+                      cm.exception.args[0])
+        self.assertIn(self.node['replication_ip'], cm.exception.args[0])
+        self.assertIn(self.node['replication_port'], cm.exception.args[0])
         self.assertEqual(self.node['replication_ip'], cm.exception.http_host)
         self.assertEqual(self.node['replication_port'], cm.exception.http_port)
         self.assertEqual(self.node['device'], cm.exception.http_device)
@@ -807,7 +839,7 @@ class TestDirectClient(unittest.TestCase):
         self._test_direct_get_suffix_hashes_fail(507)
 
     def test_direct_put_object_with_content_length(self):
-        contents = six.StringIO('123456')
+        contents = io.BytesIO(b'123456')
 
         with mocked_http_conn(200) as conn:
             resp = direct_client.direct_put_object(
@@ -817,28 +849,24 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
-        self.assertEqual(md5('123456').hexdigest(), resp)
+        self.assertEqual(md5(b'123456').hexdigest(), resp)
 
     def test_direct_put_object_fail(self):
-        contents = six.StringIO('123456')
+        contents = io.BytesIO(b'123456')
 
         with mocked_http_conn(500) as conn:
-            try:
+            with self.assertRaises(ClientException) as raised:
                 direct_client.direct_put_object(
                     self.node, self.part, self.account, self.container,
                     self.obj, contents)
-            except ClientException as err:
-                pass
-            else:
-                self.fail('ClientException not raised')
             self.assertEqual(conn.host, self.node['ip'])
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
-        self.assertEqual(err.http_status, 500)
+        self.assertEqual(raised.exception.http_status, 500)
 
     def test_direct_put_object_chunked(self):
-        contents = six.StringIO('123456')
+        contents = io.BytesIO(b'123456')
 
         with mocked_http_conn(200) as conn:
             resp = direct_client.direct_put_object(
@@ -848,7 +876,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
-        self.assertEqual(md5('6\r\n123456\r\n0\r\n\r\n').hexdigest(), resp)
+        self.assertEqual(md5(b'6\r\n123456\r\n0\r\n\r\n').hexdigest(), resp)
 
     def test_direct_put_object_args(self):
         # One test to cover all missing checks
@@ -863,10 +891,10 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(self.obj_path, conn.path)
             self.assertEqual(conn.req_headers['Content-Length'], '0')
             self.assertEqual(conn.req_headers['Content-Type'], 'Text')
-        self.assertEqual(md5('0\r\n\r\n').hexdigest(), resp)
+        self.assertEqual(md5(b'0\r\n\r\n').hexdigest(), resp)
 
     def test_direct_put_object_header_content_length(self):
-        contents = six.StringIO('123456')
+        contents = io.BytesIO(b'123456')
         stub_headers = HeaderKeyDict({
             'Content-Length': '6'})
 
@@ -878,7 +906,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.port, self.node['port'])
             self.assertEqual('PUT', conn.method)
             self.assertEqual(conn.req_headers['Content-length'], '6')
-        self.assertEqual(md5('123456').hexdigest(), resp)
+        self.assertEqual(md5(b'123456').hexdigest(), resp)
 
     def test_retry(self):
         headers = HeaderKeyDict({'key': 'value'})
@@ -905,14 +933,11 @@ class TestDirectClient(unittest.TestCase):
                                     retries=2, error_log=logger.error)
         self.assertEqual('DELETE', conn.method)
         self.assertEqual(err_ctx.exception.http_status, 500)
-        self.assertIn('DELETE', err_ctx.exception.message)
-        self.assertIn(quote('/%s/%s/%s/%s/%s'
-                            % (self.node['device'].encode('utf-8'),
-                               self.part, self.account,
-                               self.container, self.obj)),
-                      err_ctx.exception.message)
-        self.assertIn(self.node['ip'], err_ctx.exception.message)
-        self.assertIn(self.node['port'], err_ctx.exception.message)
+        self.assertIn('DELETE', err_ctx.exception.args[0])
+        self.assertIn(self.obj_path,
+                      err_ctx.exception.args[0])
+        self.assertIn(self.node['ip'], err_ctx.exception.args[0])
+        self.assertIn(self.node['port'], err_ctx.exception.args[0])
         self.assertEqual(self.node['ip'], err_ctx.exception.http_host)
         self.assertEqual(self.node['port'], err_ctx.exception.http_port)
         self.assertEqual(self.node['device'], err_ctx.exception.http_device)
@@ -952,5 +977,6 @@ class TestUTF8DirectClient(TestDirectClient):
         self.container = self.container.encode('utf-8')
         self.obj = self.obj.encode('utf-8')
 
+
 if __name__ == '__main__':
     unittest.main()
