@@ -38,6 +38,7 @@ from swift.obj.diskfile import (
 from swift.common.ring import RingData
 from swift.common import utils
 from swift.common.header_key_dict import HeaderKeyDict
+from swift.common.swob import bytes_to_wsgi
 from swift.common.utils import (
     hash_path, normalize_timestamp, mkdirs, write_pickle)
 from swift.common.storage_policy import StoragePolicy, POLICIES
@@ -68,8 +69,8 @@ _mocked_policies = [StoragePolicy(0, 'zero', False),
 class TestObjectUpdater(unittest.TestCase):
 
     def setUp(self):
-        utils.HASH_PATH_SUFFIX = 'endcap'
-        utils.HASH_PATH_PREFIX = ''
+        utils.HASH_PATH_SUFFIX = b'endcap'
+        utils.HASH_PATH_PREFIX = b''
         self.testdir = mkdtemp()
         ring_file = os.path.join(self.testdir, 'container.ring.gz')
         with closing(GzipFile(ring_file, 'wb')) as f:
@@ -489,7 +490,7 @@ class TestObjectUpdater(unittest.TestCase):
         self.assertTrue(os.path.exists(op_path))
         self.assertEqual(ou.logger.get_increment_counts(),
                          {'failures': 1, 'unlinks': 1})
-        self.assertIsNone(pickle.load(open(op_path)).get('successes'))
+        self.assertIsNone(pickle.load(open(op_path, 'rb')).get('successes'))
 
         bindsock = listen_zero()
 
@@ -498,20 +499,20 @@ class TestObjectUpdater(unittest.TestCase):
                 with Timeout(3):
                     inc = sock.makefile('rb')
                     out = sock.makefile('wb')
-                    out.write('HTTP/1.1 %d OK\r\nContent-Length: 0\r\n\r\n' %
+                    out.write(b'HTTP/1.1 %d OK\r\nContent-Length: 0\r\n\r\n' %
                               return_code)
                     out.flush()
                     self.assertEqual(inc.readline(),
-                                     'PUT /sda1/0/a/c/o HTTP/1.1\r\n')
+                                     b'PUT /sda1/0/a/c/o HTTP/1.1\r\n')
                     headers = HeaderKeyDict()
-                    line = inc.readline()
+                    line = bytes_to_wsgi(inc.readline())
                     while line and line != '\r\n':
                         headers[line.split(':')[0]] = \
                             line.split(':')[1].strip()
-                        line = inc.readline()
-                    self.assertTrue('x-container-timestamp' in headers)
-                    self.assertTrue('X-Backend-Storage-Policy-Index' in
-                                    headers)
+                        line = bytes_to_wsgi(inc.readline())
+                    self.assertIn('x-container-timestamp', headers)
+                    self.assertIn('X-Backend-Storage-Policy-Index',
+                                  headers)
             except BaseException as err:
                 return err
             return None
@@ -546,7 +547,7 @@ class TestObjectUpdater(unittest.TestCase):
         self.assertEqual(ou.logger.get_increment_counts(),
                          {'failures': 1})
         self.assertEqual([0],
-                         pickle.load(open(op_path)).get('successes'))
+                         pickle.load(open(op_path, 'rb')).get('successes'))
 
         event = spawn(accept, [404, 201])
         ou.logger._clear()
@@ -558,7 +559,7 @@ class TestObjectUpdater(unittest.TestCase):
         self.assertEqual(ou.logger.get_increment_counts(),
                          {'failures': 1})
         self.assertEqual([0, 2],
-                         pickle.load(open(op_path)).get('successes'))
+                         pickle.load(open(op_path, 'rb')).get('successes'))
 
         event = spawn(accept, [201])
         ou.logger._clear()
@@ -710,19 +711,24 @@ class TestObjectUpdater(unittest.TestCase):
             'X-Backend-Storage-Policy-Index': str(int(policies[0])),
             'User-Agent': 'object-updater %s' % os.getpid(),
             'X-Backend-Accept-Redirect': 'true',
+            'X-Backend-Accept-Quoted-Location': 'true',
         }
-        # always expect X-Backend-Accept-Redirect to be true
+        # always expect X-Backend-Accept-Redirect and
+        # X-Backend-Accept-Quoted-Location to be true
         do_test(headers_out, expected, container_path='.shards_a/shard_c')
         do_test(headers_out, expected)
 
-        # ...unless X-Backend-Accept-Redirect is already set
+        # ...unless they're already set
         expected['X-Backend-Accept-Redirect'] = 'false'
+        expected['X-Backend-Accept-Quoted-Location'] = 'false'
         headers_out_2 = dict(headers_out)
         headers_out_2['X-Backend-Accept-Redirect'] = 'false'
+        headers_out_2['X-Backend-Accept-Quoted-Location'] = 'false'
         do_test(headers_out_2, expected)
 
         # updater should add policy header if missing
         expected['X-Backend-Accept-Redirect'] = 'true'
+        expected['X-Backend-Accept-Quoted-Location'] = 'true'
         headers_out['X-Backend-Storage-Policy-Index'] = None
         do_test(headers_out, expected)
 
@@ -746,7 +752,8 @@ class TestObjectUpdater(unittest.TestCase):
             'X-Timestamp': timestamp.internal,
             'X-Backend-Storage-Policy-Index': str(int(policy)),
             'User-Agent': 'object-updater %s' % os.getpid(),
-            'X-Backend-Accept-Redirect': 'true'}
+            'X-Backend-Accept-Redirect': 'true',
+            'X-Backend-Accept-Quoted-Location': 'true'}
         for request in requests:
             self.assertEqual('PUT', request['method'])
             self.assertDictEqual(expected_headers, request['headers'])
@@ -867,7 +874,7 @@ class TestObjectUpdater(unittest.TestCase):
         self.assertEqual([mock.ANY], async_files)
         async_path = os.path.join(
             async_dir, async_subdirs[0], async_files[0])
-        with open(async_path) as fd:
+        with open(async_path, 'rb') as fd:
             async_data = pickle.load(fd)
         return async_path, async_data
 
@@ -953,9 +960,11 @@ class TestObjectUpdater(unittest.TestCase):
             # 1st round of redirects, newest redirect should be chosen
             (301, {'Location': '/.shards_a/c_shard_old/o',
                    'X-Backend-Redirect-Timestamp': ts_redirect_1.internal}),
-            (301, {'Location': '/.shards_a/c_shard_new/o',
+            (301, {'Location': '/.shards_a/c%5Fshard%5Fnew/o',
+                   'X-Backend-Location-Is-Quoted': 'true',
                    'X-Backend-Redirect-Timestamp': ts_redirect_2.internal}),
-            (301, {'Location': '/.shards_a/c_shard_old/o',
+            (301, {'Location': '/.shards_a/c%5Fshard%5Fold/o',
+                   'X-Backend-Location-Is-Quoted': 'true',
                    'X-Backend-Redirect-Timestamp': ts_redirect_1.internal}),
             # 2nd round of redirects
             (301, {'Location': '/.shards_a/c_shard_newer/o',
