@@ -19,14 +19,15 @@ import socket
 import unittest
 
 from eventlet import Timeout
+import six
 from six.moves import urllib
 
 from swift.common.constraints import CONTAINER_LISTING_LIMIT
-from swift.common.swob import Request
+from swift.common.swob import Request, bytes_to_wsgi, str_to_wsgi, wsgi_quote
 from swift.common.utils import ShardRange, Timestamp
 from swift.proxy import server as proxy_server
-from swift.proxy.controllers.base import headers_to_container_info, Controller, \
-    get_container_info
+from swift.proxy.controllers.base import headers_to_container_info, \
+    Controller, get_container_info, get_cache_key
 from test import annotate_failure
 from test.unit import fake_http_connect, FakeRing, FakeMemcache, \
     make_timestamp_iter
@@ -48,7 +49,7 @@ class TestContainerController(TestRingBase):
         self.logger = debug_logger()
         self.container_ring = FakeRing(replicas=self.CONTAINER_REPLICAS,
                                        max_more_nodes=9)
-        self.app = proxy_server.Application(None, FakeMemcache(),
+        self.app = proxy_server.Application(None,
                                             logger=self.logger,
                                             account_ring=FakeRing(),
                                             container_ring=self.container_ring)
@@ -96,7 +97,9 @@ class TestContainerController(TestRingBase):
             with mock.patch(
                     'swift.proxy.controllers.base.http_connect',
                     fake_http_connect(*responses)):
-                req = Request.blank('/v1/a/c')
+                cache = FakeMemcache()
+                cache.set(get_cache_key('a'), {'status': 204})
+                req = Request.blank('/v1/a/c', environ={'swift.cache': cache})
                 resp = getattr(controller, method)(req)
 
             self.assertEqual(expected,
@@ -105,19 +108,37 @@ class TestContainerController(TestRingBase):
                              (expected, resp.status_int, str(responses)))
 
     def test_container_info_got_cached(self):
+        memcache = FakeMemcache()
         controller = proxy_server.ContainerController(self.app, 'a', 'c')
-        with mock.patch('swift.proxy.controllers.base.http_connect',
-                        fake_http_connect(200, 200, body='')):
-            req = Request.blank('/v1/a/c', {'PATH_INFO': '/v1/a/c'})
+        with mocked_http_conn(200, 200) as mock_conn:
+            req = Request.blank('/v1/a/c', {'swift.cache': memcache})
             resp = controller.HEAD(req)
         self.assertEqual(2, resp.status_int // 100)
+        self.assertEqual(['/a', '/a/c'],
+                         # requests are like /sdX/0/..
+                         [r['path'][6:] for r in mock_conn.requests])
         # Make sure it's in both swift.infocache and memcache
+        header_info = headers_to_container_info(resp.headers)
+        info_cache = resp.environ['swift.infocache']
         self.assertIn("container/a/c", resp.environ['swift.infocache'])
-        self.assertEqual(
-            headers_to_container_info(resp.headers),
-            resp.environ['swift.infocache']['container/a/c'])
-        from_memcache = self.app.memcache.get('container/a/c')
-        self.assertTrue(from_memcache)
+        self.assertEqual(header_info, info_cache['container/a/c'])
+        self.assertEqual(header_info, memcache.get('container/a/c'))
+
+        # The failure doesn't lead to cache eviction
+        errors = [500] * self.CONTAINER_REPLICAS * 2
+        with mocked_http_conn(*errors) as mock_conn:
+            req = Request.blank('/v1/a/c', {'swift.infocache': info_cache,
+                                            'swift.cache': memcache})
+            resp = controller.HEAD(req)
+        self.assertEqual(5, resp.status_int // 100)
+        self.assertEqual(['/a/c'] * self.CONTAINER_REPLICAS * 2,
+                         # requests are like /sdX/0/..
+                         [r['path'][6:] for r in mock_conn.requests])
+        self.assertIs(info_cache, resp.environ['swift.infocache'])
+        self.assertIn("container/a/c", resp.environ['swift.infocache'])
+        # NB: this is the *old* header_info, from the good req
+        self.assertEqual(header_info, info_cache['container/a/c'])
+        self.assertEqual(header_info, memcache.get('container/a/c'))
 
     @mock.patch('swift.proxy.controllers.container.clear_info_cache')
     @mock.patch.object(Controller, 'make_requests')
@@ -344,9 +365,12 @@ class TestContainerController(TestRingBase):
             ([200], 200),
             ([404, 200], 200),
             ([404] * nodes + [200], 200),
-            ([Timeout()] * nodes + [404] * handoffs, 404),
+            ([Timeout()] * nodes + [404] * handoffs, 503),
             ([Timeout()] * (nodes + handoffs), 503),
-            ([Timeout()] * (nodes + handoffs - 1) + [404], 404),
+            ([Timeout()] * (nodes + handoffs - 1) + [404], 503),
+            ([Timeout()] * (nodes - 1) + [404] * (handoffs + 1), 503),
+            ([Timeout()] * (nodes - 2) + [404] * (handoffs + 2), 404),
+            ([500] * (nodes - 1) + [404] * (handoffs + 1), 503),
             ([503, 200], 200),
             ([507, 200], 200),
         ]
@@ -371,12 +395,28 @@ class TestContainerController(TestRingBase):
                       '\n'.join(failures))
 
         # One more test, simulating all nodes being error-limited
+        class FakeIter(object):
+            num_primary_nodes = 3
+
+            def __iter__(self):
+                return iter([])
+
         with mocked_http_conn(), mock.patch.object(self.app, 'iter_nodes',
-                                                   return_value=[]):
+                                                   return_value=FakeIter()):
             req = Request.blank('/v1/a/c')
             resp = req.get_response(self.app)
             self.assertEqual(resp.status_int, 503)
 
+    def test_handoff_has_deleted_database(self):
+        nodes = self.app.container_ring.replicas
+        handoffs = self.app.request_node_count(nodes) - nodes
+        status = [Timeout()] * nodes + [404] * handoffs
+        timestamps = tuple([None] * nodes + ['1'] + [None] * (handoffs - 1))
+        with mocked_http_conn(*status, timestamps=timestamps):
+            req = Request.blank('/v1/a/c')
+            resp = req.get_response(self.app)
+            self.assertEqual(resp.status_int, 404)
+
     def test_response_code_for_PUT(self):
         PUT_TEST_CASES = [
             ((201, 201, 201), 201),
@@ -423,13 +463,21 @@ class TestContainerController(TestRingBase):
         self._assert_responses('POST', POST_TEST_CASES)
 
     def _make_shard_objects(self, shard_range):
-        lower = ord(shard_range.lower[0]) if shard_range.lower else ord('@')
-        upper = ord(shard_range.upper[0]) if shard_range.upper else ord('z')
+        if six.PY2:
+            lower = ord(shard_range.lower.decode('utf8')[0]
+                        if shard_range.lower else '@')
+            upper = ord(shard_range.upper.decode('utf8')[0]
+                        if shard_range.upper else u'\U0001ffff')
+        else:
+            lower = ord(shard_range.lower[0] if shard_range.lower else '@')
+            upper = ord(shard_range.upper[0] if shard_range.upper
+                        else '\U0001ffff')
 
-        objects = [{'name': chr(i), 'bytes': i, 'hash': 'hash%s' % chr(i),
+        objects = [{'name': six.unichr(i), 'bytes': i,
+                    'hash': 'hash%s' % six.unichr(i),
                     'content_type': 'text/plain', 'deleted': 0,
                     'last_modified': next(self.ts_iter).isoformat}
-                   for i in range(lower + 1, upper + 1)]
+                   for i in range(lower + 1, upper + 1)][:1024]
         return objects
 
     def _check_GET_shard_listing(self, mock_responses, expected_objects,
@@ -450,7 +498,8 @@ class TestContainerController(TestRingBase):
                 self.assertLess(name(prev), name(next_))
         container_path = '/v1/a/c' + query_string
         codes = (resp[0] for resp in mock_responses)
-        bodies = iter([json.dumps(resp[1]) for resp in mock_responses])
+        bodies = iter([json.dumps(resp[1]).encode('ascii')
+                       for resp in mock_responses])
         exp_headers = [resp[2] for resp in mock_responses]
         request = Request.blank(container_path)
         with mocked_http_conn(
@@ -471,9 +520,12 @@ class TestContainerController(TestRingBase):
             with annotate_failure('Request check at index %d.' % i):
                 # strip off /sdx/0/ from path
                 self.assertEqual(exp_path, req['path'][7:])
-                self.assertEqual(
-                    dict(exp_params, format='json'),
-                    dict(urllib.parse.parse_qsl(req['qs'], True)))
+                if six.PY2:
+                    got_params = dict(urllib.parse.parse_qsl(req['qs'], True))
+                else:
+                    got_params = dict(urllib.parse.parse_qsl(
+                        req['qs'], True, encoding='latin1'))
+                self.assertEqual(dict(exp_params, format='json'), got_params)
                 for k, v in exp_headers.items():
                     self.assertIn(k, req['headers'])
                     self.assertEqual(v, req['headers'][k])
@@ -504,10 +556,11 @@ class TestContainerController(TestRingBase):
         self.assertEqual(headers_to_container_info(info_hdrs), info)
 
     def test_GET_sharded_container(self):
-        shard_bounds = (('', 'ham'), ('ham', 'pie'), ('pie', ''))
+        # Don't worry, ShardRange._encode takes care of unicode/bytes issues
+        shard_bounds = ('', 'ham', 'pie', u'\N{SNOWMAN}', u'\U0001F334', '')
         shard_ranges = [
             ShardRange('.shards_a/c_%s' % upper, Timestamp.now(), lower, upper)
-            for lower, upper in shard_bounds]
+            for lower, upper in zip(shard_bounds[:-1], shard_bounds[1:])]
         sr_dicts = [dict(sr) for sr in shard_ranges]
         sr_objs = [self._make_shard_objects(sr) for sr in shard_ranges]
         shard_resp_hdrs = [
@@ -517,7 +570,7 @@ class TestContainerController(TestRingBase):
                  sum([obj['bytes'] for obj in sr_objs[i]]),
              'X-Container-Meta-Flavour': 'flavour%d' % i,
              'X-Backend-Storage-Policy-Index': 0}
-            for i in range(3)]
+            for i, _ in enumerate(shard_ranges)]
 
         all_objects = []
         for objects in sr_objs:
@@ -543,7 +596,9 @@ class TestContainerController(TestRingBase):
             (200, sr_dicts, root_shard_resp_hdrs),
             (200, sr_objs[0], shard_resp_hdrs[0]),
             (200, sr_objs[1], shard_resp_hdrs[1]),
-            (200, sr_objs[2], shard_resp_hdrs[2])
+            (200, sr_objs[2], shard_resp_hdrs[2]),
+            (200, sr_objs[3], shard_resp_hdrs[3]),
+            (200, sr_objs[4], shard_resp_hdrs[4]),
         ]
         expected_requests = [
             # path, headers, params
@@ -551,15 +606,29 @@ class TestContainerController(TestRingBase):
              dict(states='listing')),  # 404
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(states='listing')),  # 200
-            (shard_ranges[0].name, {'X-Backend-Record-Type': 'auto'},
+            (wsgi_quote(str_to_wsgi(shard_ranges[0].name)),
+             {'X-Backend-Record-Type': 'auto'},
              dict(marker='', end_marker='ham\x00', limit=str(limit),
                   states='listing')),  # 200
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},
              dict(marker='h', end_marker='pie\x00', states='listing',
                   limit=str(limit - len(sr_objs[0])))),  # 200
-            (shard_ranges[2].name, {'X-Backend-Record-Type': 'auto'},
-             dict(marker='p', end_marker='', states='listing',
-                  limit=str(limit - len(sr_objs[0] + sr_objs[1]))))  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[2].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='p', end_marker='\xe2\x98\x83\x00', states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1])))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='\xd1\xb0', end_marker='\xf0\x9f\x8c\xb4\x00',
+                  states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1]
+                                        + sr_objs[2])))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[4].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='\xe2\xa8\x83', end_marker='', states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1] + sr_objs[2]
+                                        + sr_objs[3])))),  # 200
         ]
 
         resp = self._check_GET_shard_listing(
@@ -575,7 +644,7 @@ class TestContainerController(TestRingBase):
             (200, sr_dicts[:2] + [dict(root_range)], root_shard_resp_hdrs),
             (200, sr_objs[0], shard_resp_hdrs[0]),
             (200, sr_objs[1], shard_resp_hdrs[1]),
-            (200, sr_objs[2], root_resp_hdrs)
+            (200, sr_objs[2] + sr_objs[3] + sr_objs[4], root_resp_hdrs)
         ]
         expected_requests = [
             # path, headers, params
@@ -598,10 +667,12 @@ class TestContainerController(TestRingBase):
         self.check_response(resp, root_resp_hdrs,
                             expected_objects=expected_objects)
 
-        # GET all objects in reverse
+        # GET all objects in reverse and *blank* limit
         mock_responses = [
             # status, body, headers
             (200, list(reversed(sr_dicts)), root_shard_resp_hdrs),
+            (200, list(reversed(sr_objs[4])), shard_resp_hdrs[4]),
+            (200, list(reversed(sr_objs[3])), shard_resp_hdrs[3]),
             (200, list(reversed(sr_objs[2])), shard_resp_hdrs[2]),
             (200, list(reversed(sr_objs[1])), shard_resp_hdrs[1]),
             (200, list(reversed(sr_objs[0])), shard_resp_hdrs[0]),
@@ -609,21 +680,38 @@ class TestContainerController(TestRingBase):
         expected_requests = [
             # path, headers, params
             ('a/c', {'X-Backend-Record-Type': 'auto'},
-             dict(states='listing', reverse='true')),
-            (shard_ranges[2].name, {'X-Backend-Record-Type': 'auto'},
-             dict(marker='', end_marker='pie', reverse='true',
-                  limit=str(limit), states='listing')),  # 200
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},
+             dict(states='listing', reverse='true', limit='')),
+            (wsgi_quote(str_to_wsgi(shard_ranges[4].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='', end_marker='\xf0\x9f\x8c\xb4', states='listing',
+                  reverse='true', limit=str(limit))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='\xf0\x9f\x8c\xb5', end_marker='\xe2\x98\x83',
+                  states='listing', reverse='true',
+                  limit=str(limit - len(sr_objs[4])))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[2].name)),
+             {'X-Backend-Record-Type': 'auto'},
+             dict(marker='\xe2\x98\x84', end_marker='pie', states='listing',
+                  reverse='true',
+                  limit=str(limit - len(sr_objs[4] + sr_objs[3])))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},
              dict(marker='q', end_marker='ham', states='listing',
-                  reverse='true', limit=str(limit - len(sr_objs[2])))),  # 200
-            (shard_ranges[0].name, {'X-Backend-Record-Type': 'auto'},
+                  reverse='true',
+                  limit=str(limit - len(sr_objs[4] + sr_objs[3]
+                                        + sr_objs[2])))),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[0].name)),
+             {'X-Backend-Record-Type': 'auto'},
              dict(marker='i', end_marker='', states='listing', reverse='true',
-                  limit=str(limit - len(sr_objs[2] + sr_objs[1])))),  # 200
+                  limit=str(limit - len(sr_objs[4] + sr_objs[3] + sr_objs[2]
+                                        + sr_objs[1])))),  # 200
         ]
 
         resp = self._check_GET_shard_listing(
             mock_responses, list(reversed(expected_objects)),
-            expected_requests, query_string='?reverse=true', reverse=True)
+            expected_requests, query_string='?reverse=true&limit=',
+            reverse=True)
         # root object count will overridden by actual length of listing
         self.check_response(resp, root_resp_hdrs,
                             expected_objects=expected_objects)
@@ -643,15 +731,18 @@ class TestContainerController(TestRingBase):
              dict(limit=str(limit), states='listing')),  # 404
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(limit=str(limit), states='listing')),  # 200
-            (shard_ranges[0].name, {'X-Backend-Record-Type': 'auto'},  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[0].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
              dict(marker='', end_marker='ham\x00', states='listing',
                   limit=str(limit))),
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
              dict(marker='h', end_marker='pie\x00', states='listing',
                   limit=str(limit - len(sr_objs[0])))),
-            (shard_ranges[2].name, {'X-Backend-Record-Type': 'auto'},   # 200
-             dict(marker='p', end_marker='', states='listing',
-                  limit=str(limit - len(sr_objs[0] + sr_objs[1]))))
+            (wsgi_quote(str_to_wsgi(shard_ranges[2].name)),
+             {'X-Backend-Record-Type': 'auto'},   # 200
+             dict(marker='p', end_marker='\xe2\x98\x83\x00', states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1])))),
         ]
         resp = self._check_GET_shard_listing(
             mock_responses, expected_objects, expected_requests,
@@ -659,31 +750,35 @@ class TestContainerController(TestRingBase):
         self.check_response(resp, root_resp_hdrs)
 
         # GET with marker
-        marker = sr_objs[1][2]['name']
-        first_included = len(sr_objs[0]) + 2
+        marker = bytes_to_wsgi(sr_objs[3][2]['name'].encode('utf8'))
+        first_included = (len(sr_objs[0]) + len(sr_objs[1])
+                          + len(sr_objs[2]) + 2)
         limit = CONTAINER_LISTING_LIMIT
         expected_objects = all_objects[first_included:]
         mock_responses = [
             (404, '', {}),
-            (200, sr_dicts[1:], root_shard_resp_hdrs),
+            (200, sr_dicts[3:], root_shard_resp_hdrs),
             (404, '', {}),
-            (200, sr_objs[1][2:], shard_resp_hdrs[1]),
-            (200, sr_objs[2], shard_resp_hdrs[2])
+            (200, sr_objs[3][2:], shard_resp_hdrs[3]),
+            (200, sr_objs[4], shard_resp_hdrs[4]),
         ]
         expected_requests = [
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(marker=marker, states='listing')),  # 404
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(marker=marker, states='listing')),  # 200
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 404
-             dict(marker=marker, end_marker='pie\x00', states='listing',
-                  limit=str(limit))),
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 200
-             dict(marker=marker, end_marker='pie\x00', states='listing',
-                  limit=str(limit))),
-            (shard_ranges[2].name, {'X-Backend-Record-Type': 'auto'},  # 200
-             dict(marker='p', end_marker='', states='listing',
-                  limit=str(limit - len(sr_objs[1][2:])))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker=marker, end_marker='\xf0\x9f\x8c\xb4\x00',
+                  states='listing', limit=str(limit))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker=marker, end_marker='\xf0\x9f\x8c\xb4\x00',
+                  states='listing', limit=str(limit))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[4].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker='\xe2\xa8\x83', end_marker='', states='listing',
+                  limit=str(limit - len(sr_objs[3][2:])))),
         ]
         resp = self._check_GET_shard_listing(
             mock_responses, expected_objects, expected_requests,
@@ -691,48 +786,95 @@ class TestContainerController(TestRingBase):
         self.check_response(resp, root_resp_hdrs)
 
         # GET with end marker
-        end_marker = sr_objs[1][6]['name']
-        first_excluded = len(sr_objs[0]) + 6
+        end_marker = bytes_to_wsgi(sr_objs[3][6]['name'].encode('utf8'))
+        first_excluded = (len(sr_objs[0]) + len(sr_objs[1])
+                          + len(sr_objs[2]) + 6)
         expected_objects = all_objects[:first_excluded]
         mock_responses = [
             (404, '', {}),
-            (200, sr_dicts[:2], root_shard_resp_hdrs),
+            (200, sr_dicts[:4], root_shard_resp_hdrs),
             (200, sr_objs[0], shard_resp_hdrs[0]),
             (404, '', {}),
-            (200, sr_objs[1][:6], shard_resp_hdrs[1])
+            (200, sr_objs[1], shard_resp_hdrs[1]),
+            (200, sr_objs[2], shard_resp_hdrs[2]),
+            (404, '', {}),
+            (200, sr_objs[3][:6], shard_resp_hdrs[3]),
         ]
         expected_requests = [
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(end_marker=end_marker, states='listing')),  # 404
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(end_marker=end_marker, states='listing')),  # 200
-            (shard_ranges[0].name, {'X-Backend-Record-Type': 'auto'},  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[0].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
              dict(marker='', end_marker='ham\x00', states='listing',
                   limit=str(limit))),
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 404
-             dict(marker='h', end_marker=end_marker, states='listing',
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 404
+             dict(marker='h', end_marker='pie\x00', states='listing',
                   limit=str(limit - len(sr_objs[0])))),
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 200
-             dict(marker='h', end_marker=end_marker, states='listing',
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker='h', end_marker='pie\x00', states='listing',
                   limit=str(limit - len(sr_objs[0])))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[2].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker='p', end_marker='\xe2\x98\x83\x00', states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1])))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 404
+             dict(marker='\xd1\xb0', end_marker=end_marker, states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1]
+                                        + sr_objs[2])))),
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
+             dict(marker='\xd1\xb0', end_marker=end_marker, states='listing',
+                  limit=str(limit - len(sr_objs[0] + sr_objs[1]
+                                        + sr_objs[2])))),
         ]
         resp = self._check_GET_shard_listing(
             mock_responses, expected_objects, expected_requests,
             query_string='?end_marker=%s' % end_marker)
         self.check_response(resp, root_resp_hdrs)
 
+        # GET with prefix
+        prefix = 'hat'
+        # they're all 1-character names; the important thing
+        # is which shards we query
+        expected_objects = []
+        mock_responses = [
+            (404, '', {}),
+            (200, sr_dicts, root_shard_resp_hdrs),
+            (200, [], shard_resp_hdrs[1]),
+        ]
+        expected_requests = [
+            ('a/c', {'X-Backend-Record-Type': 'auto'},
+             dict(prefix=prefix, states='listing')),  # 404
+            ('a/c', {'X-Backend-Record-Type': 'auto'},
+             dict(prefix=prefix, states='listing')),  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[1].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 404
+             dict(prefix=prefix, marker='', end_marker='pie\x00',
+                  states='listing', limit=str(limit))),
+        ]
+        resp = self._check_GET_shard_listing(
+            mock_responses, expected_objects, expected_requests,
+            query_string='?prefix=%s' % prefix)
+        self.check_response(resp, root_resp_hdrs)
+
         # marker and end_marker and limit
         limit = 2
         expected_objects = all_objects[first_included:first_excluded]
         mock_responses = [
-            (200, sr_dicts[1:2], root_shard_resp_hdrs),
-            (200, sr_objs[1][2:6], shard_resp_hdrs[1])
+            (200, sr_dicts[3:4], root_shard_resp_hdrs),
+            (200, sr_objs[3][2:6], shard_resp_hdrs[1])
         ]
         expected_requests = [
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(states='listing', limit=str(limit),
                   marker=marker, end_marker=end_marker)),  # 200
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
              dict(marker=marker, end_marker=end_marker, states='listing',
                   limit=str(limit))),
         ]
@@ -742,17 +884,18 @@ class TestContainerController(TestRingBase):
             % (marker, end_marker, limit))
         self.check_response(resp, root_resp_hdrs)
 
-        # reverse with marker, end_marker
+        # reverse with marker, end_marker, and limit
         expected_objects.reverse()
         mock_responses = [
-            (200, sr_dicts[1:2], root_shard_resp_hdrs),
-            (200, list(reversed(sr_objs[1][2:6])), shard_resp_hdrs[1])
+            (200, sr_dicts[3:4], root_shard_resp_hdrs),
+            (200, list(reversed(sr_objs[3][2:6])), shard_resp_hdrs[1])
         ]
         expected_requests = [
             ('a/c', {'X-Backend-Record-Type': 'auto'},
              dict(marker=end_marker, reverse='true', end_marker=marker,
                   limit=str(limit), states='listing',)),  # 200
-            (shard_ranges[1].name, {'X-Backend-Record-Type': 'auto'},  # 200
+            (wsgi_quote(str_to_wsgi(shard_ranges[3].name)),
+             {'X-Backend-Record-Type': 'auto'},  # 200
              dict(marker=end_marker, end_marker=marker, states='listing',
                   limit=str(limit), reverse='true')),
         ]
@@ -1010,7 +1153,7 @@ class TestContainerController(TestRingBase):
             'X-Backend-Sharding-State': 'unsharded',
             'X-Container-Object-Count': 0,
             'X-Container-Bytes-Used': 0,
-            'X-Container-Meta-Flavour': 'flavour%d' % i,
+            'X-Container-Meta-Flavour': 'flavour',
             'X-Backend-Storage-Policy-Index': 0}
 
         # empty first shard range
